{******************************************************************************}
{                                                                              }
{ Project UML Component Factory                                                }
{                                                                              }
{ The contents of this file are subject to the Mozilla Public License Version  }
{ 1.1 (the "License"); you may not use this file except in compliance with the }
{ License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ }
{                                                                              }
{ Software distributed under the License is distributed on an "AS IS" basis,   }
{ WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for }
{ the specific language governing rights and limitations under the License.    }
{                                                                              }
{ The Original Code is UCFNote.pas                                             }
{                                                                              }
{ The Initial Developer of the Original Code is Thierry.Coq                    }
{ Portions created by this individual are Copyright (C)                        }
{ 2000 of this individual.                                                     }
{                                                                              }
{ Last modified: October 18, 2000                                              }
{ Version 1.0                                                                  }
{                                                                              }
{******************************************************************************}
unit UCFNote;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

const
   DEFAULT_TAB_SIZE = 12;

type
  TPoints = array of TPoint;
  TUCFNote = class(TGraphicControl)
  private
    FText : TStringList;
    FNoteTabSize: Integer;
    FParentBrush: Boolean;
    FBorder : TPoints;
    FNoteTabPoints : TPoints;
    FComputingSize : Boolean;
    function GetText: TStrings;
    procedure SetText(const Value: TStrings);
    procedure SetNoteTabSize(const Value: Integer);
    procedure SetFont(const Value: TFont);
    function GetFont: TFont;
    function GetBrush: TBrush;
    procedure SetBrush(const Value: TBrush);
    procedure SetParentBrush(const Value: Boolean);
    { Dclarations prives }
  protected
    { Dclarations protges }
    procedure AdjustSize; override;
    procedure BrushHasChanged( anObject : TObject); virtual;
     function CanResize(var NewWidth, NewHeight: Integer): Boolean; override;
    procedure ComputeBorder; virtual;
    procedure ComputeSize; virtual;
    procedure Paint; override;
    procedure PaintBack; virtual;
    procedure PaintBorder; virtual;
    procedure PaintText; virtual;
    procedure ChangeText( Sender : TObject); virtual;

  public
    { Dclarations publiques }
    constructor Create( anOwner : TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
  published
    { Dclarations publies }
    property AutoSize;
    property Brush : TBrush read GetBrush write SetBrush;
    property Font : TFont read GetFont write SetFont;
    property NoteTabSize : Integer read FNoteTabSize write SetNoteTabSize
      default DEFAULT_TAB_SIZE;
    property ParentBrush : Boolean read FParentBrush write SetParentBrush
      default True;
    property Text : TStrings read GetText write SetText;
  end;

procedure Register;
function CompareBrushes( Brush1, Brush2 : TBrush) : Boolean;


implementation


procedure Register;
begin
  RegisterComponents('UCF', [TUCFNote]);
end;

{ TUCFNote }

procedure TUCFNote.AdjustSize;
begin
  inherited;
  ComputeSize;
end;

procedure TUCFNote.BrushHasChanged( anObject: TObject);
begin
  Self.Update;
end;

function TUCFNote.CanResize(var NewWidth, NewHeight: Integer): Boolean;
begin
  result := true;
  //result := FComputingSize or not AutoSize;
end;

procedure TUCFNote.ChangeText(Sender: TObject);
begin
  if AutoSize then
    AdjustSize;
  Self.Invalidate;
end;

procedure TUCFNote.ComputeBorder;
var
  x1, y1, x2, y2, x3, y3 : integer;
begin
  x1 := 0; y1 := 0; // upper, left
  x3 := Width-1; y3 := Height-1; // lower, right
  x2 := x3 -FNoteTabSize; // upper, right start of node id.
  y2 := FNoteTabSize; // upper, right, end of note id.

  SetLength( FBorder, 5);
  FBorder[0] := Point(x1,y1);
  FBorder[1] := Point(x2,y1);
  FBorder[2] := Point(x3, y2);
  FBorder[3] := Point(x3,y3);
  FBorder[4] := Point(x1,y3);

  SetLength( FNoteTabPoints, 3);
  FNoteTabPoints[0] := Point( x2, y1);
  FNoteTabPoints[1] := Point( x2, y2);
  FNoteTabPoints[2] := Point( x3, y2);
end;

procedure TUCFNote.ComputeSize;
var
  aLine : String;
  iLine : Integer;
  wt, ht : Integer;
  x, y : integer;
begin
  if FComputingSize then Exit;
  FComputingSize := true;
  try
    x:= 1;
    y:= 1;
    for iLine := 0 to FText.Count-1 do
    begin
      aLine := FText[ iLine];
      wt := Canvas.TextWidth( aLine);
      ht := Canvas.TextHeight( aLine);
      if x<wt then x := wt;
      inc( y, ht);
    end;
    inc( y, 3); inc( x, FNoteTabSize);
    if x < FNoteTabSize *2 then
      x := FNoteTabSize *2;
    if y < FNoteTabSize *2 then
      y := FNoteTabSize *2;
    SetBounds( Left, Top, X, Y);
  finally
    FComputingSize := false;
  end;
end;

constructor TUCFNote.Create(anOwner: TComponent);
begin
  inherited;
  FComputingSize := False;
  FNoteTabSize := DEFAULT_TAB_SIZE;
  FParentBrush := True;
  FText := TStringList.Create;
  FText.OnChange := Self.ChangeText;
  Width  := 200;
  Height := 100;

  SetLength( FBorder, 0);
  SetLength( FNoteTabPoints, 0);
end;

destructor TUCFNote.Destroy;
begin
  inherited;
  FreeAndNil( FText);
end;

function TUCFNote.GetBrush: TBrush;
begin
  result := Canvas.Brush;
end;

function TUCFNote.GetFont: TFont;
begin
  result := Canvas.Font;
end;

function TUCFNote.GetText: TStrings;
begin
  Result := FText;
end;

procedure TUCFNote.Paint;
begin
  inherited;
  ComputeBorder;
  PaintBack;
  PaintText;
  PaintBorder;
end;

procedure TUCFNote.PaintBack;
begin
  if FParentBrush and Assigned( Parent) then
    Canvas.Brush := Parent.Brush;
  Canvas.Polygon( FBorder);
  Canvas.Polyline( FNoteTabPoints);
end;

procedure TUCFNote.PaintBorder;
begin
  Canvas.Polyline( FBorder);
end;

procedure TUCFNote.PaintText;
var
  aLine : String;
  iLine : Integer;
  w, h, ht : Integer;
  x, y : integer;
begin
  w:= Width-FNoteTabSize-1;
  h:= Height-2;
  x:= 1;
  y:= 1;
  for iLine := 0 to FText.Count-1 do
  begin
    aLine := FText[ iLine];
    ht := Canvas.TextHeight( aLine);
    Canvas.TextRect( Rect( x, y, w, h), x, y, aLine);
    inc( y, ht);
  end;
end;

procedure TUCFNote.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited;
  //ComputeBorder;
  //Invalidate;
end;

procedure TUCFNote.SetBrush(const Value: TBrush);
begin
  if CompareBrushes( Canvas.Brush, Value) then Exit;
  Canvas.Brush := Value;
  if assigned( Parent) then
    if CompareBrushes( Value, Parent.Brush) then
      FParentBrush := true
    else
      FParentBrush := false
  else
    FParentBrush := false;
  Self.Invalidate;
  Canvas.Brush.OnChange := Self.BrushHasChanged; // doesn't work, not cabled in
                                                 // Delphi VCL...
end;

procedure TUCFNote.SetFont(const Value: TFont);
begin
  Canvas.Font := Value;
  if AutoSize then ComputeSize;
  Self.Invalidate;
end;

procedure TUCFNote.SetNoteTabSize(const Value: Integer);
begin
  FNoteTabSize := Value;
  Self.Invalidate;
end;

procedure TUCFNote.SetParentBrush(const Value: Boolean);
begin
  if FParentBrush = Value then Exit;
  FParentBrush := Value;
  if Value and assigned( Parent) then
    Canvas.Brush := Parent.Brush;
  Self.Invalidate;
end;

procedure TUCFNote.SetText(const Value: TStrings);
begin
  FText.Assign( Value);
  if AutoSize then
    AdjustSize;
  Self.Invalidate;
end;

function CompareBrushes( Brush1, Brush2 : TBrush) : Boolean;
begin
  result := false;
  if Assigned( Brush1) and Assigned( Brush2) then
    if (Brush1.Color = Brush2.Color) and
       (Brush1.Style = Brush2.Style) then
       result := true;
end;

end.
