unit DynProps;

interface

uses
  Classes;

type
  TDynPropertyCollection = class;
  TDynPropertiesHolder = class;

  TDynProperty = class (TCollectionItem)
  private
    FPropertyName : string;
    FValue : string;
    FOnGetValue : TNotifyEvent;
    FOnSetValue : TNotifyEvent;
    function GetCollection : TDynPropertyCollection;
    procedure SetPropertyName (const Value : string);
    function GetValue : string;
    procedure SetValue (AValue : string);
  protected
    function GetDisplayName: string; override;
  public
    destructor Destroy; override;
    property Collection : TDynPropertyCollection read GetCollection;
  published
    property PropertyName : string read FPropertyName Write SetPropertyName;
    property Value : string read GetValue Write SetValue;
    property OnGetValue : TNotifyEvent read FOnGetValue Write FOnGetValue;
    property OnSetValue : TNotifyEvent read FOnSetValue Write FOnSetValue;
  end;

  TDynPropertyCollection = class (TCollection)
  private
    FDynPropertiesHolder : TDynPropertiesHolder;
    function GetItems (Index : Integer) : TDynProperty;
    procedure SetItems (Index : Integer; item : TDynProperty);
  protected
    function GetOwner : TPersistent; override;
  public
    function Add : TDynProperty;
    property Items [Index : Integer] : TDynProperty read GetItems Write SetItems; default;
    property DynPropertiesHolder : TDynPropertiesHolder read FDynPropertiesHolder;
  end;

  TDynPropertiesHolder = class (TComponent)
  private
    FDynProperties : TDynPropertyCollection;
    FAutoCreatePropertiesOnAccess : Boolean;
    procedure SetDynProperties (Value : TDynPropertyCollection);
    function GetValues (const PropName : string) : string;
    procedure SetValues (const PropName, AValue : string);
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    function FindProperty (const PropName : string) : TDynProperty;
    property Values [const PropName : string] : string read GetValues Write SetValues; default;
  published
    property AutoCreatePropertiesOnAccess : Boolean read FAutoCreatePropertiesOnAccess Write FAutoCreatePropertiesOnAccess default False;
    property DynProperties : TDynPropertyCollection read FDynProperties Write SetDynProperties;
  end;

implementation

uses
  SysUtils, PropsSaver;

resourcestring
  ERR_PropertyDoesnExists = 'Property doesn''t exists';

function DynPropToString (Obj : TObject; const PropertyName : string; PropItem : TPropInfoItem = nil) : string; forward;
procedure StringToDynProp (Obj : TObject; const PropertyName, Value : string; PropItem : TPropInfoItem = nil); forward;

{ TDynProperty }

destructor TDynProperty.Destroy;
begin
  if FPropertyName <> ''
    then UnRegisterNonPublishedProperty (TDynPropertiesHolder, FPropertyName);
  inherited;
end;

function TDynProperty.GetCollection;
begin
  Result := Collection as TDynPropertyCollection;
end;

function TDynProperty.GetDisplayName;
begin
  if FValue <> ''
    then Result := FPropertyName + '=' + FValue
    else Result := FPropertyName;
end;

procedure TDynProperty.SetPropertyName;
begin
  if Value <> FPropertyName
    then
    begin
      if FPropertyName <> ''
        then UnRegisterNonPublishedProperty (TDynPropertiesHolder, FPropertyName);
      FPropertyName := Value;
      if FPropertyName <> ''
        then RegisterNonPublishedProperty (TDynPropertiesHolder, FPropertyName, DynPropToString, StringToDynProp);
    end;
end;

function TDynProperty.GetValue;
begin
  Result := FValue;
  if Assigned (FOnGetValue)
    then FOnGetValue (Self);
end;

procedure TDynProperty.SetValue;
begin
  FValue := AValue;
  if Assigned (FOnSetValue)
    then FOnSetValue (Self);
end;

{ TDynPropertyCollection }

function TDynPropertyCollection.GetItems;
begin
  Result := inherited Items [Index] as TDynProperty;
end;

function TDynPropertyCollection.GetOwner;
begin
  Result := FDynPropertiesHolder;
end;

function TDynPropertyCollection.Add;
begin
  Result := inherited Add as TDynProperty;
end;

procedure TDynPropertyCollection.SetItems;
begin
  inherited Items [Index] := item;
end;

{ TDynPropertiesHolder }

constructor TDynPropertiesHolder.Create;
begin
  inherited;
  FDynProperties := TDynPropertyCollection.Create (TDynProperty);
  FDynProperties.FDynPropertiesHolder := Self;
  FAutoCreatePropertiesOnAccess := False;
end;

destructor TDynPropertiesHolder.Destroy;
begin
  FDynProperties.Free;
  inherited;
end;

procedure TDynPropertiesHolder.SetDynProperties;
begin
  FDynProperties.Assign (Value);
end;

function TDynPropertiesHolder.GetValues;
var
  Prop : TDynProperty;
begin
  Prop := FindProperty (PropName);
  if Prop <> nil
    then Result := Prop.Value
    else if FAutoCreatePropertiesOnAccess
      then with FDynProperties.Add do
        begin
          try
            PropertyName := PropName;
            Result := '';
          except
            Free;
            raise;
          end;
        end
      else raise Exception.Create (ERR_PropertyDoesnExists);
end;

procedure TDynPropertiesHolder.SetValues;
var
  Prop : TDynProperty;
begin
  Prop := FindProperty (PropName);
  if Prop <> nil
    then Prop.Value := AValue
    else if FAutoCreatePropertiesOnAccess
      then with FDynProperties.Add do
        begin
          try
            PropertyName := PropName;
            Value := AValue;
          except
            Free;
            raise;
          end;
        end
      else raise Exception.Create (ERR_PropertyDoesnExists);
end;

function TDynPropertiesHolder.FindProperty;
var
  i : Integer;
begin
  for i := 0 to FDynProperties.Count - 1 do
    if uppercase (FDynProperties [i].PropertyName) = uppercase (PropName)
      then
      begin
        Result := FDynProperties [i];
        Exit;
      end;
  Result := nil;
end;

{ Translation Functions }

function DynPropToString;
begin
  Result := (Obj as TDynPropertiesHolder).Values [PropertyName];
end;

procedure StringToDynProp;
begin
  (Obj as TDynPropertiesHolder).Values [PropertyName] := Value;
end;

end.
