{***************************************************************
 *
 * Unit Name: EStringGrid
 * Purpose  : Extends a normal drawgrid to provide for:
 *            exposing fixedcell click, cell click and
              double clicks
 *            exposing mouseenter and mouseleave events for cells
 *
 * Author   : Andrew Baylis (c)
 * History  : Ver 1.00 - July 24th,2000
 *
 ****************************************************************}

unit EDrawGrid;                              { TExtDrawGrid component. }
{ Extends the standard draw grid with mouse events }

interface

uses
    Windows,
    SysUtils,
    Messages,
    Classes,
    Graphics,
    Controls,
    Forms,
    Dialogs,
    Menus,
    StdCtrls,
    ExtCtrls,
    Grids,
    contnrs;

type
    { TMouseEnter / TMouseLeave sends the Cell X and Y coords NOT the
      screen X,Y values.
      TMouseClickEvent sends the type of button, status of shift key,
      screen X & Y coords relative to the top-left corner of the control,
      Cell X & Y values }
    TMouseEnterEvent = procedure(Sender: TObject; Shift: TShiftState; CellX,
        CellY:
        Integer) of object;
    TMouseLeaveEvent = procedure(Sender: TObject; Shift: TShiftState; CellX,
        CellY:
        Integer) of object;
    TMouseClickEvent = procedure(Sender: TObject; Button: TMouseButton; Shift:
        TShiftState;
        X, Y, CellX, CellY: Integer) of object;

    TExtDrawGrid = class(TDrawGrid)
    private
    { Private declarations }
        FOnMouseEnter: TMouseEnterEvent;
        FOnMouseLeave: TMouseLeaveEvent;
        FOnCellRightClick: TMouseClickEvent;
        FOnFixedRightClick: TMouseClickEvent;
        FCellBitmap: TBitmap;
        FPaintLock: integer;
        FCurrentCol, FCurrentRow: integer;
        FOnCellDblClick: TMouseClickEvent;
        FOnFixedDblClick: TMouseClickEvent;
        FOnCellLeftClick: TMouseClickEvent;
        FOnFixedLeftClick: TMouseClickEvent;
        procedure SetOnCellClick(const Value: TMouseClickEvent);
        procedure SetOnFixedCellClick(const Value: TMouseClickEvent);
        procedure SetOnCellDblClick(const Value: TMouseClickEvent);
        procedure SetOnFixedCellDblClick(const Value: TMouseClickEvent);
        procedure SetOnCellLeftClick(const Value: TMouseClickEvent);
        procedure SetOnFixedLeftClick(const Value: TMouseClickEvent);

    protected
    { Protected declarations }
        procedure TriggerMouseEnterEvent(Shift: TShiftState; X, Y: Integer);
            virtual;
        procedure TriggerMouseLeaveEvent(Shift: TShiftState; X, Y: Integer);
            virtual;
        procedure Paint; override;
        procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
        procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
            X, Y: Integer); override;
        procedure WMLeftDblClk(var Message: TMessage); message WM_LBUTTONDBLCLK;
        procedure WMLeftDown(var Message: TMessage); message WM_LBUTTONDOWN;
    public
    { Public declarations }
        constructor Create(AOwner: TComponent); override;
        destructor destroy; override;
        procedure LockPaint;                 // provides a way to stop painting
        procedure UnlockPaint;
        property CellBitmap:TBitmap read FCellBitmap;
        property CurrentCol:integer read FCurrentCol;
        property CurrentRow:integer read FCurrentRow;
    published
    { Published properties and events }
    // default cell horiz & vert alignment
        property OnMouseEnter: TMouseEnterEvent read FOnMouseEnter write
        FOnMouseEnter;
        property OnMouseLeave: TMouseLeaveEvent read FOnMouseLeave write
        FOnMouseLeave;
        property OnFixedRightClk: TMouseClickEvent read FOnFixedRightClick write
        SetOnFixedCellClick;
        property OnCellRightClk: TMouseClickEvent read FOnCellRightClick write
        SetOnCellClick;
        property OnCellDblClick: TMouseClickEvent read FOnCellDblClick write
        SetOnCellDblClick;
        property OnFixedDblClick: TMouseClickEvent read FOnFixedDblClick
        write SetOnFixedCellDblClick;
        property OnFixedLeftClick: TMouseClickEvent read FOnFixedLeftClick write
        SetOnFixedLeftClick;
        property OnCellLeftClick: TMouseClickEvent read FOnCellLeftClick write
        SetOnCellLeftClick;
    end;

    TExtDragObject = class(TDragControlObject)
    private
        FImages: TDragImageList;
    protected
        function GetDragImages: TDragImageList; override;
    public
        constructor Create(AControl: TControl); override;
        destructor destroy; override;
        procedure HideDragImage; override;
        procedure ShowDragImage; override;
        property ImageList: TDragImageList read FImages;
    end;

implementation

{ TExtDrawGrid }

constructor TExtDrawGrid.Create(AOwner: TComponent);
{ Creates an object of type TExtDrawGrid, and initializes properties. }
begin
    inherited Create(AOwner);
    FCellBitmap := TBitmap.Create;
    FixedCols := 1;
    FixedRows := 1;
    FPaintLock := 0;
    FCurrentCol := -1;
    FCurrentRow := -1;
    ControlStyle:=ControlStyle+[csDisplayDragImage];
end;

destructor TExtDrawGrid.destroy;
begin
    FCellBitmap.Free;
    inherited;
end;

procedure TExtDrawGrid.LockPaint;
begin
    Inc(FPaintLock);
end;

procedure TExtDrawGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
var
    Coord: TGridCoord;
begin
    inherited;
    Coord := MouseCoord(X, Y);

    if (FCurrentCol <> Coord.X) or (FCurrentRow <> Coord.Y) then
    begin
        if (FCurrentCol > -1) and (FCurrentRow > -1) then
            TriggerMouseLeaveEvent(Shift, FCurrentCol, FCurrentRow);
        if (Coord.X > -1) and (Coord.Y > -1) then
            TriggerMouseEnterEvent(Shift, Coord.X, Coord.Y);
        FCurrentCol := Coord.X;
        FCurrentRow := Coord.Y;
    end;
end;                                         { MouseMove }

procedure TExtDrawGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X,
    Y: Integer);
var
    coord: TGridCoord;
begin
    inherited;
    coord := MouseCoord(X, Y);
    if (coord.X >= 0) and (coord.Y >= 0) then
    begin
         FCurrentCol := Coord.X;
        FCurrentRow := Coord.Y;
        if ((coord.X < FixedCols) or (coord.Y < FixedRows)) then
        begin
            if (ssDouble in Shift) then
            begin
                if Assigned(FOnFixedDblClick) then
                    FOnFixedDblClick(self, button, shift, x, y, coord.X,
                        coord.Y);
            end
            else
            begin
                if (button = mbRight) and Assigned(FOnFixedRightClick) then
                    FOnFixedRightClick(self, button, shift, x, y, coord.X,
                        coord.Y);
                if (button = mbLeft) and Assigned(FOnFixedLeftClick) then
                    FOnFixedLeftClick(self, button, shift, x, y, coord.X,
                        coord.Y);
            end;
        end
        else if (ssDouble in Shift) then
        begin
            if Assigned(FOnCellDblClick) then
                FOnCellDblClick(self, button, shift, x, y, coord.X, coord.Y);
        end
        else
        begin
            if (Button = mbRight) and Assigned(FOnCellRightClick) then
                FOnCellRightClick(self, button, shift, x, y, coord.X, coord.Y);
            if (Button = mbLeft) and Assigned(FOnCellLeftClick) then
                FOnCellLeftClick(self, button, shift, x, y, coord.X, coord.Y);
        end;
    end;
end;

procedure TExtDrawGrid.Paint;
begin
    if (FPaintLock = 0) then inherited;
end;

procedure TExtDrawGrid.SetOnCellClick(const Value: TMouseClickEvent);
begin
    FOnCellRightClick := Value;
end;

procedure TExtDrawGrid.SetOnCellDblClick(const Value: TMouseClickEvent);
begin
    FOnCellDblClick := Value;
end;

procedure TExtDrawGrid.SetOnCellLeftClick(const Value: TMouseClickEvent);
begin
    FOnCellLeftClick := Value;
end;

procedure TExtDrawGrid.SetOnFixedCellClick(
    const Value: TMouseClickEvent);
begin
    FOnFixedRightClick := Value;
end;

procedure TExtDrawGrid.SetOnFixedCellDblClick(
    const Value: TMouseClickEvent);
begin
    FOnFixedDblClick := Value;
end;

procedure TExtDrawGrid.SetOnFixedLeftClick(const Value: TMouseClickEvent);
begin
    FOnFixedLeftClick := Value;
end;

procedure TExtDrawGrid.TriggerMouseEnterEvent(Shift: TShiftState; X, Y:
    Integer);
{ Triggers the OnMouseEnter event.  }
begin
    if assigned(FOnMouseEnter) then
        FOnMouseEnter(Self, Shift, X, Y);
end;                                         { TriggerMouseEnterEvent }

procedure TExtDrawGrid.TriggerMouseLeaveEvent(Shift: TShiftState; X, Y:
    Integer);
{ Triggers the OnMouseLeave event. }
begin
    if assigned(FOnMouseLeave) then
        FOnMouseLeave(Self, Shift, X, Y);
end;                                         { TriggerMouseLeaveEvent }

procedure TExtDrawGrid.UnlockPaint;
begin
    if (FPaintLock > 0) then Dec(FPaintLock);
    if (FPaintLock = 0) then Repaint;
end;

procedure TExtDrawGrid.WMLeftDblClk(var Message: TMessage);
var
    x, y: integer;
begin
    y := message.LParamHi;
    x := message.LParamLo;
    MouseUp(mbLeft, [ssDouble], x, y);
end;

procedure TExtDrawGrid.WMLeftDown(var Message: TMessage);
var
    coord: TGridCoord;
    arect: TRect;
    w, h: integer;
begin
    inherited; // allow for normal processing

    coord := MouseCoord(message.lParamLo, message.lParamHi);
    aRect := CellRect(coord.x, coord.y);
    w := arect.Right - arect.Left + 1;
    h := arect.Bottom - arect.Top + 1;
    Fcellbitmap.Width := w;
    FCellbitmap.Height := h;
    FCellbitmap.Canvas.CopyRect(Rect(0, 0, w - 1, h - 1),
    Canvas, aRect);
    FCurrentCol := Coord.X;
    FCurrentRow := Coord.Y;
    if DragDetect(Handle,Point(message.lParamLo, message.lParamHi)) then
       BeginDrag(True)
    else MouseUp(mbLeft,[],message.lParamLo, message.lParamHi);
end;

{ TExtDragObject }

constructor TExtDragObject.Create;
begin
    inherited;
    FImages := TDragImageList.Create(nil);
    FImages.ShareImages := False;
end;

destructor TExtDragObject.destroy;
begin
    FImages.Free;
    inherited;
end;

function TExtDragObject.GetDragImages: TDragImageList;
begin
    Result := FImages;
end;

procedure TExtDragObject.HideDragImage;
begin
    FImages.HideDragImage;
end;

procedure TExtDragObject.ShowDragImage;
begin
    FImages.ShowDragImage;
end;

end.

