{***************************************************************
 *
 * Unit Name: EStringGrid
 * Purpose  : Extends a normal stringgrid to provide for:
 *            individual cell colors / fonts / locking
 *            merging groups of cells
 *            exposing mouseenter and mouseleave events
 *
 * Author   : Andrew Baylis (c)
 * History  : Ver 1.00 - May 6th, 2000
 *            Ver 1.10 - June 3rd, 2000
                  AlignHoriz & AlignVert
                  OnCellClick
                  OnFixedCellClick
                  + correctly drawing fixed cells
              Ver 1.20 - June 20th, 2000
                  Locked cell fore/back colours
              Ver 1.30 - June 30th, 2000
                  Column properties
              Ver 1.31 - July 17th, 2000
                  Fixed some issues with grouping cells
                  Fixed bug in Column properties
                  Fixed memory allocation bug
 ****************************************************************}

unit EStringGrid;                            { TExtStringGrid component. }
{ Extends the standard string grid with mouse events + format cells }

interface

uses
    Windows,
    SysUtils,
    Messages,
    Classes,
    Graphics,
    Controls,
    Forms,
    Dialogs,
    Menus,
    StdCtrls,
    ExtCtrls,
    Grids,
    contnrs;

type
    EBlockOverlap = class(Exception);
    EOutOfRange = class(Exception);
    TExtHAlign = (a_left, a_center, a_right);
    TExtVAlign = (a_top, a_middle, a_bottom);
    TTextAttr = (EditControl, EndEllipsis, ExpandTabs);
    TTextAttrs = set of TTextAttr;
    { TMouseEnter / TMouseLeave sends the Cell X and Y coords NOT the
      screen X,Y values.
      TMouseClickEvent sends the type of button, status of shift key,
      screen X & Y coords relative to the top-left corner of the control,
      Cell X & Y values }
    TMouseEnterEvent = procedure(Sender: TObject; Shift: TShiftState; X, Y:
        Integer) of object;
    TMouseLeaveEvent = procedure(Sender: TObject; Shift: TShiftState; X, Y:
        Integer) of object;
    TMouseClickEvent = procedure(Sender: TObject; Button: TMouseButton; Shift:
        TShiftState;
        X, Y, CellX, CellY: Integer) of object;

    TCellProperties = class(TObject)
    private
        Fcolor: TColor;
        FHAlign: TExtHAlign;
        FVAlign: TExtVAlign;
        Ffont: TFont;
        FGroupBitmap: TBitmap;
        FGBitmapFresh: boolean;
        procedure Setcolor(const Value: TColor);
        procedure Setfont(const Value: TFont);
        procedure SetHAlign(const Value: TExtHAlign);
        procedure SetVAlign(const Value: TExtVAlign);
    public
        locked, merged: boolean;
        parentfont, parentcolor, parentalign: boolean;
        block: TGridRect;
        constructor create;
        destructor destroy; override;
        procedure MakeBitmap;
        property font: TFont read Ffont write Setfont;
        property color: TColor read Fcolor write Setcolor;
        property GroupBitmap: TBitmap read FGroupBitmap;
        property HAlign: TExtHAlign read FHAlign write SetHAlign;
        property VAlign: TExtVAlign read FVAlign write SetVAlign;
    end;

    TColumnProperties = class(TObject)
    private
        Ffont: TFont;
        Fcolor: TColor;
        FHAlign: TExtHAlign;
        FVAlign: TExtVAlign;
        procedure Setfont(const Value: TFont);
        procedure Setcolor(const Value: TColor);
        procedure SetHAlign(const Value: TExtHAlign);
        procedure SetVAlign(const Value: TExtVAlign);
    public
        parentfont, locked, parentcolor, parentalign: Boolean;
        FCellProps: TObjectList;
        constructor create;
        destructor destroy; override;
        procedure Clear;
        property font: TFont read Ffont write Setfont;
        property color: TColor read Fcolor write Setcolor;
        property HAlign: TExtHAlign read FHAlign write SetHAlign;
        property VAlign: TExtVAlign read FVAlign write SetVAlign;
    end;

    TLockedCell = class(TPersistent)
    private
        FBackColor: TColor;
        FFont: TFont;
        FUseLockedFont: Boolean;
        procedure SetBackColor(const Value: TColor);
        procedure SetForeColor(const Value: TColor);
        function GetForeColor: TColor;
        procedure SetUseLockedFont(const Value: Boolean);
        procedure SetFont(const Value: TFont);
    public
        constructor Create;
        destructor Destroy; override;
        procedure Assign(const Source: TLockedCell); reintroduce;
    published
        property ForeColor: TColor read GetForeColor write SetForeColor;
        property BackColor: TColor read FBackColor write SetBackColor;
        property LockedFont: TFont read FFont write SetFont;
        property UseLockedFont: Boolean read FUseLockedFont write
        SetUseLockedFont;
    end;

    TFixedFont = class(TPersistent)
    private
        FFixedFont: TFont;
        FUseFixed: Boolean;
        FParentAlign: Boolean;
        FAlignHoriz: TExtHAlign;
        FAlignVert: TExtVAlign;
        procedure SetAlignHoriz(const Value: TExtHAlign);
        procedure SetAlignVert(const Value: TExtVAlign);
        procedure SetFont(const Value: TFont);
    protected

    public
        constructor Create;
        destructor Destroy; override;
    published
        property FixedFont: TFont read FFixedFont write SetFont;
        property UseFixedFont: Boolean read FUseFixed write FUseFixed;
        property ParentAlign: Boolean read FParentAlign write FParentAlign;
        property AlignHoriz: TExtHAlign read FAlignHoriz write SetAlignHoriz;
        property AlignVert: TExtVAlign read FAlignVert write SetAlignVert;
    end;

    TExtStringGrid = class(TStringGrid)
    private
    { Private declarations }
        FOnMouseEnter: TMouseEnterEvent;
        FOnMouseLeave: TMouseLeaveEvent;
        FCurrentCol, FCurrentRow: Integer;
        FTextHAlign: TExtHAlign;
        FTextVAlign: TExtVAlign;
        FOnCellClick: TMouseClickEvent;
        FOnFixedCellClick: TMouseClickEvent;
        FLockedCell: TLockedCell;
        FOnAfterDrawCell: TDrawCellEvent;
        FPaintLock: integer;
        FColumnProps: TObjectList;
        FTextOptions: TTextAttrs;
        FFixedFont: TFixedFont;
        FOnBeforeDrawCell: TDrawCellEvent;
        function GetCellProps(ACol, ARow: integer): TCellProperties;
        function GetCellColor(ACol, ARow: integer): TColor;
        function GetCellFont(ACol, ARow: integer): TFont;
        function GetCellLocked(ACol, ARow: integer): boolean;
        procedure SetCellColor(ACol, ARow: integer; const Value: TColor);
        procedure SetCellFont(ACol, ARow: integer; const Value: TFont);
        procedure SetCellLocked(ACol, ARow: integer; const Value: boolean);
        procedure CheckIndexRange(Acol, ARow: integer);
        function GetCellHAlignment(ACol, ARow: integer): TExtHAlign;
        function GetCellVAlignment(ACol, ARow: integer): TExtVAlign;
        procedure SetAlignHoriz(const Value: TExtHAlign);
        procedure SetAlignVert(const Value: TExtVAlign);
        procedure SetCellHAlignment(ACol, ARow: integer;
            const Value: TExtHAlign);
        procedure SetCellVAlignment(ACol, ARow: integer;
            const Value: TExtVAlign);
        procedure SetOnCellClick(const Value: TMouseClickEvent);
        procedure SetOnFixedCellClick(const Value: TMouseClickEvent);
        procedure SetColCount(const Value: Integer);
        procedure SetRowCount(const Value: Integer);
        function GetRowCount: integer;
        function GetColCount: Integer;
        procedure SetOnAfterDrawCell(const Value: TDrawCellEvent);
        procedure SetTextOptions(const Value: TTextAttrs);
        function GetColumnProps(Index: integer): TColumnProperties;
        function ResizeGroupBitmap(CellProps: TCellProperties): TRect;
        procedure CopyGroup(ACol, ARow: longint; mRect: TRect;
            GroupProp: TCellProperties);
        function MakeCellProp(ACol, ARow: LongInt): TCellProperties;
        procedure SetOnBeforeDrawCell(const Value: TDrawCellEvent);
        procedure SetLockedCell(const Value: TLockedCell);
    procedure SetFixedFont(const Value: TFixedFont);
    protected
    { Protected declarations }
        procedure DrawCell(ACol, ARow: Longint; mRect: TRect;
            AState: TGridDrawState); override;
        function MouseCoord(X, Y: Integer): TGridCoord;
        function CanEditModify: Boolean; override;
        function CanEditShow: Boolean; override;
        procedure TopLeftChanged; override;
        procedure MouseDown(Button: TMouseButton;
            Shift: TShiftState; X, Y: Integer); override;
    { Event triggers: }
        procedure TriggerMouseEnterEvent(Shift: TShiftState; X, Y: Integer);
            virtual;
        procedure TriggerMouseLeaveEvent(Shift: TShiftState; X, Y: Integer);
            virtual;
        procedure Paint; override;
        function SelectCell(ACol, ARow: Longint): Boolean; override;
        procedure ColumnMoved(FromIndex, ToIndex: LongInt); override;
        procedure RowMoved(FromIndex, ToIndex: LongInt); override;
    public
    { Public declarations }
        constructor Create(AOwner: TComponent); override;
        destructor Destroy; override;
        procedure Clear;
        procedure ClearColumnProps;
        procedure GroupCells(left, top, right, bottom: integer);
        procedure UngroupCells(ACol, ARow: integer);
        procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
        procedure LockPaint;                 // provides a way to stop painting
        procedure UnlockPaint;
        property CellProperties[ACol, ARow: integer]: TCellProperties read
        GetCellProps;
        property CellFont[ACol, ARow: integer]: TFont read GetCellFont write
        SetCellFont;
        property CellLocked[ACol, ARow: integer]: boolean read GetCellLocked
        write SetCellLocked;
        property CellColor[ACol, ARow: integer]: TColor read GetCellColor write
        SetCellColor;
        property CellHAlignment[ACol, ARow: integer]: TExtHAlign read
        GetCellHAlignment write SetCellHAlignment;
        property CellVAlignment[ACol, ARow: integer]: TExtVAlign read
        GetCellVAlignment write SetCellVAlignment;
        property ColumnProperties[Index: integer]: TColumnProperties read
        GetColumnProps;
    published
    { Published properties and events }
    // default cell horiz & vert alignment
        property AlignHoriz: TExtHAlign read FTextHAlign write SetAlignHoriz;
        property AlignVert: TExtVAlign read FTextVAlign write SetAlignVert;
        property ColCount read GetColCount write SetColCount;
        property RowCount read GetRowCount write SetRowCount;
        property LockedCell: TLockedCell read FLockedCell write SetLockedCell;
        property FixedFont: TFixedFont read FFixedFont Write SetFixedFont;

        property TextOptions: TTextAttrs read FTextOptions write SetTextOptions;
        property OnMouseEnter: TMouseEnterEvent read FOnMouseEnter write
        FOnMouseEnter;
        property OnMouseLeave: TMouseLeaveEvent read FOnMouseLeave write
        FOnMouseLeave;
        property OnFixedCellClick: TMouseClickEvent read FOnFixedCellClick write
        SetOnFixedCellClick;
        property OnCellClick: TMouseClickEvent read FOnCellClick write
        SetOnCellClick;
        property OnAfterDrawCell: TDrawCellEvent read FOnAfterDrawCell write
        SetOnAfterDrawCell;
        property OnBeforeDrawCell: TDrawCellEvent read FOnBeforeDrawCell write
        SetOnBeforeDrawCell;
    end;

implementation

{ TCellProperties }

constructor TCellProperties.create;
begin
    inherited;
    locked := false;
    merged := false;
    parentfont := true;
    parentcolor := true;
    parentalign := true;
    ffont := TFont.Create;
    fvalign := a_top;
    fhalign := a_left;
    FGroupBitmap := nil;
end;

destructor TCellProperties.destroy;
begin
    ffont.Free;
    if Assigned(FGroupBitmap) then FGroupBitmap.free;
    inherited;
end;

procedure TCellProperties.MakeBitmap;
begin
    if not Assigned(FGroupBitmap) then
        FGroupBitmap := TBitmap.Create;
    FGBitmapFresh := false;
end;

procedure TCellProperties.Setcolor(const Value: TColor);
begin
    Fcolor := Value;
    parentcolor := false;
end;

procedure TCellProperties.Setfont(const Value: TFont);
begin
    Ffont.assign(Value);
    parentfont := false;
end;

procedure TCellProperties.SetHAlign(const Value: TExtHAlign);
begin
    FHAlign := Value;
    parentalign := false;
end;

procedure TCellProperties.SetVAlign(const Value: TExtVAlign);
begin
    FVAlign := Value;
    parentalign := false;
end;

{ TColumnProperties }

procedure TColumnProperties.Clear;
begin
    Color := clWindow;
    HAlign := a_left;
    VAlign := a_top;
    locked := False;
    parentcolor := True;
    parentfont := True;
    parentalign := True;
    if Assigned(FCellProps) then
    begin
        FCellProps.Free;
        FCellProps := nil;
    end;
end;

constructor TColumnProperties.create;
begin
    inherited;
    FFont := TFont.Create;
    FCellProps := nil;
    Clear;
end;

destructor TColumnProperties.destroy;
begin
    FFont.Free;
    if Assigned(FCellProps) then FCellProps.Free;
    inherited;
end;

procedure TColumnProperties.Setcolor(const Value: TColor);
begin
    Fcolor := Value;
    parentcolor := False;
end;

procedure TColumnProperties.Setfont(const Value: TFont);
begin
    Ffont.Assign(Value);
    parentfont := False;
end;

procedure TColumnProperties.SetHAlign(const Value: TExtHAlign);
begin
    FHAlign := Value;
    parentalign := False;
end;

procedure TColumnProperties.SetVAlign(const Value: TExtVAlign);
begin
    FVAlign := Value;
    parentalign := False;
end;

{ TExtStringGrid }

function TExtStringGrid.CanEditModify: Boolean;
var
    cellprops: TCellProperties;
begin
    cellprops := CellProperties[Col, Row];
    if (cellprops <> nil) then
    begin // don't allow merged cells to be edited
        if (cellprops.merged) or (Cellprops.locked)
            then
            result := false
        else
            result := inherited CanEditModify;
    end
    else
        result := inherited CanEditModify;
end;

function TExtStringGrid.CanEditShow: Boolean;
var
    cellprops: TCellProperties;
begin
    cellprops := CellProperties[Col, Row];
    if (cellprops <> nil) then
    begin // don't allow merged cells to be edited
        if (cellprops.merged) or (Cellprops.locked)
            then
            result := false
        else
            result := inherited CanEditShow;
    end
    else
        result := inherited CanEditShow;
end;

procedure TExtStringGrid.CheckIndexRange(ACol, ARow: integer);
begin
    if ((ACol >= 0) and (ACol < ColCount)
        and (ARow >= 0) and (ARow < RowCount)) then
        exit
    else
        raise EOutOfRange.CreateFmt('Grid coords [%d,%d] out of range', [ACol,
            ARow]);
end;

procedure TExtStringGrid.Clear;
var
    i: integer;
begin
    LockPaint;
    for i := 0 to ColCount - 1 do
    begin
        ColumnProperties[i].Clear;
        Cols[i].Clear;
    end;
    UnlockPaint;
end;

procedure TExtStringGrid.ClearColumnProps;
var
    i: integer;
begin
    for i := 0 to ColCount - 1 do
        ColumnProperties[i].Clear;
end;

procedure TExtStringGrid.ColumnMoved(FromIndex, ToIndex: Integer);
var
    oList: TObjectList;
    i: integer;
begin
    oList := ColumnProperties[FromIndex].FCellProps;
    if FromIndex < ToIndex then
    begin
        for i := FromIndex to ToIndex - 1 do
            ColumnProperties[i].FCellProps := ColumnProperties[i +
                1].FCellProps;
    end
    else
        for i := ToIndex to FromIndex - 1 do
            ColumnProperties[i + 1].FCellProps :=
                ColumnProperties[i].FCellProps;
    ColumnProperties[ToIndex].FCellProps := oList;
    inherited;
end;

procedure TExtStringGrid.CopyGroup(ACol, ARow: Integer; mRect: TRect;
    GroupProp: TCellProperties);
var
    i: integer;
    r: TRect;
    DrawInfo: TGridDrawInfo;
begin
    CalcDrawInfo(DrawInfo);
    with DrawInfo do
    begin
        r.left := Horz.EffectiveLineWidth;
        r.top := Vert.EffectiveLineWidth;

        if (groupProp.block.Right <> ACol) then
            Inc(mRect.Right, Horz.EffectiveLineWidth);
        if (groupProp.block.Bottom <> ARow) then
            Inc(mRect.Bottom, Vert.EffectiveLineWidth);
        for i := groupProp.block.Left to ACol - 1 do
            Inc(r.left, Horz.GetExtent(i) + Horz.EffectiveLineWidth);
        for i := groupProp.block.Top to ARow - 1 do
            Inc(r.top, Vert.GetExtent(i) + Vert.EffectiveLineWidth);
        r.right := r.Left + mRect.Right - mRect.left;
        r.Bottom := r.top + mRect.Bottom - mRect.top;
        // need next two statements to ensure r is a valid rect within the
        // bitmap
        if r.Bottom > groupProp.GroupBitmap.Height then
            r.Bottom := groupProp.GroupBitmap.Height;
        if r.Right > groupProp.GroupBitmap.Width then
            r.Right := groupProp.GroupBitmap.Width;
    end; // now r is rect on bitmap holding contents of cell
    Canvas.CopyMode := cmSrcCopy;
    Canvas.CopyRect(mRect, groupProp.GroupBitmap.Canvas, r);
end;

constructor TExtStringGrid.Create(AOwner: TComponent);
{ Creates an object of type TExtStringGrid, and initializes properties. }
var
    i: integer;
    o: TGridOptions;
begin
    inherited Create(AOwner);
    FCurrentCol := -1;
    FCurrentRow := -1;
    FixedCols := 1;
    FixedRows := 1;
    DefaultDrawing := false;
    VirtualView := true;
    o := Options;
    Exclude(o, goRangeSelect);
    Options := o;
    FTextHAlign := a_left;
    FTextVAlign := a_top;
    FLockedCell := TLockedCell.Create;
    FLockedCell.BackColor := clGreen;
    FFixedFont := TFixedFont.Create;
    FPaintLock := 0;
    FColumnProps := TObjectList.Create;
    FColumnProps.OwnsObjects := True;
    for i := 0 to ColCount - 1 do
        FColumnProps.Add(TColumnProperties.Create);
end;

destructor TExtStringGrid.Destroy;
begin
    Clear;
    FLockedCell.Free;
    FFixedFont.Free;
    FColumnProps.Free;
    inherited Destroy;
end;

procedure TExtStringGrid.DrawCell(ACol, ARow: Integer; mRect: TRect;
    AState: TGridDrawState);
var
    cellprop: TCellProperties;
    colProp: TColumnProperties;
    ARect, BRect: TRect;
    halign: TExtHAlign;
    valign: TExtVAlign;
    sTemp: string;
    vgap, hgap: integer;
    uFormat: Cardinal;
    nFont: TFont;
    nColor: TColor;
    PaintCanvas: TCanvas;
begin
    if Assigned(FOnBeforeDrawCell) then
        FOnBeforeDrawCell(Self, ACol, ARow, mRect, AState);
    if DefaultDrawing then
    begin
        inherited;
        exit;
    end;
    ARect := mRect;
    cellprop := CellProperties[ACol, ARow];
    sTemp := Cells[ACol, ARow];
    colProp := ColumnProperties[ACol];
    nFont := Font;
    nColor := Color;
    halign := FTextHAlign;
    valign := FTextVAlign;
    if (not colProp.parentfont) then nFont := colProp.Font;
    if (not colProp.parentcolor) then nColor := colProp.Color;
    if (not colProp.parentAlign) then
    begin
        halign := colProp.HAlign;
        valign := colProp.VAlign;
    end;
    PaintCanvas := Canvas;
    if (cellprop <> nil) and (cellprop.merged) then
    begin
    // properties of top left cell = region
        cellprop := CellProperties[cellprop.block.left, cellprop.block.top];
        if (cellprop.block.left = ACol) and (cellprop.block.Top = ARow) then
        begin
            ARect := ResizeGroupBitmap(cellprop);  // current cell is top-left of group
            cellprop.FGBitmapFresh := false;
            PaintCanvas := cellprop.GroupBitmap.Canvas;
        end;
    end;

    if (cellprop <> nil) then
    begin
        if (not cellProp.parentFont) then nFont := cellProp.Font;
        if (not cellProp.parentColor) then nColor := cellProp.Color;
        if (not cellProp.parentalign) then
        begin
            halign := cellprop.HAlign;
            valign := cellprop.VAlign;
        end;
    end;
    if (cellprop <> nil) and (cellprop.merged) and (cellprop.FGBitmapFresh) then
       // copy rect from bitmap onto screen;
        CopyGroup(ACol, ARow, mRect, cellprop)
    else
    begin
    // only draw stuff if not merged or if merge bitmap is not fresh

    // now assign colors and fonts
        PaintCanvas.Font.Assign(nFont);
        PaintCanvas.Brush.Color := nColor;

        // locked overrides color & font
        if ((cellprop <> nil) and (cellprop.locked)) or
            (colprop.locked) then
        begin
            PaintCanvas.Brush.Color := FLockedCell.BackColor;
            if FLockedCell.UseLockedFont then
                PaintCanvas.font.Assign(FLockedCell.LockedFont)  // use all font attributes
            else
                PaintCanvas.Font.Color := FLockedCell.LockedFont.Color;  // just change font color
        end;

    // focussed overrides all brush colors
        if (gdSelected in AState) and
            (not (gdFocused in AState) or
            ([goDrawFocusSelected, goRowSelect] * Options <> [])) then
        begin
            PaintCanvas.Brush.Color := clHighlight;
            PaintCanvas.Font.Color := clHighlightText;
        end;

    // check if cell is a fixed cell
        if gdFixed in AState then
        begin
            PaintCanvas.brush.color := fixedcolor;
            DrawEdge(PaintCanvas.Handle, ARect, BDR_RAISEDINNER,
                BF_BOTTOMRIGHT);
            DrawEdge(PaintCanvas.Handle, ARect, BDR_RAISEDINNER, BF_TOPLEFT);
            InflateRect(ARect, -1, -1);
            PaintCanvas.FillRect(ARect);
            InflateRect(ARect, -1, -1);
            if FixedFont.UseFixedFont then
                PaintCanvas.Font.Assign(FixedFont.FixedFont);
            if (not FixedFont.ParentAlign) then
            begin
                halign := FixedFont.AlignHoriz;
                valign := FixedFont.AlignVert;
            end;
        end
        else
        begin
    // blank out cell region
            PaintCanvas.FillRect(ARect);
            if gdFocused in AState then
                PaintCanvas.DrawFocusRect(Arect); // put focus rectangle around
            InflateRect(ARect, -2, -2);
        end;

    // now draw text in region
        if (halign <> a_left) or (valign <> a_top) then
        begin
            BRect := ARect;

        // need to find size of text
            DrawText(PaintCanvas.handle, PChar(sTemp), -1, BRect, DT_CALCRECT or
                DT_WORDBREAK);
            vgap := ARect.Bottom - BRect.Bottom;
            hgap := ARect.Right - BRect.Right;
            if (vgap > 0) then
                case valign of
                    a_middle:
                        begin
                            Inc(ARect.Top, vgap div 2);
                            Dec(ARect.Bottom, vgap div 2);
                        end;
                    a_bottom: Inc(ARect.top, vgap);
                end;                         // case
            if (hgap > 0) then
                case halign of
                    a_right: Inc(ARect.Left, hgap);
                    a_center:
                        begin
                            Inc(ARect.Left, hgap div 2);
                            Dec(ARect.Right, hgap div 2);
                        end;
                end;                         //case
        end;

    // DT_WORDBREAK wraps words in paragraph
        uFormat := DT_WORDBREAK;
        if EditControl in FTextOptions then
            uFormat := uFormat or DT_EDITCONTROL;
        if EndEllipsis in FTextOptions then
            uFormat := uFormat or DT_END_ELLIPSIS;
        if ExpandTabs in FTextOptions then uFormat := uFormat or DT_EXPANDTABS;
        DrawText(PaintCanvas.handle, PChar(sTemp), -1, ARect, uFormat);
        if (cellprop <> nil) and (cellprop.merged) then
        begin
            cellprop.FGBitmapFresh := true;
            CopyGroup(ACol, ARow, mRect, cellprop);
        end;
    end;
    if Assigned(FOnAfterDrawCell) then
        FOnAfterDrawCell(Self, ACol, ARow, mRect, AState);
end;

function TExtStringGrid.GetCellColor(ACol, ARow: integer): TColor;
var
    cellprop: TCellProperties;
    colProp: TColumnProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    colProp := ColumnProperties[ACol];
    if (cellprop <> nil) and (not cellprop.Parentcolor) then
        result := cellprop.color
    else if (not colProp.parentcolor) then
        Result := colProp.Color
    else
        result := Color;                     // default is color of grid
end;

function TExtStringGrid.GetCellFont(ACol, ARow: integer): TFont;
var
    cellprop: TCellProperties;
    colProp: TColumnProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    colProp := ColumnProperties[ACol];
    if not Assigned(cellprop) then
       cellprop:=MakeCellProp(ACol,ARow);
    if not cellprop.ParentFont then
        result := cellprop.Font
    else if (not colProp.parentfont) then
        Result := colProp.Font
    else
        result := Font;                      // default is grid font
end;

function TExtStringGrid.GetCellHAlignment(ACol, ARow: integer): TExtHAlign;
var
    cellprop: TCellProperties;
    colProp: TColumnProperties;
begin
    CheckIndexRange(ACol, ARow);
    colProp := ColumnProperties[ACol];
    cellprop := CellProperties[ACol, ARow];
    if (cellprop <> nil) then
        result := cellprop.HAlign
    else if (not colProp.parentalign) then
        Result := colProp.HAlign
    else
        result := AlignHoriz;                // grid default
end;

function TExtStringGrid.GetCellLocked(ACol, ARow: integer): boolean;
var
    cellprop: TCellProperties;
    colProp: TColumnProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    colProp := ColumnProperties[ACol];
    if (cellprop <> nil) then
        result := cellprop.locked
    else
        result := colprop.locked;
end;

function TExtStringGrid.GetCellProps(ACol, ARow: integer): TCellProperties;
begin
    CheckIndexRange(ACol, ARow);             // raise exception if out of range
    with ColumnProperties[ACol] do
        if (Assigned(FCellProps)) and (ARow < FCellProps.Count) then
            result := TCellProperties(FCellProps[ARow])
        else
            result := nil;
    if Result=nil then Result:=MakeCellProp(ACol,ARow);
end;

function TExtStringGrid.GetCellVAlignment(ACol, ARow: integer): TExtVAlign;
var
    cellprop: TCellProperties;
    colProp: TColumnProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    colProp := ColumnProperties[ACol];
    if (cellprop <> nil) then
        result := cellprop.vAlign
    else if (not colProp.parentalign) then
        Result := colProp.VAlign
    else
        result := AlignVert;                 // grid default
end;

function TExtStringGrid.GetColCount: Integer;
begin
    result := inherited ColCount;
end;

function TExtStringGrid.GetColumnProps(Index: integer): TColumnProperties;
begin
    Result := TColumnProperties(FColumnProps[Index]);
end;

function TExtStringGrid.GetRowCount: integer;
begin
    result := inherited RowCount;
end;

procedure TExtStringGrid.GroupCells(left, top, right, bottom: integer);
var
    i, j: integer;
    cellprops: TCellProperties;
    block: TGridRect;
begin
    CheckIndexRange(left, top);              // raise exception if exceed bounds
    CheckIndexRange(right, bottom);
    HideEditor; // in case editor is in group's block
    begin
        block.left := left;
        block.right := right;
        block.top := top;
        block.bottom := bottom;
        // check for overlap with other merge
        for i := left to right do
            for j := top to bottom do
            begin
                cellprops := CellProperties[i, j];
                if (cellprops <> nil) and (cellprops.merged) then
                // error - overlap of blocks
                    raise EBlockOverlap.Create('Blocks overlap!');
            end;
        // OK, no overlap
        for i := left to right do
            for j := top to bottom do
            begin
                cellprops := CellProperties[i, j];
                if cellprops = nil then
                    cellprops := MakeCellProp(i, j);
                cellprops.merged := true;
                cellprops.block := block;
            end;
    end;
    Refresh;
end;                                         { GroupCells }

procedure TExtStringGrid.LockPaint;
begin
    Inc(FPaintLock);
end;

function TExtStringGrid.MakeCellProp(ACol, ARow: Integer): TCellProperties;
begin
    with ColumnProperties[ACol] do
    begin
        if not Assigned(FCellProps) then FCellProps := TObjectList.Create;
        if FCellProps.Count <= ARow then FCellProps.Count := ARow + 1;
        if not Assigned(FCellProps[ARow]) then
            FCellProps[ARow] := TCellProperties.Create;
        result := TCellProperties(FCellProps[ARow]);
    end;
end;

function TExtStringGrid.MouseCoord(X, Y: Integer): TGridCoord;
var
    cellprops: TCellProperties;
begin
     // get actual co-ordinate
    result := inherited MouseCoord(X, Y);
    // check if cell the mouse is over is part of a merge
    // then change Coord.X, Y to top-left of merge
    if (result.X < 0) or (result.Y < 0) then exit;
    cellprops := CellProperties[Result.X, Result.Y];
    if (cellprops <> nil) and (cellprops.merged) then
    begin
        result.X := cellprops.block.Left;
        result.Y := cellprops.block.Top;
    end;
end;

procedure TExtStringGrid.MouseDown(Button: TMouseButton; Shift: TShiftState;
    X, Y: Integer);
var
    coord: TGridCoord;
begin
    inherited;
    coord := MouseCoord(X, Y);
    if (coord.X >= 0) and (coord.Y >= 0) then
    begin
        if ((coord.X < FixedCols) or (coord.Y < FixedRows)) then
        begin
            if Assigned(FOnFixedCellClick) then
                FOnFixedCellClick(self, button, shift, x, y, coord.X, coord.Y);
        end
        else if Assigned(FOnCellClick) then
            FOnCellClick(self, button, shift, x, y, coord.X, coord.Y);
    end;
end;

procedure TExtStringGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
var
    Coord: TGridCoord;
begin
    inherited;
    Coord := MouseCoord(X, Y);

    if (FCurrentCol <> Coord.X) or (FCurrentRow <> Coord.Y) then
    begin
        if (FCurrentCol > -1) and (FCurrentRow > -1) then
            TriggerMouseLeaveEvent(Shift, FCurrentCol, FCurrentRow);
        if (Coord.X > -1) and (Coord.Y > -1) then
            TriggerMouseEnterEvent(Shift, Coord.X, Coord.Y);
        FCurrentCol := Coord.X;
        FCurrentRow := Coord.Y;
    end;
end;                                         { MouseMove }

procedure TExtStringGrid.Paint;
begin
    if (FPaintLock = 0) then inherited;
end;

function TExtStringGrid.ResizeGroupBitmap(CellProps: TCellProperties): TRect;
var
    i: integer;
    hsize, vsize: integer;
    c: TCellProperties;
    DrawInfo: TGridDrawInfo;
begin
    Result := Rect(0, 0, 0, 0);              //blank result
    if (not CellProps.merged) then exit;
    c := CellProperties[Cellprops.block.Left, CellProps.block.Top];
    c.MakeBitmap;
    CalcDrawInfo(DrawInfo);
    with DrawInfo do
    begin
        hsize := 0;
        vsize := 0;
        for i := c.block.Left to c.block.Right do
            Inc(hsize, Horz.GetExtent(i) + Horz.EffectiveLineWidth);
        Dec(hsize, Horz.EffectiveLineWidth); // remove width of final grid line
        for i := c.block.Top to c.block.Bottom do
            Inc(vsize, Vert.GetExtent(i) + Vert.EffectiveLineWidth);
        Dec(vsize, Vert.EffectiveLineWidth); // remove width of bottom grid line
        c.GroupBitmap.Height := vsize;
        c.GroupBitmap.Width := hsize;
        result := Rect(0, 0, hsize, vsize);
    end;
end;

procedure TExtStringGrid.RowMoved(FromIndex, ToIndex: Integer);
var
    fromcell, tocell: TCellProperties;
    i, j: integer;
begin
    for i := 0 to ColCount - 1 do
    begin
        fromCell := CellProperties[i, FromIndex];
        toCell := CellProperties[i, ToIndex];
        if (fromCell <> nil) or (toCell <> nil) then
        begin
            with ColumnProperties[i].FCellProps do
            begin
                if FromIndex < ToIndex then
                begin
                    if Count < ToIndex then Count := ToIndex + 1;
                    for j := FromIndex to ToIndex - 1 do
                        items[j] := items[j + 1];
                end
                else
                begin
                    if Count < FromIndex then Count := FromIndex + 1;
                    for j := ToIndex to FromIndex - 1 do
                        items[j + 1] := items[j];
                end;
                items[ToIndex] := fromCell;
            end;
        end;
    end;
    inherited;
end;

function TExtStringGrid.SelectCell(ACol, ARow: Integer): Boolean;
var
    c: TCellProperties;
begin
    c := CellProperties[ACol, ARow];
    if (c <> nil) and (c.merged) then
        result := false
    else
        result := inherited SelectCell(ACol, ARow);
end;

procedure TExtStringGrid.SetAlignHoriz(const Value: TExtHAlign);
begin
    FTextHAlign := Value;
    Invalidate;
end;

procedure TExtStringGrid.SetAlignVert(const Value: TExtVAlign);
begin
    FTextVAlign := Value;
    Invalidate;
end;

procedure TExtStringGrid.SetCellColor(ACol, ARow: integer;
    const Value: TColor);
var
    cellprop: TCellProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    if (cellprop = nil) then
        cellprop := MakeCellProp(ACol, ARow);
    if CellProp.merged then
        cellprop := CellProperties[cellprop.block.Left, cellprop.block.top];
    cellprop.color := Value;
    InvalidateCell(ACol, ARow);
end;

procedure TExtStringGrid.SetCellFont(ACol, ARow: integer;
    const Value: TFont);
var
    cellprop: TCellProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    if (cellprop = nil) then
        cellprop := MakeCellProp(ACol, ARow);
    if CellProp.merged then
        cellprop := CellProperties[cellprop.block.Left, cellprop.block.top];

    cellprop.font.Assign(Value);
    InvalidateCell(ACol, ARow);
end;

procedure TExtStringGrid.SetCellHAlignment(ACol, ARow: integer;
    const Value: TExtHAlign);
var
    cellprop: TCellProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    if (cellprop = nil) then
        cellprop := MakeCellProp(ACol, ARow);
    if CellProp.merged then
        cellprop := CellProperties[cellprop.block.Left, cellprop.block.top];

    cellprop.halign := Value;
    InvalidateCell(ACol, ARow);
end;

procedure TExtStringGrid.SetCellLocked(ACol, ARow: integer;
    const Value: boolean);
var
    cellprop: TCellProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    if (cellprop = nil) then
        cellprop := MakeCellProp(ACol, ARow);
    if CellProp.merged then
        cellprop := CellProperties[cellprop.block.Left, cellprop.block.top];

    cellprop.locked := Value;
    InvalidateCell(ACol, ARow);
end;

procedure TExtStringGrid.SetCellVAlignment(ACol, ARow: integer;
    const Value: TExtVAlign);
var
    cellprop: TCellProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprop := CellProperties[ACol, ARow];
    if (cellprop = nil) then
        cellprop := MakeCellProp(ACol, ARow);
    if CellProp.merged then
        cellprop := CellProperties[cellprop.block.Left, cellprop.block.top];

    cellprop.valign := Value;
    InvalidateCell(ACol, ARow);
end;

procedure TExtStringGrid.SetColCount(const Value: Integer);
var
    i: integer;
begin
    if Value > ColCount then
        for i := ColCount to Value - 1 do
            FColumnProps.Add(TColumnProperties.Create);
    if Value < ColCount then
    begin
        for i := Value to ColCount - 1 do
            Cols[i].Clear;
        FColumnProps.Count := Value;
    end;
    inherited ColCount := Value;
end;

procedure TExtStringGrid.SetFixedFont(const Value: TFixedFont);
begin
  FFixedFont.Assign(Value);
end;

procedure TExtStringGrid.SetLockedCell(const Value: TLockedCell);
begin
    FLockedCell.Assign(Value);
end;

procedure TExtStringGrid.SetOnAfterDrawCell(const Value: TDrawCellEvent);
begin
    FOnAfterDrawCell := Value;
end;

procedure TExtStringGrid.SetOnBeforeDrawCell(const Value: TDrawCellEvent);
begin
    FOnBeforeDrawCell := Value;
end;

procedure TExtStringGrid.SetOnCellClick(const Value: TMouseClickEvent);
begin
    FOnCellClick := Value;
end;

procedure TExtStringGrid.SetOnFixedCellClick(
    const Value: TMouseClickEvent);
begin
    FOnFixedCellClick := Value;
end;

procedure TExtStringGrid.SetRowCount(const Value: Integer);
var
    i, j: integer;
begin
    if Value < RowCount then
    begin
        for i := Value to RowCount - 1 do
        begin
            for j := 0 to ColCount - 1 do
                with ColumnProperties[j] do
                    if Assigned(FCellProps) and (FCellProps.Count > Value)
                        then FCellProps.Count := Value;
            Rows[i].Clear;
        end;
    end;
    inherited RowCount := Value;
end;

procedure TExtStringGrid.SetTextOptions(const Value: TTextAttrs);
begin
    FTextOptions := Value;
    Refresh;
end;

procedure TExtStringGrid.TopLeftChanged;
begin
    HideEditor;
    inherited;
end;

procedure TExtStringGrid.TriggerMouseEnterEvent(Shift: TShiftState; X, Y:
    Integer);
{ Triggers the OnMouseEnter event.  }
begin
    if assigned(FOnMouseEnter) then
        FOnMouseEnter(Self, Shift, X, Y);
end;                                         { TriggerMouseEnterEvent }

procedure TExtStringGrid.TriggerMouseLeaveEvent(Shift: TShiftState; X, Y:
    Integer);
{ Triggers the OnMouseLeave event. }
begin
    if assigned(FOnMouseLeave) then
        FOnMouseLeave(Self, Shift, X, Y);
end;                                         { TriggerMouseLeaveEvent }

procedure TExtStringGrid.UngroupCells(ACol, ARow: integer);
var
    i, j: integer;
    cellprops, c: TCellProperties;
begin
    CheckIndexRange(ACol, ARow);
    cellprops := CellProperties[ACol, ARow];
    if (cellprops <> nil) and (cellprops.merged) then
        for i := cellprops.block.Left to cellprops.block.Right do
            for j := cellprops.block.Top to cellprops.block.Bottom do
            begin
                c := CellProperties[i, j];
                if (c <> nil) then c.merged := false;
            end;
    Refresh;
end;

procedure TExtStringGrid.UnlockPaint;
begin
    if (FPaintLock > 0) then Dec(FPaintLock);
    if (FPaintLock = 0) then Repaint;
end;

{ TFixedFont }

constructor TFixedFont.Create;
begin
    inherited;
    FFixedFont := TFont.Create;
    FUseFixed := False;
    FParentAlign := True;
    FAlignHoriz := a_left;
    FAlignVert := a_top;
end;

destructor TFixedFont.Destroy;
begin
    FFixedFont.Free;
    inherited;
end;

procedure TFixedFont.SetAlignHoriz(const Value: TExtHAlign);
begin
    FAlignHoriz := Value;
    FParentAlign := False;
end;

procedure TFixedFont.SetAlignVert(const Value: TExtVAlign);
begin
    FAlignVert := Value;
    FParentAlign := False;
end;

procedure TFixedFont.SetFont(const Value: TFont);
begin
    FFixedFont.Assign(Value);
end;

{ TLockedCell }

procedure TLockedCell.Assign(const Source: TLockedCell);
begin
    FFont.Assign(Source.FFont);
    FBackColor := Source.FBackColor;
    FUseLockedFont := Source.FUseLockedFont;
end;

constructor TLockedCell.Create;
begin
    inherited;
    FFont := TFont.Create;
    FBackColor := clWindow;
    FUseLockedFont := False;
end;

destructor TLockedCell.Destroy;
begin
    FFont.Free;
    inherited;
end;

function TLockedCell.GetForeColor: TColor;
begin
    Result := FFont.Color;
end;

procedure TLockedCell.SetBackColor(const Value: TColor);
begin
    FBackColor := Value;
end;

procedure TLockedCell.SetFont(const Value: TFont);
begin
    FFont.Assign(Value);
end;

procedure TLockedCell.SetForeColor(const Value: TColor);
begin
    FFont.Color := Value;
end;

procedure TLockedCell.SetUseLockedFont(const Value: Boolean);
begin
    FUseLockedFont := Value;
end;

end.

