{***************************************************************
 *
 * Unit Name : HotImage
 * Purpose   : Hot Image
 * Author    : Eugen Mihailescu
 * Company   : EM Quicksoft Romania SRL
 * Copyright : 1998,2002  All rights reserved.
 * Web Site  : http://www.em-quicksoft.com
 * Support   : support@em-quicksoft.com
 * History   : 03/19/2000
 *
 * Disclaimer: The objectives of this program is only educational.
 *             You MUST NOT use it without getting any material benefit on it
 *             You have the right to share this code with others without removing
 *             the copyright and/or disclaimer notice.
 *             You are allowed to include its code in your application
 *             only if it was designed for non commercial purpose,
 *             otherwise we claim $10 for each copy of your
 *             application which contains this code inside.
 ****************************************************************}

Unit HotImage;

Interface

Uses
  Windows,
  Messages,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  ExtCtrls,
  StdCtrls;

Type
  TOnProgress = Procedure(Sender: TObject; PercComplete: Integer) Of Object;
  PRGBTriple = ^TRGBTriple;
  TRGBTriple = Packed Record
    B: Byte;                            //easier to type than rgbtBlue...
    g: Byte;
    R: Byte;
  End;

  PRow = ^TRow;
  TRow = Array[0..1000000] Of TRGBTriple;

  PPRows = ^TPRows;
  TPRows = Array[0..1000000] Of PRow;

Const
  MaxKernelSize     = 100;

Type

  TKernelSize = 1..MaxKernelSize;

  TKernel = Record
    Size: TKernelSize;
    Weights: Array[-MaxKernelSize..MaxKernelSize] Of Single;
  End;
  //the idea is that when using a TKernel you ignore the Weights
  //except for Weights in the range -Size..Size.

  THotImage = Class(TImage)
  private
    fBackupCanvas: TCanvas;
    fProgress: TOnProgress;
    fOldPict: TPicture;
    fOrigHotPict: TPicture;
    fHotImage: TPicture;
    fLabel: TLabel;
    FColor: TColor;
    fBackupColor: TColor;
    fBackupImage: TPicture;
    fBlurSize: Word;
    fRotate: Boolean;
    { Private declarations }
    Procedure SetHotImage(Value: TPicture);
    Procedure SetBlurSize(Value: Word);
    Procedure SetFRotate(Value: Boolean);
  protected
    Procedure CMMouseEnter(Var Msg: TMessage); message CM_MOUSEENTER;
    Procedure CMMouseLeave(Var Msg: TMessage); message CM_MOUSELEAVE;
  public
    Constructor Create(AOwner: TComponent); override;
    Destructor Destroy; override;
    Procedure GBlur(theBitmap: TBitmap; radius: Double);
    Procedure RotateBitmap90Degrees(ABitmap: TBitmap);
    { Public declarations }
  published
    Property HotImage: TPicture read fHotImage write SetHotImage;
    Property TextLabel: TLabel read fLabel write fLabel;
    Property TextColor: TColor read FColor write FColor;
    Property BlurSize: Word read fBlurSize write SetBlurSize;
    Property OnBlurProgress: TOnProgress read fProgress write fProgress;
    Property Rotate90Degrees: Boolean read fRotate write SetFRotate;
    { Published declarations }
  End;

Procedure Register;

Implementation
//******************* MakeGaussianKernel *************************

Procedure MakeGaussianKernel(Var k: TKernel; radius: Double;
  MaxData, DataGranularity: Double);
//makes K into a gaussian kernel with standard deviation = radius.
//for the current application you set MaxData = 255,
//DataGranularity = 1. Now the procedure sets the value of
//K.Size so that when we use K we will ignore the Weights
//that are so small they can't possibly matter. (Small Size
//is good because the execution time is going to be
//propertional to K.Size.)
Var
  j                 : Integer;
  Temp, delta       : Double;
  KernelSize        : TKernelSize;
Begin
  For j := Low(k.Weights) To High(k.Weights) Do
    Begin
      Temp := j / radius;
      k.Weights[j] := exp(-Temp * Temp / 2);
    End;

  //now divide by constant so sum(Weights) = 1:

  Temp := 0;
  For j := Low(k.Weights) To High(k.Weights) Do
    Temp := Temp + k.Weights[j];
  For j := Low(k.Weights) To High(k.Weights) Do
    k.Weights[j] := k.Weights[j] / Temp;

  //now discard (or rather mark as ignorable by setting Size)
  //the entries that are too small to matter -
  //this is important, otherwise a blur with a small radius
  //will take as long as with a large radius...
  KernelSize := MaxKernelSize;
  delta := DataGranularity / (2 * MaxData);
  Temp := 0;
  While (Temp < delta) And (KernelSize > 1) Do
    Begin
      Temp := Temp + 2 * k.Weights[KernelSize];
      Dec(KernelSize);
    End;

  k.Size := KernelSize;

  //now just to be correct go back and jiggle again so the
  //sum of the entries we'll be using is exactly 1:

  Temp := 0;
  For j := -k.Size To k.Size Do
    Temp := Temp + k.Weights[j];
  For j := -k.Size To k.Size Do
    k.Weights[j] := k.Weights[j] / Temp;

End;
//******************* TrimInt *************************

Function TrimInt(Lower, Upper, theInteger: Integer): Integer;
Begin
  If (theInteger <= Upper) And (theInteger >= Lower) Then
    Result := theInteger
  Else
    If theInteger > Upper Then
    Result := Upper
  Else
    Result := Lower;
End;
//******************* TrimReal *************************

Function TrimReal(Lower, Upper: Integer; X: Double): Integer;
Begin
  If (X < upper) And (X >= lower) Then
    Result := Trunc(X)
  Else
    If X > Upper Then
    Result := Upper
  Else
    Result := Lower;
End;
//******************* BlurRow *************************

Procedure BlurRow(Var theRow: Array Of TRGBTriple; k: TKernel; p: PRow);
Var
  j, n, LocalRow    : Integer;
  tr, tg, tb        : Double;           //tempRed, etc
  W                 : Double;
Begin

  For j := 0 To High(theRow) Do
    Begin
      tb := 0;
      tg := 0;
      tr := 0;
      For n := -k.Size To k.Size Do
        Begin
          W := k.Weights[n];

          //the TrimInt keeps us from running off the edge of the row...
          With theRow[TrimInt(0, High(theRow), j - n)] Do
            Begin
              tb := tb + W * B;
              tg := tg + W * g;
              tr := tr + W * R;
            End;
        End;
      With p[j] Do
        Begin
          B := TrimReal(0, 255, tb);
          g := TrimReal(0, 255, tg);
          R := TrimReal(0, 255, tr);
        End;
    End;

  Move(p[0], theRow[0], (High(theRow) + 1) * SizeOf(TRGBTriple));
End;
//******************* THotImage.GBlur *************************

Procedure THotImage.GBlur(theBitmap: TBitmap; radius: Double);
Var
  Row, col          : Integer;
  theRows           : PPRows;
  k                 : TKernel;
  ACol              : PRow;
  p                 : PRow;
Begin
  If (theBitmap.HandleType <> bmDIB) Or (theBitmap.PixelFormat <> pf24Bit) Then
    Raise Exception.Create('GBlur only works for 24-bit bitmaps');

  MakeGaussianKernel(k, radius, 255, 1);
  GetMem(theRows, theBitmap.Height * SizeOf(PRow));
  GetMem(ACol, theBitmap.Height * SizeOf(TRGBTriple));

  //record the location of the bitmap data:
  For Row := 0 To theBitmap.Height - 1 Do
    theRows[Row] := theBitmap.Scanline[Row];

  //blur each row:
  p := AllocMem(theBitmap.Width * SizeOf(TRGBTriple));
  For Row := 0 To theBitmap.Height - 1 Do
    BlurRow(Slice(theRows[Row]^, theBitmap.Width), k, p);

  //now blur each column
  ReAllocMem(p, theBitmap.Height * SizeOf(TRGBTriple));
  For col := 0 To theBitmap.Width - 1 Do
    Begin
      Application.ProcessMessages;
      If Assigned(fProgress) Then
        fProgress(Nil, Trunc(100 * ((col + 1) / theBitmap.Width)));
      //- first read the column into a TRow:
      For Row := 0 To theBitmap.Height - 1 Do
        ACol[Row] := theRows[Row][col];

      BlurRow(Slice(ACol^, theBitmap.Height), k, p);

      //now put that row, um, column back into the data:
      For Row := 0 To theBitmap.Height - 1 Do
        theRows[Row][col] := ACol[Row];
    End;

  FreeMem(theRows);
  FreeMem(ACol);
  ReAllocMem(p, 0);
End;
//******************* THotImage.Create *************************

Constructor THotImage.Create(AOwner: TComponent);
Begin
  fBackupCanvas := TCanvas.Create;
  fHotImage := TPicture.Create;
  fBackupImage := TPicture.Create;
  fOldPict := TPicture.Create;
  fOrigHotPict := TPicture.Create;
  fBlurSize := 1;
  If Assigned(fProgress) Then
    fProgress(Nil, 0);
  Inherited Create(AOwner);
End;
//******************* THotImage.Destroy *************************

Destructor THotImage.Destroy;
Begin
  fHotImage.Free;
  fBackupImage.Free;
  fOldPict.Free;
  fOrigHotPict.Free;
  fBackupCanvas.Free;
  Inherited Destroy;
End;
//******************* THotImage.SetHotImage *************************

Procedure THotImage.SetHotImage(Value: TPicture);
Begin
  fHotImage.Assign(Value);
  fOrigHotPict.Bitmap.Canvas.Draw(0, 0, Value.Graphic);
End;
//******************* THotImage.CMMouseEnter *************************

Procedure THotImage.CMMouseEnter(Var Msg: TMessage);
Begin
  If Assigned(Picture) Then
    fBackupImage.Assign(Picture);
  If Assigned(fHotImage) Then
    Picture.Assign(fHotImage);

  If Assigned(textlabel) Then
    Begin
      fbackupcolor := flabel.Font.Color;
      flabel.Font.Color := textcolor;
      flabel.Font.Style := flabel.Font.Style + [fsBold]
    End;
End;
//******************* THotImage.CMMouseLeave *************************

Procedure THotImage.CMMouseLeave(Var Msg: TMessage);
Begin
  If Assigned(fBackupImage) Then
    Picture.Assign(fBackupImage);
  If Assigned(textlabel) Then
    Begin
      flabel.Font.Color := fbackupcolor;
      flabel.Font.Style := flabel.Font.Style - [fsBold]
    End;
End;
//******************* THotImage.SetBlurSize *************************

Procedure THotImage.SetBlurSize(Value: Word);
Begin
  Picture.Bitmap.PixelFormat := pf24Bit;
  GBlur(Picture.Bitmap, Value);
  Picture.Bitmap.Canvas.Draw(0, 0, Picture.Bitmap);
  fBlurSize := Value;
End;
//******************* THotImage.RotateBitmap90Degrees *************************

Procedure THotImage.RotateBitmap90Degrees(ABitmap: TBitmap);

Const
  BITMAP_HEADER_SIZE = SizeOf(TBitmapFileHeader) +
    SizeOf(TBitmapInfoHeader);

Var
  { Things that end in R are for the rotated image. }
  PbmpInfoR         : PBitmapInfoHeader;
  bmpBuffer, bmpBufferR: PByte;
  MemoryStream, MemoryStreamR: TMemoryStream;
  PbmpBuffer, PbmpBufferR: PByte;
  PbmpBufferRFirstScanLine, PbmpBufferColumnZero: PByte;
  BytesPerPixel, BytesPerScanLine, BytesPerScanLineR: Integer;
  X, Y, T           : Integer;

Begin
  {
    Don't *ever* call GetDIBSizes! It screws up your bitmap.
    I'll be posting about that shortly.
  }

  MemoryStream := TMemoryStream.Create;

  {
   To do: Put in a SetSize, which will eliminate any reallocation
   overhead.
  }

  ABitmap.SaveToStream(MemoryStream);

  {
   Don't need you anymore. We'll make a new one when the time comes.
  }
  ABitmap.Free;

  bmpBuffer := MemoryStream.Memory;

  { Set PbmpInfoR to point to the source bitmap's info header. }
  { Boy, these headers are getting annoying. }
  Inc(bmpBuffer, SizeOf(TBitmapFileHeader));
  PbmpInfoR := PBitmapInfoHeader(bmpBuffer);

  { Set bmpBuffer to point to the original bitmap bits. }
  Inc(bmpBuffer, SizeOf(PbmpInfoR^));
  { Set the ColumnZero pointer to point to, uh, column zero. }
  PbmpBufferColumnZero := bmpBuffer;

  With PbmpInfoR^ Do
    Begin
      BytesPerPixel := biBitCount Shr 3;
      { ScanLines are DWORD aligned. }
      BytesPerScanLine := ((((biWidth * biBitCount) + 31) Div 32) * SizeOf(DWORD));
      BytesPerScanLineR := ((((biHeight * biBitCount) + 31) Div 32) * SizeOf(DWORD));

      { The TMemoryStream that will hold the rotated bits. }
      MemoryStreamR := TMemoryStream.Create;
      {
       Set size for rotated bitmap. Might be different from source size
       due to DWORD aligning.
      }
      MemoryStreamR.SetSize(BITMAP_HEADER_SIZE + BytesPerScanLineR * biWidth);
    End;

  { Copy the headers from the source bitmap. }
  MemoryStream.Seek(0, soFromBeginning);
  MemoryStreamR.CopyFrom(MemoryStream, BITMAP_HEADER_SIZE);

  { Here's the buffer we're going to rotate. }
  bmpBufferR := MemoryStreamR.Memory;
  { Skip the headers, yadda yadda yadda... }
  Inc(bmpBufferR, BITMAP_HEADER_SIZE);

  {
   Set up PbmpBufferRFirstScanLine and advance it to end of the first scan
   line of bmpBufferR.
  }
  PbmpBufferRFirstScanLine := bmpBufferR;
  Inc(PbmpBufferRFirstScanLine, (PbmpInfoR^.biHeight - 1) * BytesPerPixel);

  { Here's the meat. Loop through the pixels and rotate appropriately. }

  { Remember that DIBs have their origins opposite from DDBs. }
  For Y := 1 To PbmpInfoR^.biHeight Do
    Begin
      PbmpBuffer := PbmpBufferColumnZero;
      PbmpBufferR := PbmpBufferRFirstScanLine;

      For X := 1 To PbmpInfoR^.biWidth Do
        Begin
          For T := 1 To BytesPerPixel Do
            Begin
              PbmpBufferR^ := PbmpBuffer^;
              Inc(PbmpBufferR);
              Inc(PbmpBuffer);
            End;
          Dec(PbmpBufferR, BytesPerPixel);
          Inc(PbmpBufferR, BytesPerScanLineR);
        End;

      Inc(PbmpBufferColumnZero, BytesPerScanLine);
      Dec(PbmpBufferRFirstScanLine, BytesPerPixel);
    End;

  { Done with the source bits. }
  MemoryStream.Free;

  { Now set PbmpInfoR to point to the rotated bitmap's info header. }
  PbmpBufferR := MemoryStreamR.Memory;
  Inc(PbmpBufferR, SizeOf(TBitmapFileHeader));
  PbmpInfoR := PBitmapInfoHeader(PbmpBufferR);

  { Swap the width and height of the rotated bitmap's info header. }
  With PbmpInfoR^ Do
    Begin
      T := biHeight;
      biHeight := biWidth;
      biWidth := T;
    End;

  ABitmap := TBitmap.Create;

  { Spin back to the very beginning. }
  MemoryStreamR.Seek(0, soFromBeginning);
  ABitmap.LoadFromStream(MemoryStreamR);

  MemoryStreamR.Free;
End;
//******************* THotImage.SetfRotate *************************

Procedure THotImage.SetfRotate(Value: Boolean);
Var
  B                 : TBitmap;
Begin
  If Not Value Then Exit;
  B := TBitmap.Create;
  Try
    B.Canvas.Draw(0, 0, Picture.Bitmap);
    Try
      RotateBitmap90Degrees(B);
    Except
    End;
    Picture.Bitmap.Canvas.Draw(0, 0, B);
  Finally
    B.Free;
  End;
  fRotate := Value;
End;

Procedure Register;
Begin
  RegisterComponents('EM-Quicksoft', [THotImage]);
End;

End.

