unit LayoutManager;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, TwoColsIntf,
  TwoColsComponent;

type
  TLayoutParameter = class;
  TLayoutItem = class;
  TLayoutManager = class;
  TLayoutItemCollection = class;
  TStringArray = array of string;
  PStringArray = ^TStringArray;

  TLayoutCollectionItem = class (TTwoColsCollectionItem)
  private
    function GetLayoutManager : TLayoutManager;
  public
    property LayoutManager : TLayoutManager read GetLayoutManager;
  end;

  TLayoutParameterCollection = class (TCollection)
  private
    FLayoutItem : TLayoutItem;
    function GetItems (Index : Integer) : TLayoutParameter;
    procedure SetItems (Index : Integer; item : TLayoutParameter);
  protected
    function GetOwner : TPersistent; override;
  public
    function Add : TLayoutParameter;
    property Items [Index : Integer] : TLayoutParameter read GetItems Write SetItems; default;
    property LayoutItem : TLayoutItem read FLayoutItem;
  end;

  TLayoutParameter = class (TLayoutCollectionItem)
  private
    FParameter : string;
    FValue : string;
    function GetLayoutParameterCollection : TLayoutParameterCollection;
    procedure SetValue (const AValue : string);
    procedure SetParameter (const AParameter : string);
  protected
    function GetTwoColsComponent : TTwoColsComponent; override;
    function GetDisplayName: string; override;
  public
    property Collection : TLayoutParameterCollection read GetLayoutParameterCollection;
  published
    property Parameter : string read FParameter Write SetParameter;
    property Value : string read FValue Write SetValue;
  end;

  TLayoutItemCollection = class (TCollection)
  private
    FLayoutManager : TLayoutManager;
    function GetItems (Index : Integer) : TLayoutItem;
    procedure SetItems (Index : Integer; item : TLayoutItem);
  protected
    function GetOwner : TPersistent; override;
  public
    function Add : TLayoutItem;
    function FindItem (const LayoutName : string) : TLayoutItem;
    property Items [Index : Integer] : TLayoutItem read GetItems Write SetItems; default;
    property LayoutManager : TLayoutManager read FLayoutManager;
  end;

  TLayoutItem = class (TLayoutCollectionItem)
  private
    FLayoutId : string;
    FLayoutName : string;
    FParameters : TLayoutParameterCollection;
    procedure SetParameters (Value : TLayoutParameterCollection);
    function GetCollection : TLayoutItemCollection;
    procedure SetLayoutName (const Value : string);
  protected
    function GetTwoColsComponent : TTwoColsComponent; override;
    function GetDisplayName: string; override;
    procedure FillArrays (wildcards, replacements : PStringArray); virtual;
  public
    constructor Create (ACollection : TCollection); override;
    destructor Destroy; override;
    procedure SaveLayout; virtual;
    procedure RestoreLayout; virtual;
    property Collection : TLayoutItemCollection read GetCollection;
  published
    property LayoutName : string read FLayoutName Write SetLayoutName;
    property LayoutId : string read FLayoutId Write FLayoutId;
    property Parameters : TLayoutParameterCollection read FParameters Write SetParameters;
  end;

  TLayoutManager = class(TTwoColsComponent)
  private
    FLayoutItems : TLayoutItemCollection;
    FBeforeSaveLayout,
    FAfterSaveLayout,
    FBeforeRestoreLayout,
    FAfterRestoreLayout : TNotifyEvent;
    procedure SetLayoutItems (Value : TLayoutItemCollection);
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    procedure SaveLayout (const LayoutName : string); virtual;
    procedure RestoreLayout (const LayoutName : string); virtual;
  published
    property LayoutItems : TLayoutItemCollection read FLayoutItems Write SetLayoutItems;
    property BeforeSaveLayout : TNotifyEvent read FBeforeSaveLayout Write FBeforeSaveLayout;
    property AfterSaveLayout : TNotifyEvent read FAfterSaveLayout Write FAfterSaveLayout;
    property BeforeRestoreLayout : TNotifyEvent read FBeforeRestoreLayout Write FBeforeRestoreLayout;
    property AfterRestoreLayout : TNotifyEvent read FAfterRestoreLayout Write FAfterRestoreLayout;
  end;

implementation

uses
  PropsSaver;

{ TLayoutCollectionItem }

function TLayoutCollectionItem.GetLayoutManager;
begin
  Result := TwoColsComponent as TLayoutManager;
end;

{ TLayoutParameterCollection }

function TLayoutParameterCollection.GetItems;
begin
  Result := inherited Items [Index] as TLayoutParameter;
end;

procedure TLayoutParameterCollection.SetItems;
begin
  inherited Items [Index] := item;
end;

function TLayoutParameterCollection.GetOwner;
begin
  Result := FLayoutItem;
end;

function TLayoutParameterCollection.Add;
begin
  Result := inherited Add as TLayoutParameter;
end;

{ TLayoutParameter }

function TLayoutParameter.GetLayoutParameterCollection;
begin
  Result := (inherited Collection as TLayoutParameterCollection);
end;

function TLayoutParameter.GetDisplayName;
begin
  Result := FParameter + '=' + FValue;
end;

function TLayoutParameter.GetTwoColsComponent;
begin
  Result := Collection.LayoutItem.Collection.LayoutManager;
end;

procedure TLayoutParameter.SetValue;
begin
  if AValue <> FValue
    then
    begin
      FValue := AValue;
      if (csDesigning in LayoutManager.ComponentState) and
         (Collection <> nil) and (LayoutManager.EditorForm <> nil)
        then LayoutManager.EditorForm.ItemModified (Self);
    end;
end;

procedure TLayoutParameter.SetParameter;
begin
  if AParameter <> FParameter
    then
    begin
      FParameter := AParameter;
      if (csDesigning in LayoutManager.ComponentState) and
         (Collection <> nil) and (LayoutManager.EditorForm <> nil)
        then LayoutManager.EditorForm.ItemModified (Self);
    end;
end;

{ TLayoutItemCollection }

function TLayoutItemCollection.GetItems;
begin
  Result := inherited Items [Index] as TLayoutItem;
end;

function TLayoutItemCollection.GetOwner;
begin
  Result := FLayoutManager;
end;

procedure TLayoutItemCollection.SetItems;
begin
   inherited Items [Index] := item;
end;

function TLayoutItemCollection.FindItem;
var
  i : Integer;
begin
  for i := 0 to Count - 1 do
    if Items [i].LayoutName = LayoutName
      then
      begin
        Result := Items [i];
        Exit;
      end;
  Result := nil;
end;

function TLayoutItemCollection.Add;
begin
  Result := inherited Add as TLayoutItem;
end;

{ TLayoutItem }

constructor TLayoutItem.Create;
begin
  inherited;
  FParameters := TLayoutParameterCollection.Create (TLayoutParameter);
  FParameters.FLayoutItem := Self;
end;

destructor TLayoutItem.Destroy;
begin
  FParameters.Free;
  inherited;
end;

function TLayoutItem.GetDisplayName;
begin
  Result := FLayoutName;
end;

procedure TLayoutItem.SetParameters;
begin
  FParameters.Assign (Value);
end;

function TLayoutItem.GetCollection;
begin
  Result := inherited Collection as TLayoutItemCollection;
end;

procedure TLayoutItem.FillArrays;
var
  i : Integer;
begin
  SetLength (wildcards^, Parameters.Count);
  SetLength (replacements^, parameters.Count);
  for i := 0 to parameters.Count - 1 do
    begin
      wildcards^ [i] := Parameters [i].Parameter;
      replacements^ [i] := Parameters [i].Value;
    end;
end;

procedure TLayoutitem.SaveLayout;
var
  wildcards, replacements : array of string;
begin
  FillArrays (@wildcards, @replacements);
  PropsSaver.SaveLayout (FLayoutId, wildcards, replacements);
end;

procedure TLayoutItem.RestoreLayout;
var
  wildcards, replacements : array of string;
begin
  FillArrays (@wildcards, @replacements);
  PropsSaver.RestoreLayout (FLayoutId, wildcards, replacements);
end;

function TLayoutItem.GetTwoColsComponent;
begin
  Result := Collection.LayoutManager;
end;

procedure TLayoutItem.SetLayoutName;
begin
  if Value <> FLayoutName
    then
    begin
      FLayoutName := Value;
      if (csDesigning in LayoutManager.ComponentState) and
         (Collection <> nil) and (LayoutManager.EditorForm <> nil)
        then LayoutManager.EditorForm.ItemModified (Self);
    end;
end;

{ TLayoutManager }

constructor TLayoutManager.Create;
begin
  inherited;
  FLayoutItems := TLayoutItemCollection.Create (TLayoutItem);
  FLayoutItems.FLayoutManager := Self;
end;

destructor TLayoutManager.Destroy;
begin
  FLayoutItems.Free;
  inherited;
end;

procedure TLayoutManager.SetLayoutItems;
begin
  FLayoutItems.Assign (Value);
end;

procedure TLayoutManager.SaveLayout;
var
  Item : TLayoutItem;
begin
  Item := LayoutItems.FindItem (LayoutName);
  if Item <> nil
    then
    begin
      if Assigned (FBeforeSaveLayout)
        then FBeforeSaveLayout (Self);
      Item.SaveLayout;
      if Assigned (FAfterSaveLayout)
        then FAfterSaveLayout (Self);
    end
    else raise Exception.Create ('Layout item not found');
end;

procedure TLayoutManager.RestoreLayout;
var
  Item : TLayoutItem;
begin
  Item := LayoutItems.FindItem (LayoutName);
  if Item <> nil
    then
    begin
      if Assigned (FBeforeRestoreLayout)
        then FBeforeRestoreLayout (Self);
      Item.RestoreLayout;
      if Assigned (FAfterRestoreLayout)
        then FAfterRestoreLayout (Self);
    end
    else raise Exception.Create ('Layout item not found');
end;

end.
