unit MtkPPPMgr;

//******************************************************************************
// This software is covered under the MorphTek.com Inc. Open Source License
// for a full text, see http://www.morphtek.com/oslicense.html
//
// MorphTek Peer to Peer Protocol Manager
// MorphTek Custom Peer to Peer Protocol
//
// Copyright (c) 2000 - MorphTek.com Inc.
//
// Description: The TMtkPPPMgr Component accepts and manages multiple protocols
//              of class TMtkCustomPPP.
//
//              TMtkPPPMgr reads and writes all data via blocking sockets, this
//              ensures that this component suite will work equally well in
//              the MS Windows environment as well as Linux.
//
//              Each message received by TMtkPPPMgr is firstly examined for a
//              valid message Header which is comprised of a ProtocolID,
//              a Message Type and the IP Address of the Requestor/Responder.
//
//              The Protocol ID determines which Protocol should handle the
//              attached data and the data is then passed to that protocol for
//              further handling.
//
//              The TMtkCustomPPP component should never be used directly,
//              rather a new component should be inherited from it and events
//              should be overridden and new events created to facilitate
//              specific protocol handling. See the sample TMtkFileExchange
//              component and sample application.
//
// Please download P2PSock.zip for the blocking socket components required by this code
//
// Author : Michael M. Michalak MACS
//
// License: for the full license text - please refer to :
// http://www.MorphTek.com/oslicense.html
//
//******************************************************************************

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, dsgnintf, ExtCtrls, Buttons, ComCtrls, inifiles,
  PsiThreadMgr, PsiThreadMgrDefault, PsiTCPServer, PsiBaseComponent,
  PsiComponent, PsiTCPConnection, PsiTCPClient,PsiWinSock, PsiStackWinsock;

const
  DefaultPort = 6463;   // This is the default listening port
  MaxTTL = 7;           // The Maximum Time to Live for a Message
  MaxMsgSize = 8192;    // The Maximum size of the Message List
  DefIPFile = '\MtkDefIP.ini';
  DefDirFile = 'C:\MtkDefDir.ini';

type
  // This is the Main Protocol's Message Type - it is used to indicate if the
  // current message is a Request or a Response. These two types simplify message
  // handling since Request messages are routed to all outgoing connections whilst
  // Response messages are routed back along the path the Request came in on.

  TMtkMsgType = (mtRequest, mtResponse);

  // This is the Main Protocol Header - All messages use this to identify the
  // protocol that is to manage the attached message.

  PMtkPPPHeader = ^TMtkPPPHeader;
  TMtkPPPHeader = packed record
                    ProtocolID : Word;    // I.E. 65536 possible protocols
                    MessageTyp : TMtkMsgType;
                    IPAddress  : Integer; // The IP Address of the Requestor/Responder
                  end;

  // This is the Custom PPP Protocol Message Header.
  // All messages have this secondary header that is handled by the Custom Protocol Object
  // It identifies a unique message, who created it, the function that is to be executed,
  // how long it has to live, how many times it was passed on and the length of the attached
  // message data.

  PMtkMsgHeader = ^TMtkMsgHeader;
  TMtkMsgHeader = packed record
                    MsgID        : TGUID;   // A Globally Unique Message Identifier
                    MsgCreator   : TGUID;   // A Globally Unique Identifier of the message Creator
                    FunctionCode : Byte;    // The Function of this message
                    TTLRemaining : Byte;    // Time to Live Indicator
                    JumpsTaken   : Byte;    // How many times this message was passed on
                    DataLength   : Integer; // The length of the attached data
                  end;

  // The combination of the Main Protocol Header and the Custom Protocol Message Header

  PThisMsg = ^TThisMsg;
  TThisMsg = record
               IDHdr  : TMtkPPPHeader;
               MsgHdr : TMtkMsgHeader;
             end;

  // This is the File Request Header

  PRequestFileHdr = ^TRequestFileHdr;
  TRequestFileHdr = packed record
                      HostIP : array[0..16] of Char;
                      FileName : array[0..127] of Char;
                    end;

  // These are message handling flags

  TMsgHandling = (mhDoNothing, mhProcessOnly, mhProcessAndPass);
  TDefDirs = (ddShared, ddDownload);

  // Events for Custom Peer to Peer Protocol

  TInterceptHostEvent = procedure(HostIP: String) of object;
  TAssignMessageEvent = procedure(MsgHdr: TMtkMsgHeader; DataStream: TMemoryStream) of object;

  TMtkPPPMgr = class;

  // This is the Parent Class from which all specific protocols are inherited
  // It provides a number of properties and methods which may be overridden by
  // inherited protocol components. (See Sample; TMtkFileExchange)

  TMtkCustomPPP = class(TComponent)
  private
    { Private declarations }
    FManager : TMtkPPPMgr;
    FProtocolID : Word;
    FCreator : String;
    FMsgList : TStrings;
    FMsgSrcList : TStrings;
    FDataStream : TMemoryStream;
    FThisMsg : TThisMsg;
    FdoProcessMsg : TMsgHandling;
    FCurrentPeer : String;
    FOnAssignMessage : TAssignMessageEvent;
    function GUIDToString(const ClassID: TGUID): String;
    function StringToGUID(const StrClassID: String): TGUID;
    function AddMsgToList(PeerIP: String) : TMsgHandling;
    procedure PassMessageBack(PeerIP: String); virtual;
  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AssignMessage(PeerIP: String; AStream: TMemoryStream); virtual;
    procedure PassMessageOn; virtual;
    procedure SendNewMessage(FunctionCode: Word; Data: TMemoryStream); virtual;
    procedure ReturnToSender(FunctionCode: Word; Data: TMemoryStream); virtual;
    property doProcessMsg : TMsgHandling read FdoProcessMsg write FdoProcessMsg;
    property ThisMsg : TThisMsg read FThisMsg write FThisMsg;
    property Manager : TMtkPPPMgr read FManager write FManager;
    property DataStream : TMemoryStream read FDataStream write FDataStream;
  published
    { Published declarations }
    property ProtocolID : Word read FProtocolID write FProtocolID;
    property OnAssignMessage : TAssignMessageEvent read FOnAssignMessage write FOnAssignMessage;
  end;

  // This is the Peer to Peer Protocol Manager. It listens for incoming messages,
  // sends those messages to the indicated protocol (if one exists) and dispatches
  // reply messages via the socket from which the original message was receoived.

  TMtkPPPMgr = class(TComponent)
  private
    { Private declarations }
    FServer : TPsiTCPServer;
    FClient : TPsiTCPClient;
    FThreadMgr : TPsiThreadMgrDefault;
    FClientList : TList;

    FCatcherList : TListBox;
    FDefaultHosts : TListBox;

    FDefSharedDir : TEdit;
    FDefDownloadDir : TEdit;

    FAutoListen : Boolean;
    FProtocols : TStrings;

    FLocalAddr : String;
    FCurrentProtocol : TMtkCustomPPP;
    FCurrentHdrID : Integer;

    FCurrentPeer : String;
    FCurrentStream : TMemoryStream;
    FReplyToSender : Boolean;

    FOnInterceptHost : TInterceptHostEvent;

    procedure doMsgProcess;

    function GetLocalAddress(LocalName : String) : String;
    // Server Events
    procedure doServerConnect(AThread: TPsiPeerThread);
    procedure doServerExecute(AThread: TPsiPeerThread);
    procedure doServerDisconnect(AThread: TPsiPeerThread);
    // Client Events
    procedure doClientConnected(Sender: TObject);
    procedure doClientDisconnected(Sender: TObject);
    procedure doClientStatus(axSender: TObject; const axStatus: TPsiStatus; const asStatusText: String);
    procedure doClientWork(Sender: TComponent; AWorkMode: TWorkMode; const AWorkCount: Integer);
    procedure doClientWorkBegin(Sender: TComponent; AWorkMode: TWorkMode; const AWorkCountMax: Integer);
    procedure doClientWorkEnd(Sender: TComponent; AWorkMode: TWorkMode);
    // Events
  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;

    procedure doConnectAll;
    procedure LoadDefaultAddressList;
    procedure SaveDefaultAddressList;
    procedure LoadDefaultDir(DirType:TDefDirs);
    procedure SaveDefaultDir(DirType:TDefDirs);
    procedure doSendMsg(AStream: TMemoryStream);
    procedure doReturnMsg(PeerIP: String; AStream: TMemoryStream);
    // Procedure Called By Thread;
    procedure FindProtocol;
    procedure AssignCurrentHdr(ThisHdr : Integer);
    procedure AssignCurrentProctocol(ThisProtocol: TMtkCustomPPP);

    property LocalAddr : String read FLocalAddr write FLocalAddr;
  published
    { Published declarations }
    property AutoListen : Boolean read FAutoListen write FAutoListen;
    property CatcherList : TListBox read FCatcherList write FCatcherList;
    property DefaultHosts : TListBox read FDefaultHosts write FDefaultHosts;
    property DefSharedDir : TEdit read FDefSharedDir write FDefSharedDir;
    property DefDownloadDir : TEdit read FDefDownloadDir write FDefDownloadDir;
    property Protocols : TStrings read FProtocols write FProtocols;
    property OnInterceptHost : TInterceptHostEvent read FOnInterceptHost write FOnInterceptHost;
  end;

  TMtkPPPListEditor = class(TPropertyEditor)
    procedure Edit; override;
    function  GetAttributes : TPropertyAttributes; override;
    function  GetValue : String; override;
  end;

  TfrmProtocol = class(TForm)
    ToolPanel      : TPanel;
    BtnRefresh     : TSpeedButton;
    BtnNew         : TSpeedButton;
    BtnDelete      : TSpeedButton;
    Panel1         : TPanel;
    cbProtocols    : TComboBox;
    Label1         : TLabel;
    lbProtocolList : TListBox;
    procedure BtnRefreshClick(Sender: TObject);
    procedure BtnNewClick(Sender: TObject);
    procedure BtnDeleteClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    FOwner : TComponent;
    FComp  : TComponent;
  end;

const
  Base64Table = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';

function Base64Encode(Value : TMemoryStream) : TMemoryStream; // Converts Passed Value To MIME Base64
function Base64Decode(Value : TMemoryStream) : TMemoryStream; // Converts Passed Value From MIME Base64

function CoCreateGuid(out guid: TGUID): HResult; stdcall; external 'ole32.dll';
function StringFromCLSID(const clsid: TGUID; out psz: PWideChar): HResult; stdcall; external 'ole32.dll';
function CLSIDFromString(psz: PWideChar; out clsid: TGUID): HResult; stdcall; external 'ole32.dll';
procedure CoTaskMemFree(pv: Pointer); stdcall; external 'ole32.dll';

procedure Register;

var
  DummyProtocol : TMtkCustomPPP;

implementation

{$R *.DFM}
{$R *.DCR}

// Global Functions ************************************************************

function Base64Encode(Value : TMemoryStream) : TMemoryStream;
var
  AIn   : Array[1..3] Of Byte;
  AOut  : Array[1..4] Of Byte;
  AWork : Array[1..3] Of Byte;
  I : Integer;
  O : LongInt;
  Buf : Char;
begin
  Buf := #0;
  Result := TMemoryStream.Create;
  //Result.Write(Buf,1);
  I := 1;
  O := Value.Seek(0,soFromEnd);
  case O mod 3 Of
  1 : begin
        Buf := #0;
        Value.Seek(0,soFromEnd);
        Value.Write(Buf,1);
        Value.Write(Buf,1);
      end;
  2 : begin
        Buf := #0;
        Value.Seek(0,soFromEnd);
        Value.Write(Buf,1);
      end;
   end;
   O := Value.Seek(0,soFromEnd);
   Value.Seek(0,soFromBeginning);
   while I < O do
   begin
     Value.Read(Buf,1);
     AIn[1] := Byte(Buf);
     Value.Read(Buf,1);
     AIn[2] := Byte(Buf);
     Value.Read(Buf,1);
     AIn[3] := Byte(Buf);

     AOut[1] := Byte(AIn[1] shr 2);
     AWork[1] := Byte(AIn[1] shl 4);
     AWork[2] := Byte(AWork[1] and $30);
     AWork[3] := Byte(AIn[2] shr 4);
     AOut[2] := Byte(AWork[2] or AWork[3]);
     AWork[1] := Byte(AIn[2] shl 2);
     AWork[2] := Byte(AWork[1] and $3C);
     AWork[3] := Byte(AIn[3] shr 6);
     AOut[3] := Byte(AWork[2] or AWork[3]);
     AOut[4] := Byte(AIn[3] and $3F);
     Buf := Base64Table[AOut[1]+1];
     Result.Write(Buf,1);
     Buf := Base64Table[AOut[2]+1];
     Result.Write(Buf,1);
     Buf := Base64Table[AOut[3]+1];
     Result.Write(Buf,1);
     Buf := Base64Table[AOut[4]+1];
     Result.Write(Buf,1);
     inc(I,3);
   end;
   if O mod 3 > 0 then
   begin
     Result.Seek(0,soFromEnd);
     Buf := '=';
     Result.Write(Buf,1);
   end;
   if O mod 3 = 1 then
   begin
     Result.Seek(1,soFromEnd);
     Buf := '=';
     Result.Write(Buf,1);
   end;
end;

function Base64Decode(Value : TMemoryStream) : TMemoryStream;
var
  AIn   : Array[1..4] Of Byte;
  AOut  : Array[1..3] Of Byte;
  AWork : Array[1..3] Of Byte;
  I : Integer;
  C : Integer;
  O : LongInt;
  Buf0 : Char;
  Buf1 : Char;
  Buf2 : Char;
  Buf3 : Char;
begin
  Buf0 := #0;
  Buf1 := #0;
  Buf2 := #0;
  Buf3 := #0;
  Result := TMemoryStream.Create;
  //Result := '';
  I := 1;
  O := Value.Seek(0,soFromEnd);
  Value.Seek(0,soFromBeginning);
  while I < O do
  begin
    C := 3;
    FillChar(AWork, SizeOf(AWork), #0);
    FillChar(AOut, SizeOf(AWork), #0);
    Value.Read(Buf0,1);
    AIn[1] := Byte(Pos(Buf0,Base64Table)-1);
    Value.Read(Buf1,1);
    AIn[2] := Byte(Pos(Buf1,Base64Table)-1);
    Value.Read(Buf2,1);
    AIn[3] := Byte(Pos(Buf2,Base64Table)-1);
    Value.Read(Buf3,1);
    AIn[4] := Byte(Pos(Buf3,Base64Table)-1);
    if Buf3 = '=' then
    begin
      C := 2;
      AIn[4] := 0;
      if Buf2 = '=' then
      begin
        C := 1;
        AIn[3] := 0;
      end;
    end;
    AWork[2] := Byte(AIn[1] Shl 2);
    AWork[3] := Byte(AIn[2] Shr 4);
    AOut[1] := Byte(AWork[2] Or AWork[3]);
    AWork[2] := Byte(AIn[2] Shl 4);
    AWork[3] := Byte(AIn[3] Shr 2);
    AOut[2] := Byte(AWork[2] Or AWork[3]);
    AWork[2] := Byte(AIn[3] Shl 6);
    AOut[3] := Byte(AWork[2] Or AIn[4]);
    Result.Write(Char(AOut[1]),1);
    if C > 1 then
      Result.Write(Char(AOut[2]),1);
    if C > 2 then
      Result.Write(Char(AOut[3]),1);
    inc(I, 4);
  end;
end;


// TMtkCustomPPP ***************************************************************

constructor TMtkCustomPPP.Create(AOwner: TComponent);
var
  NewGUID : TGUID;
begin
  inherited Create(AOwner);
  FManager := nil;
  FMsgList := TStringList.Create;
  FMsgSrcList := TStringList.Create;
  FDataStream := TMemoryStream.Create;
  CoCreateGUID(NewGUID); // Create a Unique ID for this Instance
  FCreator := GUIDToString(NewGUID);
end;

destructor TMtkCustomPPP.Destroy;
begin
  FDataStream.Free;
  FMsgList.Free;
  FMsgSrcList.Free;
  inherited Destroy;
end;

procedure TMtkCustomPPP.Notification(AComponent: TComponent; Operation: TOperation);
begin
  // Override this in the inherited component;
end;

function TMtkCustomPPP.GUIDToString(const ClassID: TGUID): String;
var
  P: PWideChar;
begin
  StringFromCLSID(ClassID, P);
  Result := P;
  CoTaskMemFree(P);
end;

function TMtkCustomPPP.StringToGUID(const StrClassID: String): TGUID;
begin
  CLSIDFromString(PWideChar(WideString(StrClassID)), Result);
end;

function TMtkCustomPPP.AddMsgToList(PeerIP: String) : TMsgHandling;
var
  ThisPeer, ThisMsgID, ThisCreator : String;
  MsgType : TMtkMsgType;
begin
  MsgType := FThisMsg.IDHdr.MessageTyp;
  ThisMsgID := GUIDToString(FThisMsg.MsgHdr.MsgID);// + '::' + IntToStr(FThisMsg.MsgHdr.FunctionCode);
  ThisCreator := GUIDToString(FThisMsg.MsgHdr.MsgCreator);
  // A Received Message can be a Request or a Response;
  // In the case of a Request,
  //   We must pass it on to all our outgoing connections
  //   AND we must process it if we have not done so already (if we have, it's in the list)
  // In the case of a Response,
  //   If we originated the Request (we are the creator) - just process it
  //   if we did not request it, then we should pass it back to where the request came from.
  Result := mhDoNothing;
  if MsgType = mtRequest then
  begin
    // This is a Request
    // If we have seen it before, do nothing
    if FMsgList.IndexOf(ThisMsgID) = -1 then
    begin
      // Delete the Oldest message
      if FMsgList.Count > MaxMsgSize then
      begin
        FMsgList.Delete(0);
        FMsgSrcList.Delete(0);
      end;
      FMsgList.Add(ThisMsgID);
      FMsgSrcList.Add(PeerIP);
      if ThisCreator = FCreator then
        Result := mhProcessOnly
      else
        Result := mhProcessAndPass;
    end;
  end
  else
  begin
    // This is a Response
    // If we originated the message, we only need to process it
    if ThisCreator = FCreator then
      Result := mhProcessOnly
    else
    begin
      // We did not ask for this, so return the Message to the Sender;
      // Note: A Response sent to us must have been routed by us (or asked for by us)
      // if we did not route it, we should do nothing;
      if FMsgList.IndexOf(ThisMsgID) <> -1 then
      begin
        ThisPeer := FMsgSrcList[FMsgList.IndexOf(ThisMsgID)];
        PassMessageBack(ThisPeer);
      end;
      {
      else
      begin
        // if we have not seen the message before, we need to add it to our list
        ThisPeer := PeerIP;
        if FMsgList.Count > MaxMsgSize then
        begin
          FMsgList.Delete(0);
          FMsgSrcList.Delete(0);
        end;
        FMsgList.Add(ThisMsgID);
        FMsgSrcList.Add(PeerIP);
      end;
      // Pass it back and let the sender handle it
      PassMessageBack(ThisPeer);
      }
    end;
  end;
end;

procedure TMtkCustomPPP.AssignMessage(PeerIP: String; AStream: TMemoryStream);
var
  MsgHandling : TMsgHandling;
begin
  FCurrentPeer := PeerIP;
  // The FDataStream needs to be Cleared;
  FDataStream.Clear;
  AStream.Seek(0,soFromBeginning);
  AStream.Read(FThisMsg, SizeOf(TThisMsg));
  try
    FDataStream.Seek(0,soFromBeginning);
    FDataStream.CopyFrom(AStream, FThisMsg.MsgHdr.DataLength);
  except
  end;
  // How should we handle this Message?
  FdoProcessMsg := AddMsgToList(PeerIP);
  try
    if Assigned(FOnAssignMessage) then
      FOnAssignMessage(FThisMsg.MsgHdr, FDataStream);
  except
  end;
end;

procedure TMtkCustomPPP.PassMessageOn;
var
  Response : TMemoryStream;
  SSize : Integer;
begin
  // Passing the Message On is simple;
  // Just decrement the TTL in the Header, append the Data Stream, and send it
  if FThisMsg.MsgHdr.TTLRemaining > 0 then
  begin
    try
      Response := TMemoryStream.Create;
      // if the TTL is greater than the Max, set it to the max before decrementing
      if FThisMsg.MsgHdr.TTLRemaining > MaxTTL then
        FThisMsg.MsgHdr.TTLRemaining := MaxTTL;
      dec(FThisMsg.MsgHdr.TTLRemaining);
      Response.Write(FThisMsg,SizeOf(TThisMsg));
      SSize := FDataStream.Seek(0,soFromEnd);
      FDataStream.Seek(0,soFromBeginning);
      Response.CopyFrom(FDataStream, SSize);
      Response.Seek(0,soFromBeginning);
      if FManager <> nil then
        FManager.doSendMsg(Response);
    finally
      Response.Free;
    end;
  end;
end;

procedure TMtkCustomPPP.PassMessageBack;
var
  Response : TMemoryStream;
  SSize : Integer;
begin
  // Passing the Message back along the route is even simpler;
  // No need to decrement the TTL since it is being passed back to the sender and
  // will die on receipt.
  // A sends request to B,
  // B has no data and passes it to C
  // C has data and passes it back to B
  // B finds the originator A and only sends it to A
  // A receives the reply and does not pass it on.
  // A received just one response from the original request.
  try
    Response := TMemoryStream.Create;
    Response.Write(FThisMsg,SizeOf(TThisMsg));
    SSize := FDataStream.Seek(0,soFromEnd);
    FDataStream.Seek(0,soFromBeginning);
    Response.CopyFrom(FDataStream, SSize);
    Response.Seek(0,soFromBeginning);
    if FManager <> nil then
      FManager.doReturnMsg(FCurrentPeer, Response);
  finally
    Response.Free;
  end;
end;

procedure TMtkCustomPPP.ReturnToSender(FunctionCode: Word; Data: TMemoryStream);
var
  NewMsg : TMemoryStream;
  SSize : Integer;
begin
  // Return our response to the request. This is different from PassMessageBack
  // as it contains data that is in response to a request.
  // For example
  // A sends request to B,
  // B has data and sends it to A and also sends the original request to C
  // C has data and passes it back to B
  // B finds the originator of the request and only sends it to A
  // A receives the reply and does not pass it on.
  // A has now received two responses from the original request
  if FManager <> nil then
  begin
    SSize := Data.Seek(0,soFromEnd);
    try
      NewMsg := TMemoryStream.Create;
      FThisMsg.IDHdr.MessageTyp := mtResponse;
      FThisMsg.IDHdr.IPAddress := inet_addr(PChar(FManager.FLocalAddr));
      FThisMsg.MsgHdr.FunctionCode := FunctionCode;
      FThisMsg.MsgHdr.TTLRemaining := MaxTTL;
      FThisMsg.MsgHdr.DataLength := SSize;
      NewMsg.Write(FThisMsg, SizeOf(TThisMsg));
      if SSize > 0 then
      begin
        Data.Seek(0,soFromBeginning);
        NewMsg.CopyFrom(Data,SSize);
      end;
      NewMsg.Seek(0,soFromBeginning);
      if FManager <> nil then
        FManager.doReturnMsg(FCurrentPeer, NewMsg);
    finally
      NewMsg.Free;
    end;
  end;
end;

procedure TMtkCustomPPP.SendNewMessage(FunctionCode: Word; Data: TMemoryStream);
var
  NewMsg : TMemoryStream;
  SSize : Integer;
begin
  // Send a New Message - This handles the Generic Headers and appends any data
  // New Messages are ALWAYS Requests.
  if FManager <> nil then
  begin
    SSize := Data.Seek(0,soFromEnd);
    try
      NewMsg := TMemoryStream.Create;
      FThisMsg.IDHdr.ProtocolID := ProtocolID;
      FThisMsg.IDHdr.MessageTyp := mtRequest;
      FThisMsg.IDHdr.IPAddress := inet_addr(PChar(FManager.FLocalAddr));
      CoCreateGuid(FThisMsg.MsgHdr.MsgID);
      FThisMsg.MsgHdr.MsgCreator := StringToGUID(FCreator);
      FThisMsg.MsgHdr.FunctionCode := FunctionCode;
      FThisMsg.MsgHdr.TTLRemaining := MaxTTL;
      FThisMsg.MsgHdr.JumpsTaken := 0;
      FThisMsg.MsgHdr.DataLength := SSize;
      NewMsg.Write(FThisMsg, SizeOf(TThisMsg));
      if SSize > 0 then
      begin
        Data.Seek(0,soFromBeginning);
        NewMsg.CopyFrom(Data,SSize);
      end;
      NewMsg.Seek(0,soFromBeginning);
      FManager.doSendMsg(NewMsg);
    finally
      NewMsg.Free;
    end;
  end;
end;

// TMtkPPPMgr ******************************************************************

constructor TMtkPPPMgr.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FServer := nil;
  if not (csDesigning in ComponentState) then
  begin
    // Default Thread Manager;
    FThreadMgr := TPsiThreadMgrDefault(nil);
    // This is the Listening Server
    FServer := TPsiTCPServer.Create(nil);
    FServer.DefaultPort := DefaultPort;
    FServer.ThreadMgr := FThreadMgr;
    FServer.OnConnect := doServerConnect;
    FServer.OnExecute := doServerExecute;
    FServer.OnDisconnect := doServerDisconnect;
  end;
  // This is a List of Client Connections
  FClientList := TList.Create;
  // This is a List of Connected Protocols
  FProtocols := TStringList.Create;
  FCatcherList := nil;
  FDefaultHosts := nil;
  //FDefaultHosts..OnDestroy := doSaveContent
end;

{
procedure TMtkPPPMgr.doSaveContent;
begin
  // Save Default IP Addresses;
  if FDefaultHosts <> nil then
  begin
    ThisDir := GetCurrentDir;
    ThisDir := ThisDir + DefIPFile;
    if FDefaultHosts.Items.Count > 0 then
      FDefaultHosts.Items.SaveToFile(ThisDir);
  end;
end;
}
destructor TMtkPPPMgr.Destroy;
var
  i : Integer;
  Obj : TObject;
  ThisDir : String;
begin
  FProtocols.Free;
  for i := FClientList.Count -1 downto 0 do
  begin
    TPsiTCPClient(FClientList[i]).Disconnect;
    Obj := FClientList[i];
    FreeAndNil(Obj);
  end;
  inherited Destroy;
end;

procedure TMtkPPPMgr.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if Operation = opRemove then
  begin
     if AComponent = FDefSharedDir then
       FDefSharedDir := nil;
     if AComponent = FDefDownloadDir then
       FDefDownloadDir := nil;
     if AComponent = FCatcherList then
       FCatcherList := nil;
     if AComponent = FDefaultHosts then
       FDefaultHosts := nil;
  end;
end;

procedure TMtkPPPMgr.Loaded;
var
  i : Integer;
  ThisProtocol : TMtkCustomPPP;
begin
  inherited Loaded;
  if FAutoListen then
    if FServer <> nil then
    begin
      FServer.Active := True;
      FLocalAddr := GetLocalAddress(FServer.LocalName);
    end;
  // Assign the Protocol Objects
  for i := 0 to FProtocols.Count -1 do
  begin
    ThisProtocol := TMtkCustomPPP(Owner.FindComponent(FProtocols[i]));
    if ThisProtocol <> nil then
    begin
      ThisProtocol.FManager := Self;
      if ThisProtocol.ProtocolID = FCurrentHdrID then
      begin
        FCurrentProtocol := ThisProtocol;
        break;
      end;
    end;
  end;
end;

function TMtkPPPMgr.GetLocalAddress(LocalName : String) : String;
var
  pa: PChar;
  sa: TInAddr;
  Host: PHostEnt;
begin
  Host := GetHostByName(PChar(LocalName));
  if Host <> nil then
  begin
    pa := Host^.h_addr_list^;
    sa.S_un_b.s_b1 := pa[0];
    sa.S_un_b.s_b2 := pa[1];
    sa.S_un_b.s_b3 := pa[2];
    sa.S_un_b.s_b4 := pa[3];
    with sa.S_un_b do
    begin
      result := IntToStr(Ord(s_b1)) + '.' +
                IntToStr(Ord(s_b2)) + '.' +
                IntToStr(Ord(s_b3)) + '.' +
                IntToStr(Ord(s_b4));
    end;
  end;
end;

procedure TMtkPPPMgr.AssignCurrentHdr(ThisHdr : Integer);
begin
  FCurrentHdrID := ThisHdr;
end;

procedure TMtkPPPMgr.AssignCurrentProctocol(ThisProtocol: TMtkCustomPPP);
begin
  FCurrentProtocol := ThisProtocol;
end;

procedure TMtkPPPMgr.FindProtocol;
var
  i : Integer;
  ThisProtocol : TMtkCustomPPP;
begin
  // This searches through the list of attached protocols and finds one that
  // matches the passed identifier - if not found it returns nil.
  FCurrentProtocol := nil;
  for i := 0 to FProtocols.Count -1 do
  begin
    ThisProtocol := TMtkCustomPPP(Owner.FindComponent(FProtocols[i]));
    if ThisProtocol <> nil then
    begin
      ThisProtocol.FManager := Self;
      if ThisProtocol.ProtocolID = FCurrentHdrID then
      begin
        FCurrentProtocol := ThisProtocol;
        break;
      end;
    end;
  end;
end;

procedure TMtkPPPMgr.LoadDefaultDir(DirType:TDefDirs);
var
  RegIni : TIniFile;
begin
  try
    RegIni := TIniFile.Create(DefDirFile);
    case DirType of
    ddShared   : begin
                   if FDefSharedDir <> nil then
                     FDefSharedDir.Text := RegIni.ReadString('DefDir', 'Shared', 'ERROR');
                     if FDefSharedDir.Text = 'ERROR' then
                       FDefSharedDir.Text := GetCurrentDir;
                 end;
    ddDownload : begin
                   if FDefDownloadDir <> nil then
                     FDefDownloadDir.Text := RegIni.ReadString('DefDir', 'Download', 'ERROR');
                     if FDefDownloadDir.Text = 'ERROR' then
                       FDefDownloadDir.Text := GetCurrentDir;
                 end;
    end;
  finally
    RegIni.Free;
  end;
end;

procedure TMtkPPPMgr.SaveDefaultDir(DirType:TDefDirs);
var
  RegIni : TIniFile;
begin
  try
    RegIni := TIniFile.Create(DefDirFile);
    case DirType of
    ddShared   : begin
                   if FDefSharedDir <> nil then
                     RegIni.WriteString('DefDir', 'Shared', FDefSharedDir.Text)
                 end;
    ddDownload : begin
                   if FDefDownloadDir <> nil then
                     RegIni.WriteString('DefDir', 'Download', FDefDownloadDir.Text)
                 end;
    end;
  finally
    RegIni.Free;
  end;
end;

// Server Events

procedure TMtkPPPMgr.doServerConnect(AThread: TPsiPeerThread);
var
  PeerIP : String;
begin
  // The Server gets a connection and spawns a thread.
  // This thread provides access to a blocking socket.
  if not (csDestroying in ComponentState) then
    if FCatcherList <> nil then
    begin
      PeerIP := AThread.Connection.Binding.PeerIP;
      if FCatcherList.Items.IndexOf(PeerIP) = -1 then
        FCatcherList.Items.Add(PeerIP);
      if Assigned(FOnInterceptHost) then
        FOnInterceptHost(PeerIP);
    end;
end;

procedure TMtkPPPMgr.doMsgProcess;
begin
  FCurrentProtocol.AssignMessage(FCurrentPeer, FCurrentStream);
end;

procedure TMtkPPPMgr.doServerExecute(AThread: TPsiPeerThread);
var
  AStream : TMemoryStream;
  ASize : Integer;
  Hdr : TMtkPPPHeader;
begin
  AStream := nil;
  // The thread created now executes
  try
    AStream := TMemoryStream.Create;
    AThread.Connection.ReadStream(AStream,-1,False);
    ASize := AStream.Seek(0,soFromEnd);
    if ASize > 0 then
    begin
      AStream.Seek(0,soFromBeginning);
      // Now Determine what this message is;
      AStream.Read(Hdr,SizeOf(TMtkPPPHeader));
      // OK, we need to check which protocol is being used
      FCurrentHdrID := Hdr.ProtocolID;
      AThread.Synchronize(FindProtocol);
      if FCurrentProtocol <> nil then
      begin
        FCurrentPeer   := AThread.Connection.Binding.PeerIP;
        //FCurrentThread := AThread;
        FCurrentStream := AStream;
        FReplyToSender := False;
        AThread.Synchronize(doMsgProcess);
      end;
    end;
    AThread.Connection.Disconnect;
  finally
    AStream.Free
  end;
end;

procedure TMtkPPPMgr.doServerDisconnect(AThread: TPsiPeerThread);
begin
  {
  if not (csDestroying in ComponentState) then
    if FCatcherList <> nil then
      if FCatcherList.Items.IndexOf('->'+AThread.Connection.Binding.PeerIP) <> -1 then
        FCatcherList.Items.Delete(FCatcherList.Items.IndexOf('->'+AThread.Connection.Binding.PeerIP));
  }
end;

// Client Events

procedure TMtkPPPMgr.doClientConnected(Sender: TObject);
begin
  //
end;

procedure TMtkPPPMgr.doClientDisconnected(Sender: TObject);
begin
  //
end;

procedure TMtkPPPMgr.doClientStatus(axSender: TObject;
  const axStatus: TPsiStatus; const asStatusText: String);
begin
  //
end;

procedure TMtkPPPMgr.doClientWork(Sender: TComponent;
  AWorkMode: TWorkMode; const AWorkCount: Integer);
begin
  //
end;

procedure TMtkPPPMgr.doClientWorkBegin(Sender: TComponent;
  AWorkMode: TWorkMode; const AWorkCountMax: Integer);
begin
  //
end;

procedure TMtkPPPMgr.doClientWorkEnd(Sender: TComponent;
  AWorkMode: TWorkMode);
begin
  //
end;

procedure TMtkPPPMgr.LoadDefaultAddressList;
var
  ThisDir : String;
begin
  // This Loads the Default AddressList Values;
  if FDefaultHosts <> nil then
  begin
    ThisDir := GetCurrentDir;
    ThisDir := ThisDir + DefIPFile;
    if FileExists(ThisDir) then
    begin
      FDefaultHosts.Items.LoadFromFile(ThisDir);
      doConnectAll;
    end;
  end;
end;

procedure TMtkPPPMgr.SaveDefaultAddressList;
var
  ThisDir : String;
  i : Integer;
begin
  // This Loads the Default AddressList Values;
  if FDefaultHosts <> nil then
  begin
    ThisDir := GetCurrentDir;
    ThisDir := ThisDir + DefIPFile;
    FDefaultHosts.Items.SaveToFile(ThisDir);
    for i := FClientList.Count -1 downto 0 do
    begin
      TPsiTCPClient(FClientList.Items[i]).Free;
      FClientList.Delete(i);
    end;
    FClientList.Clear;
    if not (csFreeNotification in FServer.ComponentState) then
      doConnectAll;
  end;
end;

procedure TMtkPPPMgr.doConnectAll;
var
  i : Integer;
begin
  if (FDefaultHosts <> nil) and (FServer <> nil) then
  begin
    for i := 0 to FDefaultHosts.Items.Count -1 do
    begin
      if FDefaultHosts.Items[i] <> FLocalAddr then
      begin
        FClient := TPsiTCPClient.Create(nil);
        FClient.OnConnected := doClientConnected;
        FClient.OnDisconnected := doClientDisconnected;
        FClient.OnStatus := doClientStatus;
        FClient.OnWork := doClientWork;
        FClient.OnWorkBegin := doClientWorkBegin;
        FClient.OnWorkEnd := doClientWorkEnd;
        FClient.Host := FDefaultHosts.Items[i];
        FClient.Port := DefaultPort;
        FClientList.Add(FClient);
      end;
    end;
  end;
end;

procedure TMtkPPPMgr.doReturnMsg(PeerIP: String; AStream: TMemoryStream);
var
  ThisSender : TPsiTCPClient;
  ASize : Integer;
begin
  // Return to Sender
  ASize := AStream.Seek(0,soFromEnd);
  AStream.Seek(0,soFromBeginning);
  FReplyToSender := True;
  try
    ThisSender := TPsiTCPClient.Create(nil);
    try
      ThisSender.Host := PeerIP;
      ThisSender.Port := DefaultPort;
      ThisSender.Connect;
      if ThisSender.Connected then
      begin
        ThisSender.WriteStream(AStream,True,True);
        ThisSender.Disconnect;
      end;
    except
      ShowMessage('Error in Return to Sender : '+PeerIP);
    end;
  finally
    ThisSender.Free;
  end;
end;

procedure TMtkPPPMgr.doSendMsg(AStream: TMemoryStream);
var
  i : Integer;
begin
  for i := 0 to FClientList.Count -1 do
  begin
    AStream.Seek(0,soFromBeginning);
    // if we can not connect then go to the next client;
    try
      TPsiTCPClient(FClientList[i]).Connect;
      if TPsiTCPClient(FClientList[i]).Connected then
      begin
        TPsiTCPClient(FClientList[i]).WriteStream(AStream,True,True);
        TPsiTCPClient(FClientList[i]).Disconnect;
      end;
    except
    end;
  end;
end;

// TMtkPPPListEditor ***********************************************************

procedure TMtkPPPListEditor.Edit;
var
  Comp  : TComponent;
  ThisProtocol : TMtkCustomPPP;
  i : Integer;
begin
  with TfrmProtocol.Create(Application) do
  begin
    FOwner := nil;
    try
      Comp := TComponent(GetComponent(0));
      FComp := Comp;
      FOwner := Comp.Owner;
      for i := 0 to TMtkPPPMgr(Comp).Protocols.Count -1 do
      begin
        lbProtocolList.Items.Add(TMtkPPPMgr(Comp).Protocols[i]);
      end;
      // Add all available protocol components to the list
      for i := 0 to Comp.Owner.ComponentCount -1 do
      begin
        if Comp.Owner.Components[i] is TMtkCustomPPP then
          cbProtocols.Items.Add(Comp.Owner.Components[i].Name);
      end;
      ShowModal;
      TMtkPPPMgr(Comp).Protocols.Clear;
      for i := 0 to lbProtocolList.Items.Count -1 do
      begin
        ThisProtocol := TMtkCustomPPP(FOwner.FindComponent(lbProtocolList.Items[i]));
        if ThisProtocol <> nil then
        begin
          TMtkPPPMgr(Comp).Protocols.AddObject(ThisProtocol.Name,ThisProtocol);
        end;
      end;
    finally
      Free;
    end;
  end;
end;

function TMtkPPPListEditor.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

function TMtkPPPListEditor.GetValue: String;
begin
  Result := '(TStrings)';
end;

// frmProtocolEditor ***********************************************************

procedure TfrmProtocol.BtnRefreshClick(Sender: TObject);
var
  i : Integer;
  ThisProtocol : TMtkCustomPPP;
begin
  Screen.Cursor:=crHourGlass;
  try
    if FOwner <> nil then
    begin
      lbProtocolList.Clear;
      for i := 0 to TMtkPPPMgr(FComp).Protocols.Count -1 do
      begin
        ThisProtocol := TMtkCustomPPP(FOwner.FindComponent(TMtkPPPMgr(FComp).Protocols[i]));
        if ThisProtocol <> nil then
        begin
          lbProtocolList.Items.AddObject(ThisProtocol.Name,ThisProtocol);
        end;
      end;
    end;
  finally
    Screen.Cursor:=crDefault;
  end;
end;

procedure TfrmProtocol.BtnNewClick(Sender: TObject);
var
  ThisProtocol : TComponent;
begin
  // Add a Protocol
  if cbProtocols.Text <> '' then
  begin
    if FOwner = nil then
      lbProtocolList.Items.Add('Owner is nil');
    if FOwner <> nil then
    begin
      ThisProtocol := FOwner.FindComponent(cbProtocols.Text);
      if ThisProtocol <> nil then
      begin
        if lbProtocolList.Items.IndexOf(cbProtocols.Text) = -1 then
          lbProtocolList.Items.AddObject(cbProtocols.Text, ThisProtocol);
      end;
    end;
  end;
end;

procedure TfrmProtocol.BtnDeleteClick(Sender: TObject);
begin
  // Delete the selected Protocol;
  if lbProtocolList.ItemIndex <> -1 then
    lbProtocolList.Items.Delete(lbProtocolList.ItemIndex);
end;

// Register ********************************************************************

procedure Register;
begin
  RegisterComponents('MorphTek P2P', [TMtkPPPMgr]);
  RegisterPropertyEditor(TypeInfo(TStrings), TMtkPPPMgr, 'Protocols', TMtkPPPListEditor);
end;

end.
