{

 Project UML Component Factory
 Unit Dynamic Library Element

 The contents of this file are subject to the Mozilla Public License Version
 1.1 (the "License"); you may not use this file except in compliance with the
 License. You may obtain a copy of the License at http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 the specific language governing rights and limitations under the License.

 The Original Code is DynamicLibraryElement.pas

 The Initial Developer of the Original Code is Thierry.Coq
 Portions created by this individual are Copyright (C)
 2000 of this individual.

 Last modified: 25 oct 2000
 Version 1.0

 known bugs :
 1. for some reason, finalizing dynamically loaded packages very often results
    in a AV. So I don't call finalize, which is a mistake. Therefore the
    finalizations of the packages will not work. If anyone has an idea, please
    contact me ? (thierry.coq@centraliens.net)
}
unit DynamicLibraryElement;

interface

uses Windows, Classes;

type
  TDynamicLibraryElementClass = class of TDynamicLibraryElement;
  TDynamicLibraryElement = class( TComponent)
  public
    constructor Create( anOwner : TComponent); override;
    destructor Destroy; override;
    procedure Execute; virtual;
    class function GetDescription: String; virtual;
    class procedure SetDescription(const Value: String); virtual;
    class function GetHelp : String; virtual;
    class procedure SetHelp(const Value: String); virtual;
    class function GetCommandText: String; virtual;
    class procedure SetCommandText(const Value: String); virtual;
    class function GetInstances(index: integer): TDynamicLibraryElement; virtual;
    class procedure SetInstances(index: integer;
      const Value: TDynamicLibraryElement); virtual;
    class procedure AddInstance( Value : TDynamicLibraryElement); virtual;
    class procedure RemoveInstance( Value : TDynamicLibraryElement); virtual;

    { FreeAllInstances needs the class itself as a parameter in order to free
      and remove only those instances that are really the class's and not any
      other.}
    class procedure FreeAllInstances( aClass : TDynamicLibraryElementClass); virtual;
    class function  InstanceCount : integer; virtual;
    property Instances[ index : integer] : TDynamicLibraryElement
      read GetInstances write SetInstances;
  published
    property Description : String read GetDescription write SetDescription;
    property Help : String read GetHelp write SetHelp;
    property CommandText : String read GetCommandText write SetCommandText;
  end;


  { The DLE Manager is designed to manage the addition of new elements, loaded
    dynamically from files. Of course, in order to be used, the file is a Delphi
    Execution package, which declares descendents of TDynamicLibraryElement class.
    }
  TDLEManager = class
  private
    FDLEList : TList;
    FPackageNamesList : TStringList;
    FPackageList : TList;
    FRegistryKey: String;
    FIsFinalizingPackageRequired: Boolean;
    procedure FreePackageList;
    function GetElements(index: integer): TDynamicLibraryElementClass;
    procedure SetElements(index: integer; const Value: TDynamicLibraryElementClass);
    function GetCount: integer;
    function GetPackageNames(index: integer): String;
    procedure SetRegistryKey(const Value: String);
    procedure SetIsFinalizingPackageRequired(const Value: Boolean);
  public
    function AddElement( anElement : TDynamicLibraryElementClass) : integer;
      virtual;
    function AddPackage( aHandle : HModule; aPath : String) : integer; virtual;
    constructor Create; virtual;
    destructor Destroy; override;
    function isPackageLoaded( aPackagePath : String) : boolean; virtual;
    procedure RemoveElement( anElement : TDynamicLibraryElementClass) ;

    { This procedure loads the elements, after the loading of the package
      by calling the different InitialiseUnitElements of the units of the package.
      This procedure must be called by the main program, and not by the package.
      }
    Procedure LoadElements( PackageName : String);

    { Procedure to load a package knowing its name and path. The UnitNames string
      list must be initialized. It is emptied and the list of units of the package
      is returned.}
    Function LoadPackage( PackageName : String; UnitNames : TStrings) : HMODULE;

    { This procedure must be called to unload the package from the application.
      When this occurs, all objects of the classes declared in the package must be
      freed. This is the responsibility of the classes declared in the package. The
      full path must be provided.}
    Procedure UnloadPackage( PackagePath : String);

    { This procedure allows the DLE to be registered into the Windows Registry.
      The package file name should have the full path name, in order to load
      it from the disk.}
    procedure RegisterDLEIntoRegistry( aDLEClassName : String;
                              aPackageFileName : String);

    procedure UnregisterDLE( aDLEClassName : String);

    { This procedure loads all DLE packages registered in the Windows registry.}
    procedure LoadRegisteredPackages;

    { This procedure unloads all DLE packages registered in the Windows
      registry.}
    procedure UnLoadRegisteredPackages;
    property Count : integer read GetCount;
    property Elements[ index : integer] : TDynamicLibraryElementClass read GetElements
      write SetElements;

    { This variable can be set to true, only if you are sure your package does not
      cause an AV in its finalization. Set it BEFORE calling UnloadPackage.}
    property IsFinalizingPackageRequired : Boolean read FIsFinalizingPackageRequired write SetIsFinalizingPackageRequired;

    { List of packages names linked to elements : the index is the same as for
      the elements.}
    property PackageNames[ index : integer] : String read GetPackageNames;

    { This property can be set to indicate the standard DLE storage or any
      specific one. Set it up before calling any registry-dedicated procedure}
    property RegistryKey : String read FRegistryKey write SetRegistryKey;
  end;

  { A package descriptor is used to maintain the list of packages available}
  TPackageDescriptor = class
  private
    FPackageHandle: HMODULE;
    FPackagePath: String;
    procedure SetpackageHandle(const Value: HMODULE);
    procedure SetpackagePath(const Value: String);
    function GetPackageHandle: HMODULE;
    function GetPackagePath: String;
  published
    property packageHandle : HMODULE read GetPackageHandle write SetpackageHandle;
    property packagePath : String read GetPackagePath write SetpackagePath;
  end;

{ operations below allow to manage the unique DLE manager
  in the system. There should be only one. (Adaptable Singleton pattern)
  }
function DLEManager : TDLEManager;
procedure SetDLEManager( aManager : TDLEManager);

{ Procedure example that must be implemented for each unit in packages that
  are to be linked dynamically. The DLE Manager looks for a procedure named like
  it in each and every unit loaded with the package.}
Procedure InitialiseUnitElements( aManager : TDLEManager);



type
  TLoadElementProc = Procedure( aManager : TDLEManager);

implementation

uses SysUtils, PackageUtilities,
     Registry, Dialogs;

var
  FDLEManager : TDLEManager;
  FHelp: String;
  FCommandText: String;
  FDescription : String;
  FDLEInstanceList : TList;
  FCurrentPackageName : String;

function DLEManager : TDLEManager;
begin
  result := FDLEManager;
end;

Procedure InitialiseUnitElements( aManager : TDLEManager);
begin
  DLEManager.AddElement( TDynamicLibraryElement);
end;

Procedure TDLEManager.LoadElements( PackageName : String);
var FModuleHandle : HINST;
    aLoadModuleProc : TLoadElementProc;
    procName   : String;
    unitName   : String;
    unitList   : TStringList;
    iUnit      : integer;
    lastKnownElement : Integer;
    iElt             : integer;
    n1, iP           : integer;
begin
  FModuleHandle := GetModuleHandle( PChar( PackageName));
  unitList      := TStringList.Create;
  lastKnownElement := Self.Count-1;
  GetListOfUnits( PackageName, unitList);
  try
    for iUnit := 0 to unitList.Count-1 do
    begin
      unitName := CapitalizeFirstLetterOnly( unitList[ iUnit]);
      procName := '@' +unitName
        +'@InitialiseUnitElements$qqrp33Dynamiclibraryelement@TDLEManager';
      aLoadModuleProc := GetProcAddress( FModuleHandle, PChar(procName));
      if assigned( aLoadModuleProc) then
        aLoadModuleProc( Self);
    end;
  finally
    FreeAndNil( unitList);
  end;

  // setting the package name of each new element :
  n1 := DLEManager.FPackageNamesList.Count;
  for iP := n1  to DLEManager.Count do
    DLEManager.FPackageNamesList.Add('');
  for iElt := lastKnownElement+1 to DLEManager.Count-1 do
  begin
    ShowMessage( 'Adding Package name to elt : ' + intToStr( iElt) + ' : ' +
                 FCurrentPackageName);
    DLEManager.FPackageNamesList[ iElt] := FCurrentPackageName
  end;
end;

Function TDLEManager.LoadPackage( PackageName : String;
                                  UnitNames : TStrings) : HMODULE;
begin
  result := 0;
  FCurrentPackageName := lowercase( ChangeFileExt( ExtractFileName(
    PackageName), ''));
  try
    result := SysUtils.LoadPackage( PackageName);
    AddPackage( result, PackageName);
    GetListOfUnits( PackageName, UnitNames);
    LoadElements( PackageName);
  finally
    FCurrentPackageName := '';
  end;
end;

Procedure TDLEManager.UnloadPackage( PackagePath : String);
var iElt          : integer;
    eltToRemove   : TDynamicLibraryElementClass;
    iP            : Integer;
    aPackage      : TPackageDescriptor;
    packageName   : String;
    packageHandle : HModule;
begin
  packageName   := lowercase( ChangeFileExt( ExtractFileName( PackagePath), ''));
  packageHandle := 0;

  // identifying the package
  for iP := 0 to FPackageList.Count-1 do
  begin
    aPackage := TPackageDescriptor( FPackageList[ iP]);
    if lowerCase( aPackage.FPackagePath) = lowerCase( PackagePath) then
    begin
      packageHandle := aPackage.FPackageHandle;
      FPackageList.Delete( iP);
      FreeAndNil( aPackage);
      break;
    end;
  end;
  if packageHandle = 0 then
    raise Exception.Create( 'package "'+packagePath+'" not found!');

  // removing the elements belonging to the package
  try
    for iElt := DLEManager.Count-1 downto 0 do
    begin
      ShowMessage( intToStr( iElt) +' : ' +PackageNames[ iElt]
                   +'=' +PackageName +'?');
      if PackageNames[ iElt]= PackageName then
      begin
        eltToRemove := Elements[ iElt];
        RemoveElement( eltToRemove);
        eltToRemove.FreeAllInstances( eltToRemove);
      end;
    end;
  except
    // any errors that occur here should be either ignored or logged, not more!
  end;

  // finally, removing the package itself :
  try
    if IsFinalizingPackageRequired then
      SysUtils.UnloadPackage( packageHandle)
    else
      FreeLibrary( packageHandle);
  except
  end;
end;

procedure TDLEManager.RegisterDLEIntoRegistry( aDLEClassName : String;
                                   aPackageFileName : String);
var
  reg : TRegistryIniFile;
begin
  reg := TRegistryIniFile.Create( 'UML Component Factory');
  try
    reg.WriteString( RegistryKey, aDLEClassName, aPackageFileName);
  finally
    FreeAndNil( reg);
  end;
end;

procedure TDLEManager.UnregisterDLE( aDLEClassName : String);
var
  reg : TRegistryIniFile;
begin
  reg := TRegistryIniFile.Create( 'UML Component Factory');
  try
    reg.DeleteKey( RegistryKey, aDLEClassName);
  finally
    FreeAndNil( reg);
  end;
end;

procedure TDLEManager.LoadRegisteredPackages;
var
  reg : TRegistryIniFile;
  values : TStringList;
  iElt : integer;
  aElt : String;
  aEltPackage : String;
begin
  values := nil;
  reg := TRegistryIniFile.Create( 'UML Component Factory');
  try
    values := TStringList.Create;
    reg.ReadSectionValues( RegistryKey, values);
    for iElt := 0 to values.Count-1 do
    begin
      aElt := values.Names[ iElt];
      aEltPackage := values.values[ aElt];
      if not DLEManager.isPackageLoaded( aEltPackage) then
        LoadPackage( aEltPackage, nil);
    end;
  finally
    FreeAndNil( reg);
    FreeAndNil( values);
  end;
end;

procedure TDLEManager.UnLoadRegisteredPackages;
var
  reg : TRegistryIniFile;
  values : TStringList;
  iElt : integer;
  aElt : String;
  aEltPackage : String;
begin
  values := nil;
  reg := TRegistryIniFile.Create( 'UML Component Factory');
  try
    values := TStringList.Create;
    reg.ReadSectionValues( RegistryKey, values);
    for iElt := values.Count-1 downto 0 do
    begin
      aElt := values.Names[ iElt];
      aEltPackage := values.values[ aElt];
      if DLEManager.isPackageLoaded( aEltPackage) then
        UnloadPackage( aEltPackage);
    end;
  finally
    FreeAndNil( reg);
    FreeAndNil( values);
  end;
end;

procedure SetDLEManager( aManager : TDLEManager);
begin
  FreeAndNil( FDLEManager);
  FDLEManager := aManager;
end;

{ TDLEManager }

function TDLEManager.AddElement(anElement: TDynamicLibraryElementClass): integer;
begin
  result := FDLEList.add( anElement);
end;

function TDLEManager.AddPackage(aHandle: HModule; aPath: String): integer;
var
  newPackage : TPackageDescriptor;
begin
  newPackage := TPackageDescriptor.Create;
  newPackage.FPackageHandle := aHandle;
  newPackage.FPackagePath := aPath;
  FPackageList.Add( newPackage);
end;

constructor TDLEManager.Create;
begin
  inherited;
  FDLEList := TList.Create;
  FPackageNamesList := TStringList.Create;
  FPackageList := TList.Create;
  FRegistryKey := 'Dynamic Library Elements';
  FIsFinalizingPackageRequired := false;
end;

destructor TDLEManager.Destroy;
begin
  inherited;
  FreeAndNil( FDLEList);
  FreeAndNil( FPackageNamesList);
  FreePackageList;
end;

procedure TDLEManager.FreePackageList;
var
  iP : integer;
  aP : TPackageDescriptor;
begin
  for iP := FPackageList.Count-1 downto 0 do
  begin
    aP := TPackageDescriptor(FPackageList[ iP]);
    aP.Free;
  end;
  FreeAndNil( FPackageList);
end;

function TDLEManager.GetCount: integer;
begin
  result := FDLEList.Count;
end;

function TDLEManager.GetElements(index: integer): TDynamicLibraryElementClass;
begin
  result := TDynamicLibraryElementClass( FDLEList[ index]);
end;

function TDLEManager.GetPackageNames(index: integer): String;
begin
  result := FPackageNamesList[ index];
end;

function TDLEManager.isPackageLoaded(aPackagePath: String): boolean;
var
  aPackage : TPackageDescriptor;
  iPackage : integer;
begin
  aPackagePath := lowerCase( aPackagePath);
  result       := false;
  for iPackage := 0 to FPackageList.Count-1 do
  begin
    aPackage := TPackageDescriptor( FPackageList[ iPackage]);
    if assigned( aPackage) then
    begin
      if lowerCase( aPackage.FPackagePath) = aPackagePath then
      begin
        result := true;
        exit;
      end;
    end;
  end;
end;

procedure TDLEManager.RemoveElement(anElement: TDynamicLibraryElementClass);
begin
  FDLEList.remove( anElement);
end;

procedure TDLEManager.SetElements(index: integer; const Value: TDynamicLibraryElementClass);
begin
  if Count <= Index then
    FDLEList.Count := Index+1;
  FDLEList[ index] := value;
end;

procedure TDLEManager.SetRegistryKey(const Value: String);
begin
  FRegistryKey := Value;
end;

procedure TDLEManager.SetIsFinalizingPackageRequired(const Value: Boolean);
begin
  FIsFinalizingPackageRequired := Value;
end;

{ TDynamicLibraryElement }

class procedure TDynamicLibraryElement.AddInstance(
  Value: TDynamicLibraryElement);
begin
  FDLEInstanceList.Add( Value);
end;

constructor TDynamicLibraryElement.Create(anOwner: TComponent);
begin
  inherited;
  AddInstance( Self);
end;

destructor TDynamicLibraryElement.Destroy;
begin
  RemoveInstance( Self);
  inherited;
end;

procedure TDynamicLibraryElement.Execute;
begin
  // does nothing here. To be overriden by descendents.
end;

class procedure TDynamicLibraryElement.FreeAllInstances(
  aClass : TDynamicLibraryElementClass);
var
  iInstance : integer;
  anInstance : TDynamicLibraryElement;
begin
  for iInstance := FDLEInstanceList.Count-1 downto 0 do
  begin
    anInstance := GetInstances( iInstance);
    if assigned( anInstance) then
      if anInstance.ClassType = aClass then
      begin
        SetInstances( iInstance, nil);
        anInstance.Free;
      end;
  end;
  FDLEInstanceList.Pack;
  FDLEInstanceList.Capacity := FDLEInstanceList.Count;
end;

class function TDynamicLibraryElement.GetCommandText: String;
begin
  result := FCommandText;
end;

class function TDynamicLibraryElement.GetDescription: String;
begin
  result := FDescription;
end;

class function TDynamicLibraryElement.GetHelp: String;
begin
  result := FHelp;
end;

class function TDynamicLibraryElement.GetInstances(
  index: integer): TDynamicLibraryElement;
begin
  result := TDynamicLibraryElement(FDLEInstanceList[ index]);
end;

class function TDynamicLibraryElement.InstanceCount: integer;
begin
  result := FDLEInstanceList.Count;
end;

class procedure TDynamicLibraryElement.RemoveInstance(
  Value: TDynamicLibraryElement);
begin
  FDLEInstanceList.Remove( Value);
end;

class procedure TDynamicLibraryElement.SetCommandText(const Value: String);
begin
  FCommandText := Value;
end;

class procedure TDynamicLibraryElement.SetDescription(const Value: String);
begin
  FDescription := Value
end;

class procedure TDynamicLibraryElement.SetHelp(const Value: String);
begin
  FHelp := Value;
end;

class procedure TDynamicLibraryElement.SetInstances(index: integer;
  const Value: TDynamicLibraryElement);
begin
  FDLEInstanceList[ index] := Value;
end;

{ TPackageDescriptor }

function TPackageDescriptor.GetPackageHandle: HMODULE;
begin

end;

function TPackageDescriptor.GetPackagePath: String;
begin

end;

procedure TPackageDescriptor.SetpackageHandle(const Value: HMODULE);
begin
  FPackageHandle := Value;
end;

procedure TPackageDescriptor.SetpackagePath(const Value: String);
begin
  FPackagePath := Value;
end;

initialization
  FDLEManager                 := TDLEManager.Create;
  FDLEInstanceList            := TList.Create;
finalization
  FreeAndNil( FDLEManager);
  FDescription := '';
  FreeAndNil( FDLEInstanceList);
end.
