unit PropsSaverEditors;

interface

uses
  classes, DsgnIntf;

type
  TPropertyNameEditor = class (TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TPropSaverComponentEditor = class (TComponentProperty)
  public
    procedure GetValues (Proc : TGetStrProc); override;
    procedure SetValue(const Value: string); override;
  end;

  TPropSaverHKEYProperty = class(TIntegerProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
    function GetValue: string; override;
    procedure SetValue(const Value: string); override;
  end;

  TLoadSavePerformersProperty = class (TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TLayoutIdProperty = class (TStringProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure GetValues(Proc: TGetStrProc); override;
  end;

  TPropsSaverEditor = class (TComponentEditor)
  public
    procedure Edit; override;
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

implementation

uses
  PropsSaver, TypInfo, sysUtils, Windows, TwoColsEditor, PropsSaverEditor, Forms,
  PropsSaverManageWizzards, PropsSaverFakeIniProps, PropsSaverPasteFromTIniPropSavForm,
  Controls, PropsSaverBasicClassTranslators;


{ TPropertyNameEditor }

function TPropertyNameEditor.GetAttributes;
begin
  Result := [paValueList, paSortList];
end;

procedure TPropertyNameEditor.GetValues;
var
  PropInfoItem : TPropInfoItem;
  PropList : PPropList;
  PropCount : Integer;
  i, n : Integer;
  Comp : TObject;
  PropTypes : TTypeKinds;
  NonPublishedProps : TStringList;
  NonPubProp : TNonPublishedProperty;
begin
  PropInfoItem := (GetComponent (0) as TPropInfoItem);
  if (PropInfoItem.Collection as TPropInfoCollection).CompInfoItem.Component <> nil
    then Comp := (PropInfoItem.Collection as TPropInfoCollection).CompInfoItem.Component
    else Comp := nil;
  if Comp <> nil
    then
    begin
      PropCount := GetTypeData (Comp.ClassInfo)^.PropCount;
      GetMem (PropList, PropCount * SizeOf (PPropInfo));
      try
        case PropInfoItem.PropertyType of
          ptAny : PropTypes := tkProperties + tkMethods;
          ptMethods : PropTypes := tkMethods;
          ptProperties : PropTypes := tkProperties;
        end;
        n := GetPropList (Comp.ClassInfo, PropTypes, PropList);
        for i := 0 to n - 1 do
          Proc (PropList^ [i].Name);
      finally
        FreeMem (PropList, PropCount * SizeOf (PPropInfo));
      end;
      NonPublishedProps := TStringList.Create;
      try
        GetNonPublishedProperties (Comp, NonPublishedProps);
        for i := 0 to NonPublishedProps.Count - 1 do
          Proc ('*' + NonPublishedProps [i]); // We add a * to the non-published properties to mark them
      finally
        NonPublishedProps.Free;
      end;
      if PropInfoItem.Collection.CompInfoItem.Collection.IsSubProp
        then with (PropInfoItem.Collection.CompInfoItem.Collection as TCompInfoCollectionSubProp).PropInfoItem do
          begin
            NonPublishedProps := TStringList.Create;
            try
              NonPubProp := GetNonPublishedProperty (Component, PropertyName);
              if (NonPubProp <> nil) and Assigned (NonPubProp.GetNonPublishedPropertySubProps) and (component <> nil)
                then
                begin
                  NonPubProp.GetNonPublishedPropertySubProps (Component, PropertyName, NonPublishedProps);
                  for i := 0 to NonPublishedProps.Count - 1 do
                    Proc ('+' + NonPublishedProps [i]); // We add a + to subproperties of non-published properties to mark them
                end
                else if (CompInfoItemSubProps.Component <> nil) and GetSubProperties (CompInfoItemSubProps.Component, NonPublishedProps)
                  then for i := 0 to NonPublishedProps.Count - 1 do
                    Proc ('+' + NonPublishedProps [i]) // We add a + to subproperties of non-published properties to mark them
            finally
              NonPublishedProps.Free;
            end;
          end;
    end;
end;

{ TPropSaverComponentEditor }

procedure TPropSaverComponentEditor.GetValues;
begin
  inherited GetValues (Proc);
  Proc (Designer.Form.Name);
end;

procedure TPropSaverComponentEditor.SetValue;
begin
  if uppercase (Value) = uppercase (Designer.Form.Name)
    then SetOrdValue(Longint(Designer.Form))
    else inherited;
end;

{ TPropSaverHKEYProperty }

type
  HKEYArray = array [HKEY_CLASSES_ROOT..HKEY_DYN_DATA] of string;

const
  HKEYS : HKEYArray = ('HKEY_CLASSES_ROOT', 'HKEY_CURRENT_USER',
    'HKEY_LOCAL_MACHINE', 'HKEY_USERS', 'HKEY_PERFORMANCE_DATA',
    'HKEY_CURRENT_CONFIG', 'HKEY_DYN_DATA');

function HKEYToIdent (aint : longint; var Ident : string) : boolean;
begin
  Result := aint <= longint (high (HKEYS));
  if Result
    then Ident := HKEYS [aint];
end;

function IdentToHKEY (const Ident : string; var Int : longint) : boolean;
var
  i : HKEY;
begin
  Result := true;
  for i := low (HKEYS) to high (HKEYS) do
    if HKEYS [i] = Ident
      then
      begin
        Int := i;
        exit;
      end;
  Result := false;
end;

function TPropSaverHKEYProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paValueList];
end;

procedure TPropSaverHKEYProperty.GetValues;
var
  i : HKEY;
begin
  for i := low (HKEYS) to high (HKEYS) do
    Proc (HKEYS [i]);
end;

procedure TPropSaverHKEYProperty.SetValue;
var
  Key : longint;
begin
  if IdentToHKEY (Value, Key)
    then SetOrdValue (Key)
    else inherited SetValue (Value);
end;

function TPropSaverHKEYProperty.GetValue;
begin
  if not HKEYToIdent (GetOrdValue, Result)
    then Result := inherited GetValue;
end;

{ TLoadSavePerformersProperty }

function TLoadSavePerformersProperty.GetAttributes;
begin
  Result := inherited GetAttributes + [paValueList, paSortList];
end;

procedure TLoadSavePerformersProperty.GetValues;
var
  i : Integer;
begin
  for i := 0 to TPropsSaver.LoadSavePerformers.Count - 1 do
    Proc (TLoadSavePerformerClass (TPropsSaver.LoadSavePerformers [i]).ClassName);
end;

{ TLayoutIdProperty }

function TLayoutIdProperty.getAttributes;
begin
  Result := inherited GetAttributes + [paValueList, paSortList];
end;

procedure TLayoutIdProperty.GetValues;
var
  i : integer;
  LayoutIdList : TStringList;
begin
  LayoutIdList := TStringList.Create;
  try
    LayoutIdList.sorted := True;
    LayoutIdList.duplicates := dupIgnore;
    for i := 0 to TPropsSaver.PropsSavers.Count - 1 do
      if TPropsSaver (TPropsSaver.PropsSavers [i]).LayoutId <> ''
        then LayoutIdList.Add (TPropsSaver (TPropsSaver.PropsSavers [i]).LayoutId);
    for i := 0 to LayoutIdList.Count - 1 do
      Proc (LayoutIdList [i]);
  finally
    LayoutIdList.Free;
  end;
end;

{ TPropsSaverEditor }

function TPropsSaverEditor.GetVerb;
begin
  case Index of
    0 : Result := 'PropSaver Editor...';
    1 : Result := 'Wizzards Manager...';
    2 : Result := 'Paste from TIniPropSav...';
  end;
end;

function TPropsSaverEditor.GetVerbCount;
begin
  Result := 3;
end;

procedure TPropsSaverEditor.Edit;
begin
  ShowTwoColsEditor (Designer, component as TPropsSaver, TFormPropsSaverEditor);
end;

procedure TPropsSaverEditor.ExecuteVerb;
var
  Fake : TFakeIniPropsSav;
begin
  case Index of
    0 : Edit;
    1 :
      begin
        if FormManageWizzards = nil
          then
          begin
            FormManageWizzards := TFormManageWizzards.Create (Application);
            FormManageWizzards.FramePropertiesWizzards.BaseKey := TFormTwoColsEditor.GetRegKey;
            FormManageWizzards.FramePropertiesWizzards.Obj := nil;
          end;
        FormManageWizzards.Show;
      end;
    2 :
      begin
        FormPasteFromTIniPropSav := TFormPasteFromTIniPropSav.Create (Application);
        try
          if FormPasteFromTIniPropSav.ShowModal = mrOk
            then
            begin
              Fake := TFakeIniPropsSav.Create (Component.Owner);
              try
                FormPasteFromTIniPropSav.MemoSource.Lines [0] := 'object Fake: TFakeIniPropsSav';
                StringToComponent (FormPasteFromTIniPropSav.MemoSource.Lines.Text, Fake, nil);
                Fake.AssignToPropsSaver (Component as TPropsSaver);
              finally
                Fake.Free;
              end;
              Designer.Modified;
            end;
        finally
          FormPasteFromTIniPropSav.Free;
        end;
      end;
  end;
end;

end.
