unit PropsSaverFakeIniProps;

interface

uses
  Classes, PropsSaver, sysUtils, Windows;

type
  TSpecialProperties = (npInteger, npEnumeration, npFloat);
  TSpecialPropertiesSet = set of TSpecialProperties;
  TPropertyWork = procedure (Sender : TObject; Component : TComponent; const PropName : string) of object;
  TPropertyError = procedure (Sender : TObject; Component : TComponent;
                              const PropName, Value : string; E : Exception) of object;

  TFakeIniPropsSav = class (TComponent)
  private
    FFileName : string;
    FUseAppDir : Boolean;
    FAutoLoad : TAutoLoad;
    FAutoSave : TAutoSave;
    FStoreAsRaw : TSpecialPropertiesSet;
    FLoadProperty, FSaveProperty : TPropertyWork;
    FLoadError, FSAveError : TPropertyError;
    FStartLoadingProperties, FEndLoadingProperties,
    FStartSavingProperties, FEndSavingProperties : TNotifyEvent;
    FEncrypted, FUseRegistry, FLazyWrite, FUseSoftwareKey : Boolean;
    FCompanyName, FProductName : string;
    FRootKey : HKEY;
    FSavedProperties : TStrings;
    FSections : TStrings;
    FOwner : TComponent;
    procedure UpdateObjects;
    procedure ReadProperties (Reader : TReader);
    procedure WriteProperties (Writer : TWriter);
    procedure ReadSections (Reader : TReader);
    procedure WriteSections (Writer : TWriter);
    procedure SetSavedProperties (AProp : TStrings);
    procedure SetSections (ASecc : TStrings);
    function IndexOf (Component : TComponent; const Prop : string) : Integer;
    function GetSavedComponentsCount : integer;
    function GetSavedComponents (Index : integer) : TComponent;
    function GetSavedProperties (Index : integer) : string;
    function GetSections (Component : TComponent; const Prop : string) : string;
    procedure SetSections2 (Component : TComponent; const Prop, Secc : string);
    function GetIdentifiers (Component : TComponent; const Prop : string) : string;
    procedure SetIdentifiers (Component : TComponent; const Prop, Ident : string);
    function GetMustLoad (Component : TComponent; const Prop : string) : boolean;
    procedure SetMustLoad (Component : TComponent; const Prop : string; Value : boolean);
    function GetMustSave (Component : TComponent; const Prop : string) : boolean;
    procedure SetMustSave (Component : TComponent; const Prop : string; Value : boolean);
  protected
    procedure DefineProperties (Filer : TFiler); override;
    property Properties : TStrings read FSavedProperties write SetSavedProperties;
    property Sections : TStrings read FSections write SetSections;
  public
    constructor Create (AOwner : TComponent); override;
    destructor Destroy; override;
    procedure AssignToPropsSaver (PropsSaver : TPropsSaver);
    property SavedComponentsCount : integer read GetSavedComponentsCount;
    property SavedComponents [Index : integer] : TComponent read GetSavedComponents;
    property SavedProperties [Index : integer] : string read GetSavedProperties;
    property INISections [Component : TComponent; const Prop : string] : string read GetSections write SetSections2;
    property INIIdentifiers [Component : TComponent; const Prop : string] : string read GetIdentifiers write SetIdentifiers;
    property MustLoadProperty [Component : TComponent; const Prop : string] : boolean read GetMustLoad write SetMustLoad;
    property MustSaveProperty [Component : TComponent; const Prop : string] : boolean read GetMustSave write SetMustSave;
  published
    property FileName : string read FFileName Write FFileName;
    property UseAppDir : boolean read FUseAppDir write FUseAppDir default true;
    property AutoLoadProperties : TAutoLoad read FAutoLoad write FAutoLoad default aldAfterOnCreateForm;
    property AutoSaveProperties : TAutoSave read FAutoSave write FAutoSave default asvBeforeOnDestroyForm;
    property StoreAsRawData : TSpecialPropertiesSet read FStoreAsRaw write FStoreAsRaw;
    property OnLoadProperty : TPropertyWork read FLoadProperty write FLoadProperty;
    property OnSaveProperty : TPropertyWork read FSaveProperty write FSaveProperty;
    property OnLoadPropertyError : TPropertyError read FLoadError write FLoadError;
    property OnSavePropertyError : TPropertyError read FSAveError write FSaveError;
    property OnStartLoadingProperties : TNotifyEvent read FStartLoadingProperties write FStartLoadingProperties;
    property OnEndLoadingProperties : TNotifyEvent read FEndLoadingProperties write FEndLoadingProperties;
    property OnStartSavingProperties : TNotifyEvent read FStartSavingProperties write FStartSavingProperties;
    property OnEndSavingProperties : TNotifyEvent read FEndSavingProperties write FEndSavingProperties;
    {$IfDef Win32}
    property Encrypted : boolean read FEncrypted Write FEncrypted default false;
    property UseRegistry : boolean read FUseRegistry Write FUseRegistry default False;
    property CompanyName : string read FCompanyName write FCompanyName;
    property ProductName : string read FProductName write FProductName;
    property RootKey : HKEY read FRootKey Write FRootKey default HKEY_CURRENT_USER;
    property UseSoftwareKey : boolean read FUseSoftwareKey write FUseSoftwareKey default true;
    property LazyWrite : boolean read FLazyWrite Write FLazyWrite default true;
    {$EndIf}
  end;

implementation

resourcestring
  ERR_PropNotFound = 'Property not found';

function StringForList (const Comp, Prop : string) : string; forward;
function ExtractSeccion (const Cad : string) : string; forward;
function ExtractItem (const Cad : string) : string; forward;
function ExtractValue (const Cad : string) : string; forward;
function ExtractComponent (const Cad : string) : string; forward;
function ExtractProperty (const Cad : string) : string; forward;
function ExtractGuardar (const Cad : string) : boolean; forward;
function ExtractCargar (const Cad : string) : Boolean; forward;

constructor TFakeIniPropsSav.Create;
begin
  inherited Create (AOwner);
  FSavedProperties := TStringList.Create;
  FSections := TStringList.Create;
  FUseAppDir := true;
  FAutoLoad := aldAfterOnCreateForm;
  FAutoSave := asvBeforeOnDestroyForm;
  FStoreAsRaw := [];
  {$IfDef Win32}
  FRootKey := HKEY_CURRENT_USER;
  FUseSoftwareKey := true;
  FLazyWrite := true;
  {$EndIf}
end;

destructor TFakeIniPropsSav.Destroy;
begin
  FSavedProperties.Clear;
  FSavedProperties.Free;
  FSections.Free;
  inherited Destroy;
end;

procedure TFakeIniPropsSav.DefineProperties;
begin
  inherited DefineProperties (Filer);
  Filer.DefineProperty ('Properties', ReadProperties, WriteProperties, FSavedProperties.Count > 0);
  Filer.DefineProperty ('Sections', ReadSections, WriteSections, FSections.Count > 0);
end;

procedure TFakeIniPropsSav.ReadProperties;
begin
  (FSavedProperties as TStringList).Sorted := false;
  Reader.ReadListBegin;
  while not Reader.EndOfList do
    FSavedProperties.Add (Reader.ReadString);
  Reader.ReadListEnd;
end;

procedure TFakeIniPropsSav.ReadSections;
begin
  Reader.ReadListBegin;
  while not Reader.EndOfList do
    FSections.Add (Reader.ReadString);
  Reader.ReadListEnd;
end;

procedure TFakeIniPropsSav.WriteProperties;
var
 i : integer;
begin
  Writer.WriteListBegin;
  for i := 0 to FSavedProperties.Count - 1 do
    Writer.WriteString (FSavedProperties [i]);
  Writer.WriteListEnd;
end;

procedure TFakeIniPropsSav.WriteSections;
var
 i : integer;
begin
  Writer.WriteListBegin;
  for i := 0 to FSections.Count - 1 do
    Writer.WriteString (FSections [i]);
  Writer.WriteListEnd;
end;

procedure TFakeIniPropsSav.SetSavedProperties;
begin
  if AProp <> FSavedProperties
    then FSavedProperties.Assign (AProp);
end;

procedure TFakeIniPropsSav.SetSections;
begin
  if ASecc <> FSections
    then FSections.Assign (ASecc);
end;

procedure TFakeIniPropsSav.AssignToPropsSaver;
var
  i : Integer;
  Curcomp : TComponent;
  CurCompInfo : TCompInfoItem;
begin
  FOwner := PropsSaver.Owner;
  PropsSaver.AutoLoadProperties := AutoLoadProperties;
  PropsSaver.AutoSaveProperties := AutoSaveProperties;
  PropsSaver.RootKey := RootKey;
  if UseSoftwareKey
    then if CompanyName <> ''
      then if ProductName <> ''
        then PropsSaver.BaseKey := Format ('\Software\%s\%s', [CompanyName, ProductName])
        else PropsSaver.BaseKey := Format ('\Software\%s', [CompanyName])
      else if ProductName <> ''
        then PropsSaver.BaseKey := Format ('\Software\%s', [ProductName])
        else PropsSaver.BaseKey := '\Software'
    else PropsSaver.BaseKey := FileName;
  PropsSaver.FileName := FileName;
  PropsSaver.BeforeLoadingProperties := OnStartLoadingProperties;
  PropsSaver.AfterLoadingProperties := OnEndLoadingProperties;
  PropsSaver.BeforeSavingProperties := OnStartSavingProperties;
  PropsSaver.BeforeSavingProperties := OnEndSavingProperties;
  if UseRegistry
    then PropsSaver.LoadSavePerformerClassName := 'TRegistryLoadSavePerformer'
    else PropsSaver.LoadSavePerformerClassName := 'TIniFileLoadSavePerformer';
  UpdateObjects;
  Curcomp := nil;
  CurCompInfo := nil;
  for i := 0 to SavedComponentsCount - 1 do
    begin
      if Curcomp <> SavedComponents [i]
        then
        begin
          Curcomp := SavedComponents [i];
          CurCompInfo := PropsSaver.Components.Add;
          CurCompInfo.Component := Curcomp;
        end;
      if CurCompInfo <> nil
        then
        begin
          with CurCompInfo.Properties.Add do
            begin
              BaseKey := BaseKey + '\' + INISections [Curcomp, SavedProperties [i]];
              ItemName := INIIdentifiers [Curcomp, SavedProperties [i]];
              PropertyName := SavedProperties [i];
              IfClassStoreAsLink := ExtractValue (Sections [i]) = '1';
              if not MustLoadProperty [Curcomp, SavedProperties [i]]
                then Mode := Mode - [lsmLoad];
              if not MustSaveProperty [Curcomp, SavedProperties [i]]
                then Mode := Mode - [lsmSave];
              Encrypted := Self.Encrypted;
            end;
        end;
    end;
end;

procedure TFakeIniPropsSav.UpdateObjects;
var
  i : integer;
  Comp : TComponent;
  cad : string;
begin
  for i := 0 to FSavedProperties.Count - 1 do
    begin
      Cad := ExtractComponent (FSavedProperties [i]);
      if Cad <> 'Form'
        then Comp := FOwner.FindComponent (Cad)
        else Comp := FOwner;
      FSavedProperties.Objects [i] := Comp;
    end;
end;

function TFakeIniPropsSav.GetSavedComponentsCount;
begin
  Result := FSavedProperties.Count;
end;

function TFakeIniPropsSav.GetSavedComponents (Index : integer) : TComponent;
begin
  Result := FSavedProperties.Objects [Index] as TComponent;
end;

function TFakeIniPropsSav.GetSavedProperties (Index : integer) : string;
begin
  Result := ExtractProperty (FSavedProperties [Index]);
end;

function TFakeIniPropsSav.IndexOf;
var
  CompName : string;
begin
  if Component <> FOwner
    then CompName := Component.Name
    else CompName := 'Form';
  Result := FSavedProperties.IndexOf (StringForList (CompName, Prop));
end;

function TFakeIniPropsSav.GetSections (Component : TComponent; const Prop : string) : string;
var
  Index : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then Result := ExtractSeccion (FSections [Index])
    else raise Exception.Create (ERR_PropNotFound);
end;

procedure TFakeIniPropsSav.SetSections2 (Component : TComponent; const Prop, Secc : string);
var
  Index : integer;
  Cad : string;
  p : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then
    begin
      Cad := FSections [Index];
      p := pos ('+', Cad);
      if p <> 0
        then FSections [Index] := Secc + '+' + copy (Cad, p + 1, length (Cad) - p);
    end
    else raise Exception.Create (ERR_PropNotFound);
end;

function TFakeIniPropsSav.GetIdentifiers (Component : TComponent; const Prop : string) : string;
var
  Index : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then Result := ExtractItem (FSections [Index])
    else raise Exception.Create (ERR_PropNotFound);
end;

procedure TFakeIniPropsSav.SetIdentifiers (Component : TComponent; const Prop, Ident : string);
var
  Index : integer;
  Cad : string;
  p, p2 : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then
    begin
      Cad := FSections [Index];
      p := pos ('+', Cad);
      p2 := pos ('.', Cad);
      if (p <> 0) and (p2 <> 0)
        then FSections [Index] := copy (Cad, 1, p) + Ident + copy (Cad, p2, length (Cad) - p2 + 1);
    end
    else raise Exception.Create (ERR_PropNotFound);
end;

function TFakeIniPropsSav.GetMustLoad (Component : TComponent; const Prop : string) : boolean;
var
  Index : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then Result := ExtractCargar (FSections [Index])
    else raise Exception.Create (ERR_PropNotFound);
end;

procedure TFakeIniPropsSav.SetMustLoad (Component : TComponent; const Prop : string; Value : boolean);
var
  Index : integer;
  Cad : string;
  p, p2 : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then
    begin
      Cad := FSections [Index];
      p := pos ('.', Cad);
      p2 := pos ('#', Cad);
      if (p <> 0) and (p2 <> 0)
        then FSections [Index] := copy (Cad, 1, p) + IntToStr (integer (Value)) + copy (Cad, p2, length (Cad) - p2 + 1);
    end
    else raise Exception.Create (ERR_PropNotFound);
end;

function TFakeIniPropsSav.GetMustSave (Component : TComponent; const Prop : string) : boolean;
var
  Index : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then Result := ExtractGuardar (FSections [Index])
    else raise Exception.Create (ERR_PropNotFound);
end;

procedure TFakeIniPropsSav.SetMustSave (Component : TComponent; const Prop : string; Value : boolean);
var
  Index : integer;
  Cad : string;
  p : integer;
begin
  Index := IndexOf (Component, Prop);
  if Index <> -1
    then
    begin
      Cad := FSections [Index];
      p := pos ('#', Cad);
      if p <> 0
        then FSections [Index] := copy (Cad, 1, p) + IntToStr (integer (Value));
    end
    else raise Exception.Create (ERR_PropNotFound);
end;

{ Functions }

function InvertPos (PosCad, Cad : string) : byte;
var
  i, Posit : byte;
begin
  Posit := 0;
  i := length (Cad);
  while (i >= 1) and (Posit = 0) do
    begin
      Posit := pos (PosCad, copy (Cad, i, length (Cad) - i + 1));
      dec (i);
    end;
  if Posit <> 0
    then InvertPos := i + 1
    else InvertPos := 0;
end;

function StringForList;

begin
  Result := Format ('%s+%s', [Comp, Prop]);
end;

function ExtractSeccion;
var
  p : integer;
begin
  p := pos ('+', Cad);
  if p <> 0
    then Result := copy (cad, 1, p - 1)
    else Result := Cad;
end;

function ExtractItem;
var
  p, p2 : integer;

begin
  p := pos ('+', Cad);
  p2 := pos ('=', Cad);
  if p2 = 0
    then p2 := InvertPos ('.', Cad);
  if p2 = 0
    then p2 := length (cad);
  if p <> 0
    then Result := copy (cad, p + 1, p2 - p - 1)
    else Result := Cad;
end;

function ExtractValue;
var
  p, p2 : integer;
begin
  Result := '';
  p := pos ('=', Cad);
  if p = 0
    then exit;
  p2 := InvertPos ('.', Cad);
  if p2 <> 0
    then Result := copy (Cad, p + 1, p2 - p - 1)
    else Result := copy (Cad, p + 1, length (Cad) - p);
end;

function ExtractComponent;
var
  p : integer;
begin
  p := pos ('+', Cad);
  if p <> 0
    then Result := copy (cad, 1, p - 1)
    else Result := Cad;
end;

function ExtractProperty;
var
  p : integer;
begin
  p := pos ('+', Cad);
  if p <> 0
    then Result := copy (cad, p + 1, length (cad) - p)
    else Result := Cad;
end;

function ExtractGuardar;
var
  p, p2 : integer;
begin
  p := InvertPos ('.', Cad);
  p2 := pos ('#', Cad);
  if p2 = 0
    then p2 := length (cad);
  if p <> 0
    then Result := boolean (StrToInt (copy (cad, p + 1, p2 - p - 1)))
    else Result := false;
end;

function ExtractCargar;
var
  p : integer;
begin
  p := pos ('#', Cad);
  if p <> 0
    then Result := boolean (StrToInt (copy (cad, p + 1, length (cad) - p)))
    else Result := false;
end;


end.
