
/////////////////////////////////////////////////
//                                             //
//   QMemory 1.01                              //
//                                             //
//   The alternative quick memory manager      //
//                                             //
//   Copyright (c) 2000, Andrew N. Driazgov    //
//   e-mail: andrey@asp.tstu.ru                //
//                                             //
//   Last updated: November 19, 2000           //
//                                             //
/////////////////////////////////////////////////

unit QMemory;

interface

{ QMemory is a new memory manager. You can use it as a replacement of the
  default system memory manager. To do this simply add QMemory unit in
  your project (as the first unit listed in the project file). Don't forget
  to call the QMemCompact function when your application is idle. This
  subrouting decommits the unused memory blocks and compacts an internal
  table for speeding up the memory manager. All allocated memory blocks are
  32 byte aligned. The minimum size of the block is 32 bytes. As it is
  necessary to store some information with each block a dword is attached
  to the front of each block at -4 the aligned address. Thus, memory requests
  for up to 28 bytes allocate a 32 byte block, requests for 29 to 60 bytes
  allocate a 64 byte block, etc. This idea was adopted from HPMM project
  of Robert Lee (rhlee@optimalcode.com). The memory is committed and
  decommitted in 64K blocks. The maximum amount of the memory is specified
  when QMemInstall function is called (from the initialization section of
  the unit). You can't change this value later. Now QMemory manager is not
  intended for using with DLLs. If you in need of such capability let the
  author know. This memory manager works at the same speed as a default
  memory manager with the blocks up to 4096 bytes and considerably (many
  times) faster with the larger blocks. }

(*
  QMemory should always be the first unit listed in program uses clause:

  program Project1;

  uses
    QMemory in 'QMemory.pas',        // first unit !!!
    Forms,
    MainUnit in 'MainUnit.pas' {MainForm},
    ...

  {$R *.RES}

  begin
    Application.Initialize;
    Application.Title := 'Project1';
    Application.CreateForm(TMainForm, MainForm);
    ...
    Application.Run;
  end.
*)

{ QMemInstall creates a custom heap and sets the entry points of the memory
  manager to the three functions of this unit which work with a created heap.
  The function reserves space in the virtual address space of the process and
  allocates physical storage for a specified initial portion of this block.
  InitialSize specifies the initial size, in bytes, of the heap. This value
  determines the initial amount of physical storage that is allocated for the
  heap. The value is rounded up to the next 64-kilobyte boundary. MaximumSize
  specifies the maximum size, in bytes, of the heap. The QMemInstall function
  rounds MaximumSize up to the next 512-kilobyte boundary, and then reserves
  a block of that size in the process's virtual address space for the heap.
  If allocation requests made by the GetMem or ReallocMem (or New) exceed the
  initial amount of physical storage specified by InitialSize, the system
  allocates additional pages of physical storage for the heap, up to the
  heap's maximum size. If the function succeeds, the return value is 0. If
  the function fails, it returns -1.

  QMemInstall is called from the initialization section of this unit. }

function QMemInstall(InitialSize, MaximumSize: Integer): Integer;

{ QMemRelease destroys the custom heap and restores the previous memory
  manager. QMemRelease decommits and releases all the pages of a heap. If the
  function succeeds, the return value is 0. If the function fails, the return
  value is -1.

  QMemRelease is called from the finalization section of this unit. }

function QMemRelease: Integer;

{ QMemCompact attempts to compact a custom heap. It compacts the heap by
  deleting free blocks of memory of zero length and decommitting large free
  blocks of memory. You may want to call this function from Application.Idle
  event handler. QMemCompact is only way (except QMemRelease) for decommit
  pages of physical storage. If the function succeeds, the return value is 0.
  If the function fails, it returns -1. }

function QMemCompact: Integer;

{ QMemSize returns the size, in bytes, of a memory block allocated from a
  custom heap. P is a pointer to the memory block whose size the function
  will obtain. The custom heap has to be installed. If the function succeeds,
  the return value is the size, in bytes, of the allocated memory block.
  If the function fails, the return value is -1. }

function QMemSize(P: Pointer): Integer;

{ QMemTotalAddrSpace returns the total address space of the custom heap,
  in bytes. This is fixed and will not grow as your program's dynamic memory
  usage grows. TotalUncommitted + TotalCommitted = TotalAddrSpace. The value
  is equal to MaximumSize, which you have specified when called QMemInstall
  function. If QMemTotalAddrSpace function fails, the return value is -1. }

function QMemTotalAddrSpace: Integer;

{ QMemTotalCommitted returns the total number of bytes (of TotalAddrSpace)
  for which space has been allocated in the swap file. If the function fails,
  the return value is -1. }

function QMemTotalCommitted: Integer;

{ QMemTotalUncommitted returns the total number of bytes (of TotalAddrSpace)
  for which space has not been allocated in the swap file. If the function
  fails, the return value is -1. }

function QMemTotalUncommitted: Integer;

{ QMemTotalAllocated returns the total number of bytes dynamically allocated
  by your program. It includes 4 bytes at the beginning of each memory block
  and the trailing bytes for maintain of 32-bytes align of a memory blocks.
  If the function fails, the return value is -1. }

function QMemTotalAllocated: Integer;

{ QMemTotalFree returns the total number of free bytes available in the
  custom heap for allocation by your program. If the function fails, the
  return value is -1. }

function QMemTotalFree: Integer;

{ QMemMaxFreeBlock returns the size, in bytes, of the maximum memory block
  which you can allocate in the custom heap. You can pass this value to
  the GetMem procedure (if you have physical storage of enought size).
  If the function fails, the return value is -1. }

function QMemMaxFreeBlock: Integer;

{ QMemCountOfFreeBlocks returns the total number of free blocks in the custom
  heap address space. If the function fails, the return value is -1. }

function QMemCountOfFreeBlocks: Integer;

{ QMemOverhead returns the total number of bytes required by the heap manager
  to manage all the blocks dynamically allocated by your program. More
  precisely, it returns the total size, in bytes, of additional committed
  space. If the function fails, the return value is -1. }

function QMemOverhead: Integer;

{ QMemGetHeapStatus returns the current status of the custom memory manager
  in a TQMemHeapStatus record. The fields of this record have been described
  above (they are analogues of the corresponding functions). If the function
  fails, all fields are 0. }

type
  TQMemHeapStatus = record
    TotalAddrSpace: Cardinal;
    TotalCommitted: Cardinal;
    TotalUncommitted: Cardinal;
    TotalAllocated: Cardinal;
    TotalFree: Cardinal;
    MaxFreeBlock: Cardinal;
    CountOfFreeBlocks: Cardinal;
    Overhead: Cardinal;
  end;

function QMemGetHeapStatus: TQMemHeapStatus;

{ QMemSetMaxECount assigns a new value to the internal variable which limits
  maximum number of free fragments in the custom heap. The default value is
  1048576. If you want to save some virtual address space you may set this
  variable less (or more, if it is necessary). This function should be called
  before QMemInstall (when the memory manager is not installed yet). If the
  function succeeds, it returns 0. If the function fails, it returns -1. }

function QMemSetMaxECount(Value: Integer): Integer;

implementation

uses Windows;

type
  PEntryPoint = ^TEntryPoint;
  TEntryPoint = packed record
    Address: LongWord;
    Space: LongWord;
  end;

  PEntryArr = ^TEntryArr;
  TEntryArr = array[0..$7FFFFFF] of TEntryPoint;

var
  lpCriticalSection:_RTL_CRITICAL_SECTION;

  MaxECount: Integer = 1048576;

  StartAddr: LongWord;
  SpaceBegin: LongWord;

  ECount: Integer;
  EArr: PEntryArr;

  QMemIsInstalled: Boolean = False;

  OldMemManager: TMemoryManager;

function GetRegionOfSize(P: Pointer; Size, Count: Integer): Pointer;
asm
        PUSH    EBX
        ADD     EAX,4
        MOV     EBX,ECX
        SHR     ECX,3
        JE      @@nx
@@lp:   CMP     EDX,[EAX]
        JLE     @@qt0
        CMP     EDX,[EAX+8]
        JLE     @@qt1
        CMP     EDX,[EAX+16]
        JLE     @@qt2
        CMP     EDX,[EAX+24]
        JLE     @@qt3
        CMP     EDX,[EAX+32]
        JLE     @@qt4
        CMP     EDX,[EAX+40]
        JLE     @@qt5
        CMP     EDX,[EAX+48]
        JLE     @@qt6
        CMP     EDX,[EAX+56]
        JLE     @@qt7
        ADD     EAX,64
        DEC     ECX
        JNE     @@lp
@@nx:   AND     EBX,7
        JMP     DWORD PTR @@tV[EBX*4]
@@qt0:  SUB     EAX,4
        POP     EBX
        RET
@@qt1:  ADD     EAX,4
        POP     EBX
        RET
@@qt2:  ADD     EAX,12
        POP     EBX
        RET
@@qt3:  ADD     EAX,20
        POP     EBX
        RET
@@qt4:  ADD     EAX,28
        POP     EBX
        RET
@@qt5:  ADD     EAX,36
        POP     EBX
        RET
@@qt6:  ADD     EAX,44
        POP     EBX
        RET
@@qt7:  ADD     EAX,52
        POP     EBX
        RET
@@tV:   DD      @@t0,@@t1,@@t2,@@t3
        DD      @@t4,@@t5,@@t6,@@t7
@@t7:   CMP     EDX,[EAX+EBX*8-56]
        JLE     @@xx7
@@t6:   CMP     EDX,[EAX+EBX*8-48]
        JLE     @@xx6
@@t5:   CMP     EDX,[EAX+EBX*8-40]
        JLE     @@xx5
@@t4:   CMP     EDX,[EAX+EBX*8-32]
        JLE     @@xx4
@@t3:   CMP     EDX,[EAX+EBX*8-24]
        JLE     @@xx3
@@t2:   CMP     EDX,[EAX+EBX*8-16]
        JLE     @@xx2
@@t1:   CMP     EDX,[EAX+EBX*8-8]
        JLE     @@xx1
@@t0:   XOR     EAX,EAX
        POP     EBX
        RET
@@xx7:  LEA     EAX,[EAX+EBX*8-60]
        POP     EBX
        RET
@@xx6:  LEA     EAX,[EAX+EBX*8-52]
        POP     EBX
        RET
@@xx5:  LEA     EAX,[EAX+EBX*8-44]
        POP     EBX
        RET
@@xx4:  LEA     EAX,[EAX+EBX*8-36]
        POP     EBX
        RET
@@xx3:  LEA     EAX,[EAX+EBX*8-28]
        POP     EBX
        RET
@@xx2:  LEA     EAX,[EAX+EBX*8-20]
        POP     EBX
        RET
@@xx1:  LEA     EAX,[EAX+EBX*8-12]
        POP     EBX
end;

function IntBitTest(P: Pointer; Index: Integer): Boolean;
asm
        BT      [EAX],EDX
        SETC    AL
end;

procedure IntBitSet(P: Pointer; Index: Integer);
asm
        BTS     [EAX],EDX
end;

function IntFreeBitScanForward(P: Pointer; FirstBit, LastBit: Integer): Integer;
asm
        PUSH    EBX
        PUSH    ESI
        PUSH    EDI
        LEA     ESI,[EDX+8]
        CMP     ECX,ESI
        JL      @@ut
        MOV     EBX,$FFFFFFFF
        MOV     ESI,ECX
        MOV     EDI,$0000001F
        AND     ECX,EDI
        AND     ESI,$FFFFFFE0
        SUB     EDI,ECX
        SHR     ESI,5
        MOV     ECX,EDI
        MOV     EDI,EBX
        SHR     EDI,CL
        MOV     ECX,EDX
        AND     EDX,$FFFFFFE0
        AND     ECX,$0000001F
        SHR     EDX,5
        SHL     EBX,CL
        MOV     ECX,[EAX+EDX*4]
        NOT     ECX
        AND     EBX,ECX
        SUB     ESI,EDX
        JE      @@nq
        TEST    EBX,EBX
        JNE     @@ne
        INC     EDX
        DEC     ESI
        JE      @@xx
@@lp:   MOV     EBX,[EAX+EDX*4]
        NOT     EBX
        TEST    EBX,EBX
        JNE     @@ne
        INC     EDX
        DEC     ESI
        JNE     @@lp
@@xx:   MOV     EBX,[EAX+EDX*4]
        NOT     EBX
@@nq:   AND     EBX,EDI
        JE      @@zq
@@ne:   BSF     ECX,EBX
@@qt:   SHL     EDX,5
        LEA     EAX,[ECX+EDX]
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@ut:   SUB     ECX,EDX
        JS      @@zq
@@uk:   BT      [EAX],EDX
        JNC     @@iq
        INC     EDX
        DEC     ECX
        JNS     @@uk
@@zq:   MOV     EAX,$FFFFFFFF
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@iq:   MOV     EAX,EDX
        POP     EDI
        POP     ESI
        POP     EBX
end;

function IntFreeBitScanReverse(P: Pointer; FirstBit, LastBit: Integer): Integer;
asm
        PUSH    EBX
        PUSH    ESI
        PUSH    EDI
        LEA     ESI,[EDX+8]
        CMP     ECX,ESI
        JL      @@ut
        MOV     EBX,$FFFFFFFF
        MOV     ESI,ECX
        MOV     EDI,$0000001F
        AND     ECX,EDI
        AND     ESI,$FFFFFFE0
        SUB     EDI,ECX
        SHR     ESI,5
        MOV     ECX,EDI
        MOV     EDI,EBX
        SHR     EDI,CL
        MOV     ECX,EDX
        AND     EDX,$FFFFFFE0
        AND     ECX,$0000001F
        SHR     EDX,5
        SHL     EBX,CL
        MOV     ECX,[EAX+ESI*4]
        NOT     ECX
        AND     EDI,ECX
        SUB     EDX,ESI
        JE      @@nq
        TEST    EDI,EDI
        JNE     @@ne
        NEG     EDX
        DEC     ESI
        DEC     EDX
        JE      @@xx
@@lp:   MOV     EDI,[EAX+ESI*4]
        NOT     EDI
        TEST    EDI,EDI
        JNE     @@ne
        DEC     ESI
        DEC     EDX
        JNE     @@lp
@@xx:   MOV     EDI,[EAX+ESI*4]
        NOT     EDI
@@nq:   AND     EDI,EBX
        JE      @@zq
@@ne:   BSR     ECX,EDI
@@qt:   SHL     ESI,5
        LEA     EAX,[ECX+ESI]
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@ut:   SUB     EDX,ECX
        JG      @@zq
@@uk:   BT      [EAX],ECX
        JNC     @@iq
        DEC     ECX
        INC     EDX
        JNG     @@uk
@@zq:   MOV     EAX,$FFFFFFFF
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@iq:   MOV     EAX,ECX
        POP     EDI
        POP     ESI
        POP     EBX
end;

procedure IntSetBits(P: Pointer; FirstBit, LastBit: Integer);
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        LEA     ESI,[EDX+8]
        CMP     ECX,ESI
        JL      @@ut
        MOV     EBX,$FFFFFFFF
        MOV     ESI,ECX
        MOV     EDI,$0000001F
        AND     ECX,EDI
        AND     ESI,$FFFFFFE0
        SUB     EDI,ECX
        SHR     ESI,5
        MOV     ECX,EDI
        MOV     EDI,EBX
        SHR     EDI,CL
        MOV     ECX,EDX
        AND     EDX,$FFFFFFE0
        AND     ECX,$0000001F
        SHR     EDX,5
        SHL     EBX,CL
        SUB     ESI,EDX
        JE      @@xx
        OR      [EAX+EDX*4],EBX
        INC     EDX
        DEC     ESI
        JE      @@ne
        MOV     EBX,$FFFFFFFF
@@lp:   MOV     [EAX+EDX*4],EBX
        INC     EDX
        DEC     ESI
        JNE     @@lp
@@xx:   AND     EDI,EBX
@@ne:   OR      [EAX+EDX*4],EDI
        POP     EBX
        POP     ESI
        POP     EDI
        RET
@@ut:   SUB     ECX,EDX
        JS      @@qt
@@uk:   BTS     [EAX],EDX
        INC     EDX
        DEC     ECX
        JNS     @@uk
@@qt:   POP     EBX
        POP     ESI
        POP     EDI
end;

function IntGetMem(Size: Integer): Pointer;
label
  99;
var
  P: PEntryPoint;
  I,J: Integer;
begin
  try
    if IsMultiThread then
      EnterCriticalSection(lpCriticalSection);
    Inc(Size,35);
    Size := Size and $FFFFFFE0;
    P := GetRegionOfSize(EArr,Size,ECount);
    if P <> nil then
    begin
      J := P^.Address-StartAddr;
      I := J;
      Inc(J,Size);
      I := LongWord(I) shr 16;
      Dec(J);
      J := LongWord(J) shr 16;
      if I = J then
      begin
        if not IntBitTest(Pointer(StartAddr),I) then
        begin
          if VirtualAlloc(Pointer(StartAddr+LongWord(I) shl 16), $10000,
            MEM_COMMIT, PAGE_READWRITE) = nil then
          begin
            Result := nil;
            goto 99;
          end;
          IntBitSet(Pointer(StartAddr),I);
        end;
      end else
      begin
        I := IntFreeBitScanForward(Pointer(StartAddr),I,J);
        if I >= 0 then
        begin
          J := IntFreeBitScanReverse(Pointer(StartAddr),I,J);
          if VirtualAlloc(Pointer(StartAddr+LongWord(I) shl 16),
            LongWord(J-I+1) shl 16, MEM_COMMIT, PAGE_READWRITE) = nil then
          begin
            Result := nil;
            goto 99;
          end;
          IntSetBits(Pointer(StartAddr),I,J);
        end;
      end;
      with P^ do
      begin
        Result := Pointer(Address+4);
        PInteger(Address)^ := Size;
        Inc(Address,Size);
        Dec(Space,Size);
      end;
    end else
      Result := nil;
  99:
  finally
    if IsMultiThread then
      LeaveCriticalSection(lpCriticalSection);
  end;
end;

function SearchPointerPlace(P, EArr: Pointer; ECount: Integer): Integer;
asm
        PUSH    EDI
        PUSH    ESI
        MOV     EDI,$FFFFFFFF
        XOR     ESI,ESI
        PUSH    EBX
        DEC     ECX
        JS      @@qt
@@lp:   LEA     EBX,[ESI+ECX]
        SHR     EBX,1
        CMP     EAX,[EDX+EBX*8]
        JB      @@bb
        MOV     EDI,EBX
        LEA     ESI,[EBX+1]
        CMP     ECX,ESI
        JGE     @@lp
@@qt:   POP     EBX
        MOV     EAX,EDI
        POP     ESI
        POP     EDI
        RET
@@bb:   LEA     ECX,[EBX-1]
        CMP     ESI,ECX
        JLE     @@lp
        POP     EBX
        MOV     EAX,EDI
        POP     ESI
        POP     EDI
end;

procedure ShiftListUp(P: Pointer; FromIndex, Count: Integer);
asm
        PUSH    EBX
        NOP
        LEA     EAX,[EAX+ECX*8-8]
        SUB     ECX,EDX
        DEC     ECX
        MOV     EDX,ECX
        AND     ECX,7
        SHR     EDX,3
        JE      @@nx
@@lp:   MOV     EBX,[EAX-4]
        MOV     [EAX+4],EBX
        MOV     EBX,[EAX-8]
        MOV     [EAX],EBX
        MOV     EBX,[EAX-12]
        MOV     [EAX-4],EBX
        MOV     EBX,[EAX-16]
        MOV     [EAX-8],EBX
        MOV     EBX,[EAX-20]
        MOV     [EAX-12],EBX
        MOV     EBX,[EAX-24]
        MOV     [EAX-16],EBX
        MOV     EBX,[EAX-28]
        MOV     [EAX-20],EBX
        MOV     EBX,[EAX-32]
        MOV     [EAX-24],EBX
        MOV     EBX,[EAX-36]
        MOV     [EAX-28],EBX
        MOV     EBX,[EAX-40]
        MOV     [EAX-32],EBX
        MOV     EBX,[EAX-44]
        MOV     [EAX-36],EBX
        MOV     EBX,[EAX-48]
        MOV     [EAX-40],EBX
        MOV     EBX,[EAX-52]
        MOV     [EAX-44],EBX
        MOV     EBX,[EAX-56]
        MOV     [EAX-48],EBX
        MOV     EBX,[EAX-60]
        MOV     [EAX-52],EBX
        MOV     EBX,[EAX-64]
        MOV     [EAX-56],EBX
        SUB     EAX,64
        DEC     EDX
        JNE     @@lp
@@nx:   NEG     ECX
        JMP     DWORD PTR @@tV[ECX*4+28]
@@tV:   DD      @@t7,@@t6,@@t5,@@t4
        DD      @@t3,@@t2,@@t1,@@t0
@@t7:   MOV     EBX,[EAX+ECX*8+52]
        MOV     [EAX+ECX*8+60],EBX
        MOV     EBX,[EAX+ECX*8+48]
        MOV     [EAX+ECX*8+56],EBX
@@t6:   MOV     EBX,[EAX+ECX*8+44]
        MOV     [EAX+ECX*8+52],EBX
        MOV     EBX,[EAX+ECX*8+40]
        MOV     [EAX+ECX*8+48],EBX
@@t5:   MOV     EBX,[EAX+ECX*8+36]
        MOV     [EAX+ECX*8+44],EBX
        MOV     EBX,[EAX+ECX*8+32]
        MOV     [EAX+ECX*8+40],EBX
@@t4:   MOV     EBX,[EAX+ECX*8+28]
        MOV     [EAX+ECX*8+36],EBX
        MOV     EBX,[EAX+ECX*8+24]
        MOV     [EAX+ECX*8+32],EBX
@@t3:   MOV     EBX,[EAX+ECX*8+20]
        MOV     [EAX+ECX*8+28],EBX
        MOV     EBX,[EAX+ECX*8+16]
        MOV     [EAX+ECX*8+24],EBX
@@t2:   MOV     EBX,[EAX+ECX*8+12]
        MOV     [EAX+ECX*8+20],EBX
        MOV     EBX,[EAX+ECX*8+8]
        MOV     [EAX+ECX*8+16],EBX
@@t1:   MOV     EBX,[EAX+ECX*8+4]
        MOV     [EAX+ECX*8+12],EBX
        MOV     EBX,[EAX+ECX*8]
        MOV     [EAX+ECX*8+8],EBX
@@t0:   POP     EBX
end;

procedure ShiftListDown(P: Pointer; ToIndex, Count: Integer);
asm
        PUSH    EBX
        SUB     ECX,EDX
        LEA     EAX,[EAX+EDX*8]
        DEC     ECX
        MOV     EDX,ECX
        AND     ECX,7
        SHR     EDX,3
        JE      @@nx
@@lp:   MOV     EBX,[EAX+8]
        MOV     [EAX],EBX
        MOV     EBX,[EAX+12]
        MOV     [EAX+4],EBX
        MOV     EBX,[EAX+16]
        MOV     [EAX+8],EBX
        MOV     EBX,[EAX+20]
        MOV     [EAX+12],EBX
        MOV     EBX,[EAX+24]
        MOV     [EAX+16],EBX
        MOV     EBX,[EAX+28]
        MOV     [EAX+20],EBX
        MOV     EBX,[EAX+32]
        MOV     [EAX+24],EBX
        MOV     EBX,[EAX+36]
        MOV     [EAX+28],EBX
        MOV     EBX,[EAX+40]
        MOV     [EAX+32],EBX
        MOV     EBX,[EAX+44]
        MOV     [EAX+36],EBX
        MOV     EBX,[EAX+48]
        MOV     [EAX+40],EBX
        MOV     EBX,[EAX+52]
        MOV     [EAX+44],EBX
        MOV     EBX,[EAX+56]
        MOV     [EAX+48],EBX
        MOV     EBX,[EAX+60]
        MOV     [EAX+52],EBX
        MOV     EBX,[EAX+64]
        MOV     [EAX+56],EBX
        MOV     EBX,[EAX+68]
        MOV     [EAX+60],EBX
        ADD     EAX,64
        DEC     EDX
        JNE     @@lp
@@nx:   JMP     DWORD PTR @@tV[ECX*4]
@@tV:   DD      @@t0,@@t1,@@t2,@@t3
        DD      @@t4,@@t5,@@t6,@@t7
@@t7:   MOV     EBX,[EAX+ECX*8-48]
        MOV     [EAX+ECX*8-56],EBX
        MOV     EBX,[EAX+ECX*8-44]
        MOV     [EAX+ECX*8-52],EBX
@@t6:   MOV     EBX,[EAX+ECX*8-40]
        MOV     [EAX+ECX*8-48],EBX
        MOV     EBX,[EAX+ECX*8-36]
        MOV     [EAX+ECX*8-44],EBX
@@t5:   MOV     EBX,[EAX+ECX*8-32]
        MOV     [EAX+ECX*8-40],EBX
        MOV     EBX,[EAX+ECX*8-28]
        MOV     [EAX+ECX*8-36],EBX
@@t4:   MOV     EBX,[EAX+ECX*8-24]
        MOV     [EAX+ECX*8-32],EBX
        MOV     EBX,[EAX+ECX*8-20]
        MOV     [EAX+ECX*8-28],EBX
@@t3:   MOV     EBX,[EAX+ECX*8-16]
        MOV     [EAX+ECX*8-24],EBX
        MOV     EBX,[EAX+ECX*8-12]
        MOV     [EAX+ECX*8-20],EBX
@@t2:   MOV     EBX,[EAX+ECX*8-8]
        MOV     [EAX+ECX*8-16],EBX
        MOV     EBX,[EAX+ECX*8-4]
        MOV     [EAX+ECX*8-12],EBX
@@t1:   MOV     EBX,[EAX+ECX*8]
        MOV     [EAX+ECX*8-8],EBX
        MOV     EBX,[EAX+ECX*8+4]
        MOV     [EAX+ECX*8-4],EBX
@@t0:   POP     EBX
end;

function IntFreeMem(P: Pointer): Integer;
label
  99;
var
  I: Integer;
  J: LongWord;
begin
  Dec(LongWord(P),4);
  if (LongWord(P)<SpaceBegin) or
    (PLongWord(P)^+LongWord(P)>LongWord(EArr)) then
  begin
    Inc(LongWord(P),4);
    Result := OldMemManager.FreeMem(P);
    Exit;
  end;
  try
    if IsMultiThread then
      EnterCriticalSection(lpCriticalSection);
    I := SearchPointerPlace(P,EArr,ECount);
    if I >= 0 then
    begin
      with EArr^[I] do
        J := Address+Space;
      if LongWord(P) <= J then
      begin
        if LongWord(P) = J then
        begin
          J := PLongWord(P)^;
          if I < ECount-1 then
          begin
            Inc(LongWord(P),J);
            if LongWord(P) >= EArr^[I+1].Address then
            begin
              if LongWord(P) = EArr^[I+1].Address then
              begin
                Inc(J,EArr^[I].Space);
                Inc(J,EArr^[I+1].Space);
                EArr^[I].Space := J;
                ShiftListDown(EArr,I+1,ECount);
                Dec(ECount);
                Result := 0;
              end else
                Result := -1;
              goto 99;
            end;
          end;
          Inc(EArr^[I].Space,J);
          Result := 0;
        end else
          Result := -1;
        goto 99;
      end;
    end;
    Inc(I);
    if I < ECount then
    begin
      J := PLongWord(P)^;
      Inc(J,LongWord(P));
      with EArr^[I] do
        if J >= Address then
        begin
          if J = Address then
          begin
            Address := LongWord(P);
            Inc(Space,PLongWord(P)^);
            Result := 0;
          end else
            Result := -1;
          goto 99;
        end;
    end;
    if ECount < MaxECount then
    begin
      J := LongWord(@EArr^[ECount]);
      Dec(J,StartAddr);
      J := J shr 16;
      if not IntBitTest(Pointer(StartAddr),J) then
      begin
        if VirtualAlloc(Pointer(StartAddr+J shl 16), $10000, MEM_COMMIT,
          PAGE_READWRITE) = nil then
        begin
          Result := -1;
          goto 99;
        end;
        IntBitSet(Pointer(StartAddr),J);
      end;
      Inc(ECount);
      ShiftListUp(EArr,I,ECount);
      with EArr^[I] do
      begin
        Address := LongWord(P);
        Space := PLongWord(P)^;
      end;
      Result := 0;
    end else
      Result := -1;
  99:
  finally
    if IsMultiThread then
      LeaveCriticalSection(lpCriticalSection);
  end;
end;

procedure IntCopyMem(Source, Dest: Pointer; L: Cardinal);
asm
        PUSH    EBX
        SUB     EDX,4
        SHR     ECX,5
        JMP     @@fs
@@lp:   MOV     EBX,[EAX]
        MOV     [EDX],EBX
@@fs:   MOV     EBX,[EAX+4]
        MOV     [EDX+4],EBX
        MOV     EBX,[EAX+8]
        MOV     [EDX+8],EBX
        MOV     EBX,[EAX+12]
        MOV     [EDX+12],EBX
        MOV     EBX,[EAX+16]
        MOV     [EDX+16],EBX
        MOV     EBX,[EAX+20]
        MOV     [EDX+20],EBX
        MOV     EBX,[EAX+24]
        MOV     [EDX+24],EBX
        MOV     EBX,[EAX+28]
        MOV     [EDX+28],EBX
        ADD     EAX,32
        ADD     EDX,32
        DEC     ECX
        JNE     @@lp
@@qt:   POP     EBX
end;

function IntReallocMem(P: Pointer; Size: Integer): Pointer;
label
  99;
var
  I,J,K: Integer;
begin
  Dec(LongWord(P),4);
  if (LongWord(P)<SpaceBegin) or
    (PLongWord(P)^+LongWord(P)>LongWord(EArr)) then
  begin
    Inc(LongWord(P),4);
    Result := OldMemManager.ReallocMem(P,Size);
    Exit;
  end;
  Inc(Size,35);
  Size := Size and $FFFFFFE0;
  J := PInteger(P)^;
  if Size = J then
  begin
    Result := Pointer(LongWord(P)+4);
    Exit;
  end;
  try
    if IsMultiThread then
      EnterCriticalSection(lpCriticalSection);
    I := SearchPointerPlace(P,EArr,ECount);
    if I >= 0 then
      with EArr^[I] do
        if LongWord(P) < Address+Space then
        begin
          Result := nil;
          goto 99;
        end;
    Inc(I);
    if I < ECount then
      with EArr^[I] do
        if LongWord(J)+LongWord(P) >= Address then
        begin
          if LongWord(J)+LongWord(P) = Address then
          begin
            if Size <= J+Integer(Space) then
            begin
              J := (Address-StartAddr) shr 16;
              K := (LongWord(P)+LongWord(Size)-StartAddr-1) shr 16;
              if J = K then
              begin
                if not IntBitTest(Pointer(StartAddr),J) then
                begin
                  if VirtualAlloc(Pointer(StartAddr+LongWord(J) shl 16), $10000,
                    MEM_COMMIT, PAGE_READWRITE) = nil then
                  begin
                    Result := nil;
                    goto 99;
                  end;
                  IntBitSet(Pointer(StartAddr),J);
                end;
              end else
              begin
                J := IntFreeBitScanForward(Pointer(StartAddr),J,K);
                if J >= 0 then
                begin
                  K := IntFreeBitScanReverse(Pointer(StartAddr),J,K);
                  if VirtualAlloc(Pointer(StartAddr+LongWord(J) shl 16),
                    LongWord(K-J+1) shl 16, MEM_COMMIT, PAGE_READWRITE) = nil then
                  begin
                    Result := nil;
                    goto 99;
                  end;
                  IntSetBits(Pointer(StartAddr),J,K);
                end;
              end;
              Address := LongWord(P)+LongWord(Size);
              Inc(Space,PLongWord(P)^-LongWord(Size));
              PInteger(P)^ := Size;
              Result := Pointer(LongWord(P)+4);
            end else
            begin
              Result := IntGetMem(Size-4);
              if Result <> nil then
              begin
                IntCopyMem(P,Result,PLongWord(P)^);
                Inc(LongWord(P),4);
                IntFreeMem(P);
              end;
            end;
          end else
            Result := nil;
          goto 99;
        end;
    if Size > J then
    begin
      Result := IntGetMem(Size-4);
      if Result <> nil then
      begin
        IntCopyMem(P,Result,J);
        Inc(LongWord(P),4);
        IntFreeMem(P);
      end;
    end
    else if ECount < MaxECount then
    begin
      J := LongWord(@EArr^[ECount]);
      Dec(J,StartAddr);
      J := LongWord(J) shr 16;
      if not IntBitTest(Pointer(StartAddr),J) then
      begin
        if VirtualAlloc(Pointer(StartAddr+LongWord(J) shl 16), $10000, MEM_COMMIT,
          PAGE_READWRITE) = nil then
        begin
          Result := nil;
          goto 99;
        end;
        IntBitSet(Pointer(StartAddr),J);
      end;
      Inc(ECount);
      ShiftListUp(EArr,I,ECount);
      with EArr^[I] do
      begin
        Address := LongWord(P)+LongWord(Size);
        Space := PLongWord(P)^-LongWord(Size);
      end;
      PInteger(P)^ := Size;
      Result := Pointer(LongWord(P)+4);
    end else
      Result := nil;
  99:
  finally
    if IsMultiThread then
      LeaveCriticalSection(lpCriticalSection);
  end;
end;

function RemZeroEntries(P: Pointer; Count: Integer): Cardinal;
asm
        PUSH    EDI
	PUSH	EAX
        DEC     EDX
        JS      @@qt
        MOV     ECX,$FFFFFFFF
@@lp1:  TEST    ECX,[EAX+4]
        JE      @@nx
        ADD     EAX,8
        DEC     EDX
        JNS     @@lp1
        JMP     @@qt
@@nx:   LEA     EDI,[EAX+8]
        DEC     EDX
        JS      @@qt
@@lp2:  MOV     ECX,[EDI+4]
        TEST    ECX,ECX
        JE      @@sk
        MOV     [EAX+4],ECX
        MOV     ECX,[EDI]
        MOV     [EAX],ECX
        ADD     EAX,8
@@sk:   ADD     EDI,8
        DEC     EDX
        JNS     @@lp2
@@qt:   POP	ECX
	SUB	EAX,ECX
	SHR	EAX,3
        POP     EDI
end;

function IntSetBitScanForward(P: Pointer; FirstBit, LastBit: Integer): Integer;
asm
        PUSH    EBX
        PUSH    ESI
        PUSH    EDI
        LEA     ESI,[EDX+8]
        CMP     ECX,ESI
        JL      @@ut
        MOV     EBX,$FFFFFFFF
        MOV     ESI,ECX
        MOV     EDI,$0000001F
        AND     ECX,EDI
        AND     ESI,$FFFFFFE0
        SUB     EDI,ECX
        SHR     ESI,5
        MOV     ECX,EDI
        MOV     EDI,EBX
        SHR     EDI,CL
        MOV     ECX,EDX
        AND     EDX,$FFFFFFE0
        AND     ECX,$0000001F
        SHR     EDX,5
        SHL     EBX,CL
        AND     EBX,[EAX+EDX*4]
        SUB     ESI,EDX
        JE      @@nq
        TEST    EBX,EBX
        JNE     @@ne
        INC     EDX
        DEC     ESI
        JE      @@xx
@@lp:   OR      EBX,[EAX+EDX*4]
        JNE     @@ne
        INC     EDX
        DEC     ESI
        JNE     @@lp
@@xx:   MOV     EBX,[EAX+EDX*4]
@@nq:   AND     EBX,EDI
        JE      @@zq
@@ne:   BSF     ECX,EBX
@@qt:   SHL     EDX,5
        LEA     EAX,[ECX+EDX]
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@ut:   SUB     ECX,EDX
        JS      @@zq
@@uk:   BT      [EAX],EDX
        JC      @@iq
        INC     EDX
        DEC     ECX
        JNS     @@uk
@@zq:   MOV     EAX,$FFFFFFFF
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@iq:   MOV     EAX,EDX
        POP     EDI
        POP     ESI
        POP     EBX
end;

function IntSetBitScanReverse(P: Pointer; FirstBit, LastBit: Integer): Integer;
asm
        PUSH    EBX
        PUSH    ESI
        PUSH    EDI
        LEA     ESI,[EDX+8]
        CMP     ECX,ESI
        JL      @@ut
        MOV     EBX,$FFFFFFFF
        MOV     ESI,ECX
        MOV     EDI,$0000001F
        AND     ECX,EDI
        AND     ESI,$FFFFFFE0
        SUB     EDI,ECX
        SHR     ESI,5
        MOV     ECX,EDI
        MOV     EDI,EBX
        SHR     EDI,CL
        MOV     ECX,EDX
        AND     EDX,$FFFFFFE0
        AND     ECX,$0000001F
        SHR     EDX,5
        SHL     EBX,CL
        AND     EDI,[EAX+ESI*4]
        SUB     EDX,ESI
        JE      @@nq
        TEST    EDI,EDI
        JNE     @@ne
        NEG     EDX
        DEC     ESI
        DEC     EDX
        JE      @@xx
@@lp:   OR      EDI,[EAX+ESI*4]
        JNE     @@ne
        DEC     ESI
        DEC     EDX
        JNE     @@lp
@@xx:   MOV     EDI,[EAX+ESI*4]
@@nq:   AND     EDI,EBX
        JE      @@zq
@@ne:   BSR     ECX,EDI
@@qt:   SHL     ESI,5
        LEA     EAX,[ECX+ESI]
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@ut:   SUB     EDX,ECX
        JG      @@zq
@@uk:   BT      [EAX],ECX
        JC      @@iq
        DEC     ECX
        INC     EDX
        JNG     @@uk
@@zq:   MOV     EAX,$FFFFFFFF
        POP     EDI
        POP     ESI
        POP     EBX
        RET
@@iq:   MOV     EAX,ECX
        POP     EDI
        POP     ESI
        POP     EBX
end;

procedure IntResetBits(P: Pointer; FirstBit, LastBit: Integer);
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        LEA     ESI,[EDX+8]
        CMP     ECX,ESI
        JL      @@ut
        MOV     EBX,$FFFFFFFF
        MOV     ESI,ECX
        MOV     EDI,$0000001F
        AND     ECX,EDI
        AND     ESI,$FFFFFFE0
        SUB     EDI,ECX
        SHR     ESI,5
        MOV     ECX,EDI
        MOV     EDI,EBX
        SHR     EDI,CL
        MOV     ECX,EDX
        AND     EDX,$FFFFFFE0
        AND     ECX,$0000001F
        SHR     EDX,5
        SHL     EBX,CL
        NOT     EDI
        NOT     EBX
        SUB     ESI,EDX
        JE      @@xx
        AND     [EAX+EDX*4],EBX
        INC     EDX
        DEC     ESI
        JE      @@ne
        XOR     EBX,EBX
@@lp:   MOV     [EAX+EDX*4],EBX
        INC     EDX
        DEC     ESI
        JNE     @@lp
@@xx:   OR      EDI,EBX
@@ne:   AND     [EAX+EDX*4],EDI
        POP     EBX
        POP     ESI
        POP     EDI
        RET
@@ut:   SUB     ECX,EDX
        JS      @@qt
@@uk:   BTR     [EAX],EDX
        INC     EDX
        DEC     ECX
        JNS     @@uk
@@qt:   POP     EBX
        POP     ESI
        POP     EDI
end;

function QMemCompact: Integer;
label
  99;
var
  I,L,R: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      ECount := RemZeroEntries(EArr,ECount);
      for I := 0 to ECount-1 do
        with EArr^[I] do
        begin
          if Space shr 17 > 0 then
          begin
            L := (Address+$FFFF-StartAddr) shr 16;
            R := (Address+Space-$10000-StartAddr) shr 16;
            L := IntSetBitScanForward(Pointer(StartAddr),L,R);
            if L >= 0 then
            begin
              R := IntSetBitScanReverse(Pointer(StartAddr),L,R);
              if not VirtualFree(Pointer(StartAddr+LongWord(L) shl 16),
                LongWord(R-L+1) shl 16, MEM_DECOMMIT) then
              begin
                Result := -1;
                goto 99;
              end;
              IntResetBits(Pointer(StartAddr),L,R);
            end;
          end;
        end;
      Result := 0;
    99:
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function QMemSize(P: Pointer): Integer;
begin
  Dec(LongWord(P),4);
  if (LongWord(P)>=SpaceBegin) and
      (PLongWord(P)^+LongWord(P)<=LongWord(EArr)) then
    Result := PInteger(P)^-4
  else
    Result := -1;
end;

function QMemTotalAddrSpace: Integer;
begin
  if QMemIsInstalled then
    Result := LongWord(EArr)-StartAddr
  else
    Result := -1;
end;

const
  BitTable: array[0..255] of Byte =
    (0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
     1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
     1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
     2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
     1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
     2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
     2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
     3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8);

function IntCountOfSetBits(P: Pointer; L: Cardinal): Cardinal;
asm
        PUSH    EBX
        PUSH    ESI
        MOV     EBX,EAX
        XOR     EAX,EAX
        SUB     EDX,2
        JS      @@nx
@@lp:   MOVZX   ECX,BYTE PTR [EBX+EDX]
        MOVZX   ESI,BYTE PTR [EBX+EDX+1]
        MOVZX   ECX,BYTE PTR [ECX+BitTable]
        ADD     EAX,ECX
        MOVZX   ESI,BYTE PTR [ESI+BitTable]
        ADD     EAX,ESI
        SUB     EDX,2
        JNS     @@lp
@@nx:   INC     EDX
        JZ      @@qt2
        POP     ESI
        POP     EBX
        RET
@@qt2:  MOVZX   ECX,BYTE PTR [EBX]
        MOVZX   ECX,BYTE PTR [ECX+BitTable]
        ADD     EAX,ECX
        POP     ESI
        POP     EBX
end;

function IntCountOfFreeBits(P: Pointer; L: Cardinal): Cardinal;
asm
        PUSH    EDX
        CALL    IntCountOfSetBits
        NEG     EAX
        POP     EDX
        LEA     EAX,[EAX+EDX*8]
end;

function QMemTotalCommitted: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := IntCountOfSetBits(Pointer(StartAddr),
        (LongWord(EArr)-StartAddr) shr 19) shl 16;
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function QMemTotalUncommitted: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := IntCountOfFreeBits(Pointer(StartAddr),
        (LongWord(EArr)-StartAddr) shr 19) shl 16;
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function IntGetTotalFree(P: Pointer; Count: Integer): Cardinal;
asm
        PUSH    EBX
        MOV     ECX,EDX
        LEA     EBX,[EAX+4]
        XOR     EAX,EAX
        SHR     EDX,3
        JE      @@nx
@@lp:   ADD     EAX,[EBX]
        ADD     EAX,[EBX+8]
        ADD     EAX,[EBX+16]
        ADD     EAX,[EBX+24]
        ADD     EAX,[EBX+32]
        ADD     EAX,[EBX+40]
        ADD     EAX,[EBX+48]
        ADD     EAX,[EBX+56]
        ADD     EBX,64
        DEC     EDX
        JNE     @@lp
@@nx:   AND     ECX,7
        JMP     DWORD PTR @@tV[ECX*4]
@@tV:   DD      @@qt,@@t1,@@t2,@@t3
        DD      @@t4,@@t5,@@t6,@@t7
@@t7:   ADD     EAX,[EBX+48]
@@t6:   ADD     EAX,[EBX+40]
@@t5:   ADD     EAX,[EBX+32]
@@t4:   ADD     EAX,[EBX+24]
@@t3:   ADD     EAX,[EBX+16]
@@t2:   ADD     EAX,[EBX+8]
@@t1:   ADD     EAX,[EBX]
@@qt:   POP     EBX
end;

function QMemTotalAllocated: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := LongWord(EArr)-SpaceBegin-IntGetTotalFree(EArr,ECount);
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function QMemTotalFree: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := IntGetTotalFree(EArr,ECount);
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function IntGetMaxFreeBlock(P: Pointer; Count: Integer): Integer;
asm
        ADD     EAX,4
        DEC     EDX
        JS      @@zq
        MOV     ECX,[EAX+EDX*8]
        DEC     EDX
        JS      @@qt
@@lp:   CMP     ECX,[EAX+EDX*8]
        JL      @@nw
        DEC     EDX
        JS      @@qt
        CMP     ECX,[EAX+EDX*8]
        JL      @@nw
        DEC     EDX
        JS      @@qt
        CMP     ECX,[EAX+EDX*8]
        JL      @@nw
        DEC     EDX
        JS      @@qt
        CMP     ECX,[EAX+EDX*8]
        JL      @@nw
        DEC     EDX
        JNS     @@lp
@@qt:   LEA     EAX,[ECX-4]
        RET
@@zq:   XOR     EAX,EAX
        RET
@@nw:   MOV     ECX,[EAX+EDX*8]
        DEC     EDX
        JNS     @@lp
        LEA     EAX,[ECX-4]
end;

function QMemMaxFreeBlock: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := IntGetMaxFreeBlock(EArr,ECount);
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function QMemCountOfFreeBlocks: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := ECount;
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function QMemOverhead: Integer;
begin
  if QMemIsInstalled then
  begin
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      Result := IntCountOfSetBits(Pointer(StartAddr+(LongWord(EArr)-
        StartAddr) shr 19),(MaxECount+$E000) shr 16) shl 16;
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end;
  end else
    Result := -1;
end;

function QMemGetHeapStatus: TQMemHeapStatus;
begin
  if QMemIsInstalled then
    try
      if IsMultiThread then
        EnterCriticalSection(lpCriticalSection);
      with Result do
      begin
        TotalAddrSpace := LongWord(EArr)-StartAddr;
        TotalCommitted := IntCountOfSetBits(Pointer(StartAddr),
          TotalAddrSpace shr 19) shl 16;
        TotalUncommitted := TotalAddrSpace - TotalCommitted;
        TotalFree := IntGetTotalFree(EArr,ECount);
        TotalAllocated := LongWord(EArr)-SpaceBegin-TotalFree;
        MaxFreeBlock := IntGetMaxFreeBlock(EArr,ECount);
        CountOfFreeBlocks := ECount;
        Overhead := IntCountOfSetBits(Pointer(StartAddr+TotalAddrSpace shr 19),
          (MaxECount+$E000) shr 16) shl 16;
      end;
    finally
      if IsMultiThread then
        LeaveCriticalSection(lpCriticalSection);
    end
  else
    with Result do
    begin
      TotalAddrSpace := 0;
      TotalCommitted := 0;
      TotalUncommitted := 0;
      TotalAllocated := 0;
      TotalFree := 0;
      MaxFreeBlock := 0;
      CountOfFreeBlocks := 0;
      Overhead := 0;
    end;
end;

function QMemSetMaxECount(Value: Integer): Integer;
begin
  if (not QMemIsInstalled) and (Value>0) then
  begin
    MaxECount := (Value+$1FFF) and $FFFFE000;
    Result := 0;
  end else
    Result := -1;
end;

procedure IntFillLong(Value: LongWord; P: Pointer; Count: Cardinal);
asm
        XCHG    EDI,EDX
        REP     STOSD
        MOV     EDI,EDX
end;

const
  QMemManager: TMemoryManager = (
    GetMem: IntGetMem;
    FreeMem: IntFreeMem;
    ReallocMem: IntReallocMem);

function QMemInstall(InitialSize, MaximumSize: Integer): Integer;
var
  L: LongWord;
begin
  if (not QMemIsInstalled) and (MaximumSize>0) then
  begin
    Inc(MaximumSize,$7FFFF);
    MaximumSize := MaximumSize and $FFF80000;
    L := MaxECount shl 3;
    Inc(L,MaximumSize);
    StartAddr := LongWord(VirtualAlloc(nil, L, MEM_RESERVE or MEM_TOP_DOWN,
      PAGE_READWRITE));
    if StartAddr = 0 then
    begin
      Result := -1;
      Exit;
    end;
    if InitialSize < $10000 then
      InitialSize := $10000
    else if InitialSize > MaximumSize then
      InitialSize := MaximumSize
    else
    begin
      Inc(InitialSize,$FFFF);
      InitialSize := InitialSize and $FFFF0000;
    end;
    if VirtualAlloc(Pointer(StartAddr), InitialSize, MEM_COMMIT,
      PAGE_READWRITE) = nil then
    begin
      VirtualFree(Pointer(StartAddr),0,MEM_RELEASE);
      Result := -1;
      Exit;
    end;
    L := (L+$70000) shr 19;
    IntFillLong(0,Pointer(StartAddr),(L+3) shr 2);
    IntSetBits(Pointer(StartAddr),0,InitialSize shr 16 - 1);
    EArr := Pointer(StartAddr+LongWord(MaximumSize));
    if VirtualAlloc(EArr, $10000, MEM_COMMIT, PAGE_READWRITE) = nil then
    begin
      VirtualFree(Pointer(StartAddr),0,MEM_RELEASE);
      Result := -1;
      Exit;
    end;
    IntBitSet(Pointer(StartAddr),LongWord(MaximumSize) shr 16);
    SpaceBegin := (L+35) and $FFFFFFE0 - 4;
    Inc(SpaceBegin,StartAddr);
    ECount := 1;
    with EArr^[0] do
    begin
      Address := SpaceBegin;
      Space := LongWord(EArr)-SpaceBegin;
    end;
    InitializeCriticalSection(lpCriticalSection);
    GetMemoryManager(OldMemManager);
    QMemIsInstalled := True;
    SetMemoryManager(QMemManager);
    Result := 0;
  end else
    Result := -1;
end;

function QMemRelease: Integer;
begin
  if QMemIsInstalled then
  begin
    SetMemoryManager(OldMemManager);
    QMemIsInstalled := False;
    DeleteCriticalSection(lpCriticalSection);
    if VirtualFree(Pointer(StartAddr), 0, MEM_RELEASE) then
    begin
      Result := 0;
      Exit;
    end;
  end;
  Result := -1;
end;

initialization
  QMemInstall(65536,268435456);

finalization
  QMemRelease;

end.

