{ Copyright Ralf Junker <ralfjunker@gmx.de> 2000 
  http://www.zeitungsjunge.de/delphi/ }

unit rjBase;

interface

{$I rj.inc}

uses Windows, rjTypes;

type
 TDynBuffer = class
 private
  FBuffer: Pointer;
  FBufferSize: Integer;
  function GetAsAnsiString: AnsiString;
 public
  Position: Integer;
  destructor Destroy; override;
  property AsPointer: Pointer read FBuffer;
  procedure AddAnsiChar (const c: AnsiChar);
  property AsAnsiString: AnsiString read GetAsAnsiString;
  procedure Clear;
 end;
 
procedure BufferInit (var Buf: TBuffer);
procedure BufferAdd (const Source: Pointer; const SourceLength: Integer; var Dest: TBuffer);
procedure BufferFree (var Buf: TBuffer);

function LeftMostBit (const Value: Integer): Integer;
{ Returns the number of the highest bit which is set in the passed in value.
  Note that bits are numbered right to left starting with 0.
  For example, LeftMostBit (5) returns 2 (5 is 00000101 in binary).
  Returns the number of the highest bit set, or -1 if no bits are set. }

function RightMostBit (const Value: Integer): Integer;
{ Returns the number of the lowest bit which is set in the passed in value.
  Note that bits are numbered right to left starting with 0.
  For example, RightMostBit (9) returns 2 (9 is 00001100 in binary).
  Returns the number of the lowest bit set, or -1 if no bits are set. }

function BSwapCardinal (const Value: Cardinal): Cardinal;
function BSwapInteger (const Value: Integer): Integer;

function IntToStr (const i: Integer): AnsiString;
{ Faster than the Delphi original IntToStr function. }

{$IFNDEF DELPHI5_UP}
procedure FreeAndNil (var Obj);
{ Frees an object reference and replaces the reference with nil.
  Warning: Do not pass a value for Obj if it is not an instance of TObject or one of its descendants. }
{$ENDIF}

procedure ZeroMem (const pBuff: Pointer; const Count: Integer);
{ Initializes block of memory with zeros }

function Min (const a, b: Integer): Integer; overload;
function Min (const a, b: Cardinal): Cardinal; overload;

function Max (const a, b: Integer): Integer; overload;
function Max (const a, b: Cardinal): Cardinal; overload;

function IntToHex (const i: Integer): AnsiString;

function ShiftKeyDown: Boolean;
{ Returns True if either one or both of the Shift Keys is pressed down. }

implementation

function LeftMostBit (const Value: Integer): Integer;
{ Returns the number of the highest bit which is set in the passed in value.
  Note that bits are numbered right to left starting with 0.
  For example, LeftMostBit (5) returns 2 (5 is 00000101 in binary).
  Returns the number of the highest bit set, or -1 if no bits are set. }
asm
  MOV ECX, EAX
  MOV EAX, -1
  BSR EAX, ECX
end;

function RightMostBit (const Value: Integer): Integer;
{ Returns the number of the lowest bit which is set in the passed in value.
  Note that bits are numbered right to left starting with 0.
  For example, RightMostBit (9) returns 2 (9 is 00001100 in binary).
  Returns the number of the lowest bit set, or -1 if no bits are set. }
asm
  MOV ECX, EAX
  MOV EAX, -1
  BSF EAX, ECX
end;

{ ********** }

function IntToHex (const i: Integer): AnsiString;
const
 HexChars           : array[0..15] of Char =
  ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F');
begin
 Result :=
  HexChars[(i and $F0000000) shr 28] +
  HexChars[(i and $0F000000) shr 24] +
  HexChars[(i and $00F00000) shr 20] +
  HexChars[(i and $000F0000) shr 16] +
  HexChars[(i and $0000F000) shr 12] +
  HexChars[(i and $00000F00) shr 8] +
  HexChars[(i and $000000F0) shr 4] +
  HexChars[(i and $0000000F)];
end     ;

{ ********** }

function BSwapCardinal (const Value: Cardinal): Cardinal; assembler;
asm
  BSWAP EAX
end;

{ ********** }

function BSwapInteger (const Value: Integer): Integer; assembler;
asm
  BSWAP EAX
end;

{ **************************************************************************** }

procedure BufferInit (var Buf: TBuffer);
begin
 with Buf do
  begin
   Buffer := nil;
   Size := 0;
   Position := 0;
  end;
end;

{ ********** }

procedure BufferAdd (const Source: Pointer; const SourceLength: Integer; var Dest: TBuffer);
var
 l, NewSize         : Integer;
 PCharS, PCharD     : PAnsiChar;
begin
 if (Source <> nil) and (SourceLength > 0) then
  with Dest do
   begin

    NewSize := Position + SourceLength;
    if NewSize > Size then
     begin
      Inc (NewSize, ((NewSize div 2) + 3) and -4);
      ReallocMem (Buffer, NewSize);
      Size := NewSize;
     end;

    PCharS := Source;
    Integer (PCharD) := Integer (Dest.Buffer) + Position;
    l := SourceLength;

    while l >= 4 do
     begin
      Cardinal (Pointer (PCharD)^) := Cardinal (Pointer (PCharS)^);
      Inc (PCharD, SizeOf (Cardinal));
      Inc (PCharS, SizeOf (Cardinal));
      Dec (l, SizeOf (Cardinal));
     end;

    case l of
     3:
      begin
       Word (Pointer (PCharD)^) := Word (Pointer (PCharS)^);
       //PCharD^ := PCharS^;
       //PCharD[ 1] := PCharS[ 1];
       //PCharD[ 2] := PCharS[ 2];
       Inc (PCharD, SizeOf (Word));
       Inc (PCharS, SizeOf (Word));
       PCharD^ := PCharS^;
      end;
     2:
      begin
       Word (Pointer (PCharD)^) := Word (Pointer (PCharS)^);
       //PCharD^ := PCharS^;
       //PCharD[ 1] := PCharS[ 1];
      end;
     1:
      PCharD^ := PCharS^;
    end;

    Inc (Position, SourceLength);
   end;
end;

{ ********** }

procedure BufferFree (var Buf: TBuffer);
begin
 with Buf do
  if Size > 0 then
   FreeMem (Buffer, Size);
end;

{ **************************************************************************** }

function IntToStr (const i: Integer): AnsiString;
{ Surprisingly faster than the Delphi original. }
begin
 Str (i, Result);
end;

{ ********** }

{$IFNDEF DELPHI5_UP}

procedure FreeAndNil (var Obj);
var
 o                  : TObject;
begin
 o := TObject (Obj);
 TObject (Obj) := nil;
 o.Free;
end;

{$ENDIF}

{ ********** }

procedure ZeroMem (const pBuff: Pointer; const Count: Integer);
asm
  MOV     ECX,EDX
  SAR     ECX,2
  JS      @@exit
  PUSH    EDI
  MOV     EDI,EAX { Point EDI to destination      }
  XOR     EAX,EAX
  REP     STOSD   { Fill count DIV 4 dwords       }
  MOV     ECX,EDX
  AND     ECX,3
  REP     STOSB   { Fill count MOD 4 bytes        }
  POP     EDI
@@exit:
end;

{ **************************************************************************** }
{ Minimum and Maximum functions.
{ **************************************************************************** }

function Min (const a, b: Integer): Integer;
begin
 if a < b then
  Result := a
 else
  Result := b;
end;

function Min (const a, b: Cardinal): Cardinal;
begin
 if a < b then
  Result := a
 else
  Result := b;
end;

{ ********** }

function Max (const a, b: Integer): Integer;
begin
 if a > b then
  Result := a
 else
  Result := b;
end;

function Max (const a, b: Cardinal): Cardinal;
begin
 if a > b then
  Result := a
 else
  Result := b;
end;

{ ********** }

function ShiftKeyDown: Boolean;
begin
 Result := (GetAsyncKeyState (VK_LSHIFT) < 0) or (GetAsyncKeyState (VK_RSHIFT) < 0);
end;

{ TDynBuffer }

destructor TDynBuffer.Destroy;
begin
 if FBuffer <> nil then FreeMem (FBuffer, FBufferSize);
end;

procedure TDynBuffer.AddAnsiChar (const c: AnsiChar);
begin
 if Position >= FBufferSize then
  begin
   FBufferSize := ((FBufferSize + (FBufferSize div 4) + 4) div 4) * 4;
   ReallocMem (FBuffer, FBufferSize);
  end;
 PAnsiChar (FBuffer)[Position] := c;
 Inc (Position);
end;

procedure TDynBuffer.Clear;
begin
 if FBuffer <> nil then
  begin
   FreeMem (FBuffer, FBufferSize);
   FBuffer := nil;
   FBufferSize := 0;
  end;
 Position := 0;
end;

function TDynBuffer.GetAsAnsiString: AnsiString;
begin
 SetString (Result, PAnsiChar (FBuffer), Position);
end;

end.

