{ Copyright Ralf Junker <ralfjunker@gmx.de> 2000 
  http://www.zeitungsjunge.de/delphi/ }

unit rjVectors;

interface

{$I rj.inc}

uses
 Classes;                               // Only needed for Streams

resourcestring
 SNotImplemented    = 'method not implemented';
 SMethodNotApplicable = 'method is not applicable';
 SIncompatibleClasses = 'incompatible classes';
 SErrorInParameters = 'error in parameters';
 SDataTooLarge      = 'data too large (%d)';
 SRangeError        = 'range check error (Index: %d, Count: %d)';
 SAssignError       = 'assign error';
 SDuplicateError    = 'duplicate values not allowed';
 SEmptyStack        = 'stack is empty';
 SEmptyQueue        = 'queue is empty';
 SStreamCorrupted   = 'stream is corrupted';
 SGetSizeError      = 'get size error';
 SSetSizeError      = 'set size error';
 SSeekError         = 'seek error (%d)';
 SValueExpected     = 'value expected';
 STooManyValues     = 'too many values';
 SReadAfterEnd      = 'read after end error';
 SFileCreateError   = 'can''t create file';
 SFileOpenError     = 'can''t open file';
 SFileReadError     = 'file read error';
 SFileWriteError    = 'file write error';
 SIllegalUseOfCRCStream = 'illegal use of CRC stream';
 SCRCError          = 'CRC error';
 SCanNotRollBack    = 'Can''t roll back';
 SWrongTextStreamFormat = 'wrong text stream format';
 SWrongVectorSize   = 'wrong vector size (%d)';
 SWrongMatrixSize   = 'wrong matrix size';
 SAlgorithmFailure  = 'algorithm failure';
 SCanNotConvertToInteger = 'Can''t convert ''%s'' to integer';
 SCanNotConvertToFloat = 'Can''t convert ''%s'' to float';
 SIllegalDateTime   = 'Illegal date/time ''%s''';
 SIllegalChar       = 'Illegal character ''%s''';
 SIllegalBool       = 'Illegal boolean value ''%s''';
 SInternalError     = 'internal error';
 SBOF               = 'BOF reached';
 SEOF               = 'EOF reached';
 SNoCurrentValue    = 'no current value';
 
const
 BASE_X_INIT_DEFAULT_VALUE = 0;
 CUSTOM_DEFAULT_VALUE = 0;
 
type
 
 TUntypedCompareFunc = function (const Item1, Item2): Integer;
 TUntypedMatchFunc = function (const Item1, Item2): Boolean;
 
 TVectorObject = class
 protected
  class procedure Error (const Msg: AnsiString; const Data: array of const);
 end;
 
 { ********** }
 
 TVector = class (TVectorObject)
 protected
  FVector: Pointer;
  FCount: Integer;
  procedure Grow (const Count: Integer); virtual; abstract;
  procedure SetCount (const NewCount: Integer); virtual;
  procedure InitMemory (const Index, Range: Integer); virtual; abstract;
 public
  {$IFDEF CheckFree}
  constructor Create;
  {$ENDIF}
  constructor CreateCopy (const Source: TVector);
  destructor Destroy; override;

  procedure Assign (const Source: TVector); virtual; abstract;

  property Count: Integer read FCount;  // write SetCount;

  procedure Clear; virtual;

  procedure Delete (const Index: Integer);
  procedure DeleteFirst;
  procedure DeleteLast;
  procedure DeleteRange (const Index, Count: Integer); virtual; abstract;

  procedure Exchange (const Index1, Index2: Integer); virtual; abstract;

  function IsEmpty: Boolean;
  function IsNotEmpty: Boolean;

  property Vector: Pointer read FVector;

  function ReadFromBinaryFile (const FileName: AnsiString): Boolean;
  function ReadFromBinaryStream (const Stream: TStream): Boolean; virtual; abstract;
  procedure WriteToBinaryFile (const FileName: AnsiString);
  procedure WriteToBinaryStream (const Stream: TStream); virtual; abstract;

  procedure Invert;
 end;
 
 { ********** }
 
implementation

uses
 Consts,
 rjExceptions
 {$IFDEF CheckFree}, rjCheckFree{$ENDIF};

{ **************************************************************************** }
{ Class TVectorObject
{ **************************************************************************** }

class procedure TVectorObject.Error (const Msg: AnsiString; const Data: array of const);
 function ReturnAddr: Pointer;
 asm
  mov eax, [ebp + 4]
 end;
begin
 raise EVector.CreateFmt ('Class ' + ClassName + ': ' + Msg, Data)At ReturnAddr;
end;

{ **************************************************************************** }
{ Class TVector
{ **************************************************************************** }

{$IFDEF CheckFree}
constructor TVector.Create;
begin
 RegisterObjectCreate (Self);
 inherited Create;
end;
{$ENDIF}

{ ********** }

constructor TVector.CreateCopy (const Source: TVector);
begin
 Create;
 Self.Assign (Source);
end;

{ ********** }

destructor TVector.Destroy;
begin
 Clear;
 inherited Destroy;
 {$IFDEF CheckFree}
 RegisterObjectFree (Self);
 {$ENDIF}
end;

{ ********** }

procedure TVector.Clear;
begin
 SetCount (0);
end;

{ ********** }

procedure TVector.Delete (const Index: Integer);
begin
 DeleteRange (Index, 1);
end;

{ ********** }

procedure TVector.DeleteFirst;
begin
 DeleteRange (0, 1);
end;

{ ********** }

procedure TVector.DeleteLast;
begin
 DeleteRange (FCount - 1, 1);
end;

{ ********** }

function TVector.IsEmpty: Boolean;
begin
 Result := FCount <= 0;
end;

{ ********** }

function TVector.IsNotEmpty: Boolean;
begin
 Result := FCount > 0;
end;

{ ********** }

function TVector.ReadFromBinaryFile (const FileName: AnsiString): Boolean;
var
 Stream             : TStream;
begin
 Stream := TFileStream.Create (FileName, $0000 {fmOpenRead, defined in SysUtils} or $0020 {fmShareDenyWrite, defined in SysUtils});
 try
  Result := ReadFromBinaryStream (Stream);
 finally
  Stream.Free;
 end;
end;

{ ********** }

procedure TVector.SetCount (const NewCount: Integer);
var
 DeltaCount, OldCount: Integer;
begin
 {$IFDEF RangeChecking}
 if NewCount < 0 then Error (SListCountError, [NewCount]);
 {$ENDIF}
 if NewCount <> FCount then
  begin
   OldCount := FCount;
   DeltaCount := NewCount - OldCount;

   if DeltaCount > 0 then
    begin
     Grow (DeltaCount);                 // Hint: Grow sets FCount
     InitMemory (OldCount, DeltaCount);
    end
   else
    DeleteRange (NewCount, -DeltaCount);
  end;
end;

{ ********** }

procedure TVector.WriteToBinaryFile (const FileName: AnsiString);
var
 Stream             : TStream;
begin
 Stream := TFileStream.Create (FileName, fmCreate);
 try
  WriteToBinaryStream (Stream);
 finally
  Stream.Free;
 end;
end;

{ ********** }

procedure TVector.Invert;
var
 i, N               : Integer;
begin
 N := FCount;
 for i := 0 to N div 2 - 1 do
  begin
   Dec (N);
   Exchange (i, N);
  end;
end;

end.

